// Plot_Table_Columns.java

import ij.*;
import ij.gui.*;
import ij.text.*;
import ij.util.*;
import ij.measure.ResultsTable;
import ij.plugin.*;
import ij.plugin.PlugIn;
import ij.Prefs;
import ij.process.*;
import ij.io.*;

import javax.swing.*;
import javax.swing.border.TitledBorder;
import javax.swing.event.*;
import java.util.Hashtable;

import java.awt.*;
import java.awt.geom.*;
import java.util.*;
import java.io.*;
import java.awt.event.*;

import java.lang.Math.*;
import java.text.*;

// astroj ------
import astroj.*;
// ------ astroj


/**
 * This plugin plots any number of columns from a Results or MeasurementTable.  This plugin
 * is based on the Plot_Columns plugin by F.V. Hessman.
 *
 * @author F.V. Hessman, Georg-August-Universitaet Goettingen
 * @date 2009-FEB-09
 * @version 1.0
 *
 * @author K.A. Collins, University of Louisville, Kentucky USA
 * @date 2010-JUN-09
 * @version 2.1
 * @changes New GUI interface, support for multiple column plotting, error plotting
 * support, interactive shifting and scaling, titles, subtitles, and legends, custom,
 * semi-custom, and auto scaling
 *
 *
 * 
 */

/**
 * This plugin plots any two columns from a Results or MeasurementTable.
 * @author F.V. Hessman, Georg-August-Universitaet Goettingen
 * @date 2009-FEB-09
 * @version 1.0
 *
 * To create version outside of the Astronomy plugins, comment in the "// plain ..." lines and comment out the "// astroj ..." lines
 */
public class MultiPlot_ implements PlugIn
        {
        static String title;
        static double titlePosX;
        static double titlePosY;
        static String subtitle;
        static double subtitlePosX;
        static double subtitlePosY;
        static double legendPosX;
        static double legendPosY;
        static boolean button2Drag;
        static String xlabeldefault;
        static int xcolumndefault;
        static int plotSizeX;
        static int plotSizeY;
        static double pltMinX;
        static double pltMinY;
        static double pltMaxX;
        static double pltMaxY;
        static double plotMinX;
        static double plotMinY;
        static double plotMaxX;
        static double plotMaxY;
        static int xExponent;
        static int yExponent;
        static int mmagrefs;
        static int firstCurve;
        static String xLegend;
        static String yLegend;
        static boolean legendLeft;
        static boolean legendRight;
        static boolean useTitle;
        static boolean useSubtitle;
        static boolean useXCustomName;
        static boolean useYCustomName;
        static boolean useXColumnName;
        static boolean useYColumnName;
        static boolean xTics;
        static boolean yTics;
        static boolean xGrid;
        static boolean yGrid;
        static boolean xNumbers;
        static boolean yNumbers;
        static boolean plotAutoMode;
        static boolean nextPanel;
        static boolean canceled;
        static boolean waitSecs;
        static boolean openDataSetWindow;
        static boolean rememberWindowLocations;
        static boolean autoScaleX;
        static boolean useFirstX;
        static boolean autoScaleY;
        static boolean showToolTips;
        static boolean showXScaleInfo;
        static boolean showYScaleInfo;
        static boolean showYShiftInfo;
        static boolean showLScaleInfo;
        static boolean showLShiftInfo;
        static boolean showYBinInfo;
        static boolean showLBinInfo;
        static boolean showYmmagInfo;
        static boolean showLmmagInfo;
        static boolean showYSymbolInfo;
        static boolean showLSymbolInfo;
        static boolean useTwoLineLegend;
        static boolean useWideDataPanel;
        static boolean subframeWasShowing;
        static boolean newPlotWindow;
        static boolean useDefaultSettings;
        static boolean tableHasText;
        static boolean setSubpanelVisible;
        static boolean usePixelScale;
        static boolean showVMarker1;
        static boolean showVMarker2;
        static boolean leftDragReleased;
        static boolean prioritizeColumns;
        static boolean invertYAxis;
        static int invertYAxisSign;
        static boolean negateMag;
        static boolean selectAnotherTableCanceled;
        static boolean panelShiftDown;
        static boolean panelControlDown;
        static boolean panelAltDown;
        static boolean keepFileNamesOnAppend;
        static boolean updatePlotRunning;

        static double pixelScale;
        static double zoom;
        static double xMin;
        static double xMax;
        static double xWidth;
        static double yMin;
        static double yMax;
        static double xOffset;
        static double yOffset;
        static double xautoscalemin;
        static double xautoscalemax;
        static double yautoscalemin;
        static double yautoscalemax;
        static double yRange;
        static double dashLength;
        static double numDashes;
        static double mouseX;
        static double mouseY;
        static double totalPanOffsetX;
        static double totalPanOffsetY;
        static double newPanOffsetX;
        static double newPanOffsetY;

        static int cur;
        static int plotOptions=0;
        static int mainFrameLocationX;
        static int mainFrameLocationY;
        static int subFrameLocationX;
        static int subFrameLocationY;
        static int plotFrameLocationX;
        static int plotFrameLocationY;

        static int excludedHeadSamples;
        static int excludedTailSamples;
        static int holdExcludedHeadSamples;
        static int holdExcludedTailSamples;
        static int excludedHeadSamplesStep;
        static int excludedTailSamplesStep;

        static double xWidthStep;
        static double xMaxStep;
        static double xMinStep;
        static double yMaxStep;
        static double yMinStep;
        static double yMid;
        static double firstXmin;
        static int plotSizeXStep;
        static int plotSizeYStep;
        static int nearestLine;
        static double nearestX;
        static int boldedDatum;
        static boolean useBoldedDatum;

        static String vMarker1TopText;
        static String vMarker1BotText;
        static String vMarker2TopText;
        static String vMarker2BotText;
        static String plotSuffix;
        static String configSuffix;
        static String dataSuffix;
        static String appendDestinationSuffix;
        static String appendSourceSuffix;
        static String combinedTableName;
        static String requestedTableName;

        static int n;
        static int[] nn;
        static int[] nnr;
        static String[] xlabel2;
        static int selectedRowStart, selectedRowEnd;
        static java.util.List<String> deletedRowList;
        static java.io.File[] dragAndDropFiles;
        static double[][] x;
        static double[][] y;
        static double[][] yerr;
        static double[][] yop;
        static double[][] yoperr;

        static double[][] xc1;
        static double[][] xc2;
        static double[][] yc1;
        static double[][] yc2;
        static double xMultiplierFactor;
        static double yMultiplierFactor;
        static int excluded;

        static String xlab;
        static String ylab;
        static String priorityColumns;
//        static String priorityErrColumns;
        static double dx;
        static double dy;
        static double xPlotMin;
        static double xPlotMax;
        static double yPlotMin;
        static double yPlotMax;


        static int startDragX, startDragY, endDragX, endDragY;
        static int startDragSubImageX, startDragSubImageY;
        static int startDragScreenX, startDragScreenY;
        static int newPositionX, newPositionY;

        static String temptextfield;
        static Integer tempIntField;

        static Plot plot;
        static JFrame mainFrame;
        static JFrame subFrame;
        static JFrame getMaxCurvesFrame;
//        static Frame plotFrame;
        static Panel plotpanel;
        static JPanel mainpanel;
        static JScrollPane mainscrollpane;
        static JPanel mainsubpanel;
        static JScrollPane subscrollpane;

        static JMenuBar mainmenubar;
        static JMenu filemenu;
        static JMenu preferencesmenu;
        static JCheckBoxMenuItem opendatasetCB;
        static JCheckBoxMenuItem usewidedataCB;
        static JCheckBoxMenuItem rememberwindowlocationsCB;
        static JCheckBoxMenuItem showtooltipsCB;
        static JCheckBoxMenuItem usedefaultsettingsCB;
        static JCheckBoxMenuItem usepixelscaleCB;
        static JCheckBoxMenuItem useBoldedDatumCB;
        static JMenuItem changepixelscalemenuitem, setverticalmarkertextmenuitem;
        static JMenuItem savedatamenuitem, saveimagemenuitem, savedataconfigmenuitem, saveplotconfigmenuitem,
                         opendatamenuitem, openplotconfigmenuitem, opendataconfigmenuitem,
                         opentablemenuitem;
        static JCheckBoxMenuItem usePriorityColumnsCB;
        static JMenuItem changePriorityColumnsmenuitem;
        static JMenuItem changemaxdatalengthmenuitem;

        static JTextField titleField, subtitleField, legendField, xlegendField, ylegendField;
        static JLabel statusLabel, xcolumnlabel, ycolumnslabel, ycolumnlabel, operatorlabel;
        static JLabel xcolumnlabelsub, ycolumnlabelsub, operatorlabelsub;
        static JLabel opcolumnlabel, markerlabel, markercolorlabel;
        static JLabel opcolumnlabelsub, markerlabelsub, markercolorlabelsub;
        static JSlider titlePosYSlider;
        static JSlider titlePosXSlider;
        static JSlider subtitlePosYSlider;
        static JSlider subtitlePosXSlider;
        static JSlider legendPosYSlider;
        static JSlider legendPosXSlider;
        static JComboBox xdatacolumndefault;

        static JRadioButton autoxButton, firstxButton, customxButton;
        static JRadioButton autoyButton, firstyButton, customyButton;

        static SpinnerModel mmagrefsmodel;
        static JSpinner mmagrefsspinner;
        static SpinnerModel maxcurvesmodel;
        static JSpinner maxcurvesspinner;

        static JSpinner xwidthstepspinner,xmaxstepspinner, xminstepspinner,ymaxstepspinner,yminstepspinner;
        static JSpinner plotwidthstepspinner, plotheightstepspinner;

        static JPopupMenu xwidthsteppopup, xmaxsteppopup, xminsteppopup,ymaxsteppopup,yminsteppopup;
        static JPopupMenu plotwidthsteppopup, plotheightsteppopup, trimdataheadsteppopup, trimdatatailsteppopup;
        static JPopupMenu xaxispopup, yaxispopup, legendpopup;
        static JPopupMenu vmarker1steppopup, vmarker2steppopup;

        static SpinnerModel xwidthstepmodel, xmaxstepmodel, xminstepmodel,ymaxstepmodel,yminstepmodel;
        static SpinnerModel plotwidthstepmodel, plotheightstepmodel;
        static SpinnerModel xwidthspinnermodel, xminspinnermodel, xmaxspinnermodel;
        static SpinnerModel yminspinnermodel, ymaxspinnermodel, plotheightspinnermodel, plotwidthspinnermodel;
        static SpinnerModel xmultipliermodel, ymultipliermodel;
        static SpinnerModel trimdataheadmodel, trimdatatailmodel, trimdataheadstepmodel, trimdatatailstepmodel;
        static SpinnerModel vmarker1stepmodel, vmarker2stepmodel, vmarker1spinnermodel, vmarker2spinnermodel;

        static JSpinner xwidthspinner, xminspinner, xmaxspinner, xmultiplierspinner, ymultiplierspinner;
        static JSpinner yminspinner, ymaxspinner, plotheightspinner, plotwidthspinner;
        static JSpinner trimdataheadspinner, trimdatatailspinner, trimdataheadstepspinner, trimdatatailstepspinner;
        static JSpinner vmarker1stepspinner, vmarker2stepspinner, vmarker1spinner, vmarker2spinner;


        static JCheckBox usextickscheckbox, useytickscheckbox, usexgridcheckbox;
        static JCheckBox useygridcheckbox, usexnumberscheckbox, useynumberscheckbox;
        static JCheckBox showVMarker1CB, showVMarker2CB;

        static JButton moreybutton, closebutton, grabautoxbutton, grabautoybutton, updateplotbutton;

        static String tableName;
        static String[] columns, unfilteredColumns, oldUnfilteredColumns;
        static String[] columnswd;//, columnsErr;
        
        static MeasurementTable table;
        static TextPanel tpanel;
        static JLabel dummylabel1, binsizespinnerlabel, mmagrefsspinnerlabel, dummylabel4;

        static ImageWindow plotWindow;
        static ImagePlus plotImage;
//        static ImageProcessor ip;
        static ImageCanvas plotImageCanvas;
        static OverlayCanvas plotOverlayCanvas;
        static Panel plotbottompanel;
        static Label plotcoordlabel;
        static Vector<Roi> list;
        static GeneralPath path;
        static Roi roi;
        static TitledBorder dataSectionBorder = BorderFactory.createTitledBorder("Data ()");;        
        

        static int maxCurves;
        static int maxColumnLength;
        static double vMarker1Step=0.01, vMarker2Step=0.01;
        static double vMarker1Value=0.5, vMarker2Value=0.7;
        static String[] xlabel;
        static String[] ylabel;
        static int[] xcolumn;
        static int[] ycolumn;
        static int[] errcolumn;
        static int[] operrcolumn;
        static int[] xc1column, yc1column,xc2column,yc2column;
        static String[] oplabel;
        static int[] opcolumn;
        static String[] errlabel;
        static String[] operrlabel;
        static boolean[] showErrors;
        static boolean[] showOpErrors;
        static boolean[] lines;
        static int[] marker;
        static Color[] color;
        static int[] markerIndex;
        static int[] colorIndex;
        static int[] operatorIndex;
        static String[] cblabels;
        static boolean[] moreOptions;
        static double[] autoScaleFactor;
        static double[] autoScaleStep;
        static double[] autoShiftFactor;
        static double[] autoShiftStep;
        static double[] customScaleFactor;
        static double[] customScaleStep;
        static double[] customShiftFactor;
        static double[] customShiftStep;
        static double[] baseline;
        static double[] subtotalScaleFactor;
        static double[] subtotalShiftFactor;
        static double[] totalScaleFactor;
        static double[] totalShiftFactor;
        static double[] yWidth;
        static double[] xMinimum;
        static double[] xMaximum;
        static double[] yMinimum;
        static double[] yMn;
        static double[] yMaximum;
        static double[] yMx;
        static double[] yMidpoint;
        static double[] nullX = null;
        static double[] nullY = null;

        static boolean[] mmag;
        static boolean[] plotY;
        static boolean[] useColumnName;
        static boolean[] useLegend;
        static boolean[] force;
        static boolean[] ASInclude;

        static int[] binSize;

        static String[] legend;
        static String[] xyc1label;
        static String[] xyc2label;

        static JCheckBox[] usecurvebox;
        static JCheckBox[] errorcolumnbox;
        static JCheckBox[] operrorcolumnbox;
        static JCheckBox[] usebinbox;
        static JCheckBox[] usemmagbox;
        static JCheckBox[] uselinesbox;
        static JCheckBox[] autoscaleincludebox;
        static JCheckBox[] forcebox;

        static JComboBox[] xdatacolumn;
        static JComboBox[] ydatacolumn;
        static JComboBox[] operatorselection;
        static JComboBox[] operatorcolumn;
        static JComboBox[] errorcolumn;
        static JComboBox[] operrorcolumn;
        static JComboBox[] markersymbolselection;
        static JComboBox[] markercolorselection;

        static SpinnerModel[] binsizespinnermodel;
        static JSpinner[] binsizespinner;
        static JPanel[] morelegendradiopanelgroup;
        static JPanel[] autoscalepanelgroup;
        static JPanel[] customscalepanelgroup;
        static JRadioButton[] legendnoneButton;
        static JRadioButton[] legendcustomNameButton;
        static JRadioButton[] legendcolumnNameButton;
        static JRadioButton[] autoscaleButton;
        static JRadioButton[] customscaleButton;
        static ButtonGroup[] morelegendRadioGroup;
        static ButtonGroup[] autoscaleRadioGroup;
        static JTextField[] morelegendField;
        static SpinnerModel[] autoshiftmodel;
        static SpinnerModel[] autoshiftstepmodel;
        static SpinnerModel[] autoscalemodel;
        static SpinnerModel[] autoscalestepmodel;
        static SpinnerModel[] customshiftmodel;
        static SpinnerModel[] customshiftstepmodel;
        static SpinnerModel[] customscalemodel;
        static SpinnerModel[] customscalestepmodel;
        static JSpinner[] autoshiftspinner;
        static JSpinner[] autoshiftstepspinner;
        static JSpinner[] autoscalespinner;
        static JSpinner[] autoscalestepspinner;
        static JSpinner[] customshiftspinner;
        static JSpinner[] customshiftstepspinner;
        static JSpinner[] customscalespinner;
        static JSpinner[] customscalestepspinner;
        static JPopupMenu[] autoscalesteppopup;
        static JPopupMenu[] autoshiftsteppopup;
        static JPopupMenu[] customscalesteppopup;
        static JPopupMenu[] customshiftsteppopup;
        static JPanel[] autoscalesteppanel;
        static JPanel[] customscalesteppanel;
        static JPanel[] autoshiftsteppanel;
        static JPanel[] customshiftsteppanel;
        static JLabel[] autoscalesteplabel;
        static JLabel[] autoshiftsteplabel;
        static JLabel[] customscalesteplabel;
        static JLabel[] customshiftsteplabel;
        static JLabel[] curvelabel;
        static JLabel[] othercurvelabel;
        static JPanel[] grabautopanel;
        static JButton[] grabautobutton;
        static ImageIcon forceIcon;
        static ImageIcon configureIcon;
        static ImageIcon helpIcon;
        static ImageIcon plotIcon;
        static JButton legendconfigureButton;
        static JButton xlegendconfigureButton;
        static JButton ylegendconfigureButton;
        static JButton legendslabelconfigureButton;
        static JPanel legendgroup;
        static JPanel mainpanelgroupa;
        

        static JSpinner.NumberEditor[] customscaleeditor;
        static DecimalFormat[] customscaleformat;
        static JSpinner.NumberEditor[] customshifteditor;
        static DecimalFormat[] customshiftformat;

        static DecimalFormat fourPlaces;

        static String[] markers;
        static String[] colors;

        static String[] operators;

        static String[] opSymbol;

        static String[] spinnerscalelist;

        static String[] integerspinnerscalelist;
        
        static TimerTask delayedUpdateTask;
        static java.util.Timer delayedUpdateTimer;        

       
    
	public void run(String inTableName)
		{
        //SET DEFAULT SYSTEM LOOK AND FEEL
        if (IJ.isWindows())
                {
//                        try {UIManager.setLookAndFeel("javax.swing.plaf.metal.MetalLookAndFeel");}
                try {UIManager.setLookAndFeel("com.sun.java.swing.plaf.windows.WindowsLookAndFeel");}
//                        try {UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());}
                catch (Exception e) { }
                }
        else if (IJ.isLinux())
                {
                try {UIManager.setLookAndFeel("javax.swing.plaf.metal.MetalLookAndFeel");}
                catch (Exception e) { }
                }
//                    try {UIManager.setLookAndFeel("com.sun.java.swing.plaf.windows.WindowsLookAndFeel");}
//                    try {UIManager.setLookAndFeel("com.sun.java.swing.plaf.nimbus.NimbusLookAndFeel");}
//                    try {UIManager.setLookAndFeel("com.sun.java.swing.plaf.motif.MotifLookAndFeel");}
        else if (IJ.isMacOSX())
                {

                try {UIManager.setLookAndFeel("javax.swing.plaf.metal.MetalLookAndFeel");}
//                        try {UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());}
                catch (Exception e) { }
                }

        plotAutoMode = Prefs.get("plot.automode",plotAutoMode);
        if (plotAutoMode)
            {
            updatePlot();
            plotAutoMode = false;
            Prefs.set("plot.automode",false);
            return;
            }
        
        if (mainFrame == null || !mainFrame.isShowing())
            {
            if (mainFrame != null)
                {
                mainFrame.dispose();
                subFrame.dispose();
                mainFrame=null;
                subFrame=null;
                }
            initializeVariables();
            setupArrays();
            useDefaultSettings = Prefs.get("plot.useDefaultSettings", useDefaultSettings);
            if (useDefaultSettings)
                    {
                    useDefaultSettings = false;
                    Prefs.set("plot.useDefaultSettings", useDefaultSettings);
                    }
            else
                    getPreferences();
            }

		// FIND ALL TABLES
        if (inTableName != null && !inTableName.equals("")) requestedTableName = inTableName;
        findTables(false);
        }


    static void findTables(boolean forceUpdate)
        {
        String[] tables = MeasurementTable.getMeasurementTableNames();
        if (tables == null || tables.length == 0)
            {
            makeDummyTable();
            }
        else
            {
            //filter out B&C, Log, Errors windows if they exist
            int j = 0;
            for (int i=0; i < tables.length; i++)
                if (tables[i].equals("B&C") || tables[i].startsWith("Log") || tables[i].equals("Errors"))
                    j++;
            String[] filteredTables = new String[tables.length-j];
            j = 0;
            for (int i=0; i < tables.length; i++)
                if (!tables[i].equals("B&C") && !tables[i].startsWith("Log") && !tables[i].equals("Errors"))
                    {
                    filteredTables[j] = tables[i];
                    j++;
                    }
            if (requestedTableName != null)
                {
                for (int i=0; i < filteredTables.length; i++)
                    {
                    if (requestedTableName.equals(filteredTables[i]))
                        {
                        tableName = requestedTableName;
                        requestedTableName = null;
                        setTable(MeasurementTable.getTable(tableName), forceUpdate);
                        break;
                        }
                    }
                if (requestedTableName != null)
                    {
                    requestedTableName = null;
                    makeDummyTable();
                    }
                }
            else if (filteredTables.length == 0)
                {
                makeDummyTable();
                }
            else if (filteredTables.length == 1)
                {
                tableName = filteredTables[0];
                setTable(MeasurementTable.getTable(tableName), forceUpdate);
                }
            else                            // IF MORE THAN ONE, ASK WHICH TABLE SHOULD BE USED
                {
                GenericDialog gd = new GenericDialog("Plot Table Columns");
                gd.addMessage ("Select table to be plotted.");
                gd.addChoice ("Table to be plotted",filteredTables,"Measurements");
                gd.showDialog();
                if (gd.wasCanceled())
                    {
                    makeDummyTable();
                    }
                else
                    {
                    tableName = gd.getNextChoice();
                    setTable(MeasurementTable.getTable(tableName), forceUpdate);
                    }
                }
            }
        }
    static public void setTable(MeasurementTable inTable, boolean forceUpdate)
        {
        table = inTable;
        if (table == null)
            {
            makeDummyTable();
            }
        else
            {
            tableName = table.shortTitle();
            tpanel = MeasurementTable.getTextPanel(MeasurementTable.longerName(tableName));
            dataSectionBorder.setTitle("Data ("+MeasurementTable.shorterName(tableName) +")");                 
            unfilteredColumns = table.getColumnHeadings().split("\t");
            if (unfilteredColumns == null || unfilteredColumns.length == 0)
                {
                IJ.showMessage ("No data columns to plot.");
                makeDummyTable();
                }
            else
                {
                updateColumnLists();
                Prefs.set("plot2.tableName",tableName);
                finishSetup(forceUpdate);
                }
            }
        }
    
    static public MeasurementTable getTable()
        {
        return table;
        }

    static void finishSetup(boolean forceUpdate)
        {
        if (mainFrame == null || !mainFrame.isShowing())
            showMainJPanel();
//        if (!Arrays.equals(unfilteredColumns, oldUnfilteredColumns))
//            {
//            IJ.log("columns are different");
//            for (int i=0; i<(oldUnfilteredColumns==null?0:oldUnfilteredColumns.length); i++)
//                IJ.log("oldUnfilteredColumn["+i+"]="+oldUnfilteredColumns[i]);
//            for (int i=0; i<(unfilteredColumns==null?0:unfilteredColumns.length); i++)
//                IJ.log("unfilteredColumn["+i+"]="+unfilteredColumns[i]);              
//            }
        else if (forceUpdate || !Arrays.equals(unfilteredColumns, oldUnfilteredColumns))
            updatePanels();
        oldUnfilteredColumns = unfilteredColumns.clone();
        updatePlot();
		}

//------------------------------------------------------------------------------

static public void makeDummyTable()
        {
        tableName = "No Table Selected";
        table = null; //MeasurementTable.getTable(tableName);
        dataSectionBorder.setTitle("Data ("+MeasurementTable.shorterName(tableName) +")");
        unfilteredColumns = new String[1];
        unfilteredColumns[0] = "";
        columns = unfilteredColumns.clone();
//        columnsErr = unfilteredColumns.clone();
        columnswd = unfilteredColumns.clone();
        finishSetup(true);
//        Prefs.set("plot2.tableName",tableName);
        }

static public void updateColumnLists()
        {
        int j = 0;
        for (int i=0; i < unfilteredColumns.length; i++)
                if (unfilteredColumns[i].equals("Label") || unfilteredColumns[i].equals("image"))
                       j++;
        columns = new String[unfilteredColumns.length - j];
        j = 0;
        for (int i=0; i < unfilteredColumns.length; i++)
                if (!unfilteredColumns[i].equals("Label") && !unfilteredColumns[i].equals("image") )
                        {
                        columns[j] = unfilteredColumns[i];
                        j++;
                        }
//        columnsErr = columns.clone();
        if (prioritizeColumns) sortColumns();
        columnswd = new String[columns.length+1];
        columnswd[0] = columns[0];
        columnswd[1] = "default";
        for (int i=2; i< columnswd.length; i++)
            columnswd[i]=columns[i-1];
        }

    static public void sortColumns()
        {
        String[] pcols = priorityColumns.split(",");
        if (pcols.length < 1 || columns.length < 2) return;
        String[] cols = columns.clone();
        int j = 1;
        for (int k=0; k<pcols.length; k++)
            {
            for (int i=1; i<cols.length; i++)
                {
                if (cols[i].toLowerCase().startsWith(pcols[k].toLowerCase()))
                    {
                    columns[j]=cols[i];
                    cols[i]="";
                    j++;
                    }

                }
            }
        for (int i=1; i<cols.length; i++)
            {
            if (!cols[i].equals(""))
                {
                columns[j]=cols[i];
                j++;
                }

            }

//        pcols = priorityErrColumns.split(",");
//        if (pcols.length < 1 || columnsErr.length < 2) return;
//        cols = columnsErr.clone();
//        j = 1;
//        for (int k=0; k<pcols.length; k++)
//            {
//            for (int i=1; i<cols.length; i++)
//                {
//                if (cols[i].toLowerCase().startsWith(pcols[k].toLowerCase()))
//                    {
//                    columnsErr[j]=cols[i];
//                    cols[i]="";
//                    j++;
//                    }
//
//                }
//            }
//        for (int i=1; i<cols.length; i++)
//            {
//            if (!cols[i].equals(""))
//                {
//                columnsErr[j]=cols[i];
//                j++;
//                }

//            }

        }


    static public void updatePanels()
        {
        if(mainFrame.isShowing())
            {
            if(subFrame.isShowing())
                {
                subFrameLocationX = subFrame.getLocation().x;
                subFrameLocationY = subFrame.getLocation().y;
                Prefs.set("plot2.subFrameLocationX",subFrameLocationX);
                Prefs.set("plot2.subFrameLocationY",subFrameLocationY);
                }
            subFrame.dispose();
            mainFrameLocationX = mainFrame.getLocation().x;
            mainFrameLocationY = mainFrame.getLocation().y;
            Prefs.set("plot2.mainFrameLocationX",mainFrameLocationX);
            Prefs.set("plo2.mainFrameLocationY",mainFrameLocationY);
            mainFrame.dispose();
            }
        showMainJPanel();
        oldUnfilteredColumns = unfilteredColumns.clone();
        }

    static public void setPlotAutoMode(boolean mode)
        {
        plotAutoMode = mode;
        }
    
    static public void clearPlot()
        {
        if (plot != null)
            {
            plotImage = WindowManager.getImage("Plot of "+tableName);
            if (plotImage != null)
                {
                plot = new Plot ("Plot of "+tableName,xlab,ylab,nullX,nullY,plotOptions);
                plot.setSize(plotSizeX, plotSizeY);
                plot.setLimits (plotMinX, plotMaxX, plotMinY, plotMaxY);
                plot.setJustification(plot.CENTER);
                if (useTitle)
                        {
                        plot.changeFont(new java.awt.Font("Dialog",java.awt.Font.PLAIN,18));
                        plot.addTitle(titlePosX, titlePosY/plotSizeY>1?1:titlePosY/plotSizeY,title);
                        }
                if (useSubtitle)
                        {
                        plot.changeFont(new java.awt.Font("Dialog",java.awt.Font.PLAIN,14));
                        plot.addTitle(subtitlePosX, subtitlePosY/plotSizeY>1?1:subtitlePosY/plotSizeY,subtitle);
                        }
                plot.draw();
                ImageProcessor ip = plot.getProcessor();
                plotImage.setProcessor("Plot of "+tableName,ip);
                }
            }
        else
            {
            IJ.showStatus("No plot to clear");
            }
//        IJ.log("deleting rows");
//        for (int i=table.getCounter(); i>=0; i--)
//            {
//            IJ.log("deleting row "+i);
//            table.deleteRow(i);
//            }
//        table.show();
        }
    
	/**
	 * Starts timer which handles spinner arrow clicks and holds.
	 */
	static protected void startDelayedUpdateTimer ()
        {
		try	{
            if (delayedUpdateTimer != null) delayedUpdateTimer.cancel();
            if (delayedUpdateTask != null) delayedUpdateTask.cancel();
            
			delayedUpdateTask = new TimerTask()
                {
                public void run()
                    {
                    updatePlot();
//                    IJ.log("delayed updatePlot() completed");
                    }
                };
            delayedUpdateTimer = new java.util.Timer();
            delayedUpdateTimer.schedule(delayedUpdateTask,200);
            }
        catch(Exception e)
            {
            IJ.showMessage ("Error starting delayed plot update timer : "+e.getMessage());
            }
        }    


//------------------------UPDATEPLOT()---------------------------------------

        static public void updatePlot()
                {
                if (updatePlotRunning)
                    {
//                    IJ.log("starting delayed updatePlot() timer");
                    startDelayedUpdateTimer();
                    return;
                    }
                updatePlotRunning = true;
//                table = MeasurementTable.getTable (tableName);
//                tpanel = MeasurementTable.getTextPanel(MeasurementTable.longerName(tableName));
                
                if (tableName.equals("No Table Selected"))
                    {
                    updatePlotRunning = false;
                    return;
                    }
                
                n = table.getCounter();
                if (n < 1)
                        {
//                        clearPlot();
                        updatePlotRunning = false;
                        return;
                        }
                if (n > maxColumnLength)
                    {
                    maxColumnLength = n > 2*maxColumnLength ? n : 2*maxColumnLength;
                    setupDataBuffers();
                    }
                unfilteredColumns = table.getColumnHeadings().split("\t");
                if (!Arrays.equals(unfilteredColumns, oldUnfilteredColumns))
                    {
//                    IJ.log("arrays not equal during update");
//                    for (int i=0; i<(oldUnfilteredColumns==null?0:oldUnfilteredColumns.length); i++)
//                        IJ.log("oldUnfilteredColumn["+i+"]="+oldUnfilteredColumns[i]);
//                    for (int i=0; i<(unfilteredColumns==null?0:unfilteredColumns.length); i++)
//                        IJ.log("unfilteredColumn["+i+"]="+unfilteredColumns[i]);                    
                    updateColumnLists();
                    updatePanels();
                    }

                savePreferences();

                for (int curve=0; curve<maxCurves; curve++)
                        {
                        if (binSize[curve] < 1)
                            {
                            binSize[curve] = 1;
                            binsizespinner[curve].setValue(1);
                            }
                        }

                holdExcludedHeadSamples = excludedHeadSamples;
                holdExcludedTailSamples = excludedTailSamples;
                if (excludedHeadSamples + excludedTailSamples >= n)  //Handle case for more samples excluded than in dataset
                    {
                    excludedHeadSamples = excludedHeadSamples < n ? excludedHeadSamples : n - 1;
                    excludedTailSamples = n - excludedHeadSamples - 1;
                    }
                excluded = excludedHeadSamples + excludedTailSamples;

                int magSign = negateMag?1:-1;
                for (int i=0;i<maxCurves;i++)
                        {
                        marker[i] = markerOf(markerIndex[i]);
                        color[i] = colorOf(colorIndex[i]);
                        }

                // INITIALIZE BASELINE FOR RELATIVE MMAG CALCULATIONS
                for (int i=0; i<maxCurves;i++)
                        {
                        baseline[i]=0;
                        }
                // GET DATA

                if (n > x[0].length)
                    {
                    nn = new int[maxCurves];
                    xlabel2 = new String[maxCurves];
                    x = new double[maxCurves][n];
                    y = new double[maxCurves][n];
                    yerr = new double[maxCurves][n];
                    yop = new double [maxCurves][n];
                    yoperr = new double [maxCurves][n];

                    xc1 = new double [maxCurves][n];
                    xc2 = new double [maxCurves][n];
                    yc1 = new double [maxCurves][n];
                    yc2 = new double [maxCurves][n];
                    }

                xMultiplierFactor = Math.pow(10, (double)xExponent);
                yMultiplierFactor = Math.pow(10, (double)yExponent);

                for (int curve=0; curve<maxCurves; curve++)
                        {
                        nn[curve] = (n-excluded)/binSize[curve];
                        nnr[curve] = (n-excluded)%binSize[curve];
                        if (nnr[curve]>0) nn[curve]++;
//                        x[curve] = new double[nn[curve]];
//                        y[curve] = new double[nn[curve]];
//                        yerr[curve] = new double[nn[curve]];
//                        yop[curve] = new double[nn[curve]];
//                        yoperr[curve] = new double[nn[curve]];
//                        xc1[curve] = new double[nn[curve]];
//                        xc2[curve] = new double[nn[curve]];
//                        yc1[curve] = new double[nn[curve]];
//                        yc2[curve] = new double[nn[curve]];
                        }
                for (int curve=0; curve<maxCurves; curve++)
                        {
                        if (xlabel[curve].trim().length() == 0 ||
                            (xlabel[curve].equalsIgnoreCase("default") && xlabeldefault.trim().length() == 0))
                            {
                            for (int j=0; j < nn[curve]; j++)
                                    x[curve][j] = (double)(j+1);
                            xlabel2[curve]="Sample Number";
                            }
                        else
                            {
                            if (xlabel[curve].equalsIgnoreCase("default"))
                                {
                                xcolumn[curve] = table.getColumnIndex (xlabeldefault);
                                if (xcolumn[curve] == ResultsTable.COLUMN_NOT_FOUND)
                                    {
                                    xcolumn[curve] = 0;
                                    plotY[curve] = false;
                                    usecurvebox[curve].setSelected(false);
//                                    if (xlabeldefault.trim().length() != 0)
//                                        IJ.showMessage ("Cannot access default X-axis table column "+xlabeldefault+" !");
//                                    xlabeldefault = columns[0];
//                                    return;
                                    }
                                xlabel2[curve]=xlabeldefault.trim();
                                }
                            else 
                                {
                                xcolumn[curve] = table.getColumnIndex (xlabel[curve]);
                                if (xcolumn[curve] == ResultsTable.COLUMN_NOT_FOUND)
                                        {
                                        xcolumn[curve] = 0;
                                        plotY[curve] = false;
                                        usecurvebox[curve].setSelected(false);
//                                        if (xlabel[curve].trim().length() != 0)
//                                            IJ.showMessage ("Cannot access X-axis table column "+xlabel[curve]+" for dataset "+(curve+1)+"!");
//                                        xlabel[curve] = "default";
//                                        xdatacolumn[curve].setSelectedItem(xlabel[curve]);
//                                        return;
                                        }
                                xlabel2[curve]=xlabel[curve].trim();
                                }
                            
                            if (plotY[curve])
                                {
                                int bucketSize = binSize[curve];

                                for (int j=0; j < nn[curve]; j++)
                                    {
                                    if (nnr[curve] > 0 && j == nn[curve] - 1) bucketSize = nnr[curve];
                                    x[curve][j] = 0;
                                    for (int k=excludedHeadSamples; k < (bucketSize+excludedHeadSamples); k++)
                                        x[curve][j] += table.getValueAsDouble (xcolumn[curve],j*binSize[curve]+k);
                                    x[curve][j] = x[curve][j]*xMultiplierFactor/(double)bucketSize;
                                    }
                                if (xlabel2[curve].startsWith("J.D.-2400000"))
                                    {
                                    for (int j=0; j < nn[curve]; j++)
                                        {
                                        x[curve][j] += 2400000;
                                        }
                                    }                                    
                                }
                            }
                        }
                for (int curve=0;curve<maxCurves;curve++)
                        {
                        if (plotY[curve])
                                {
                                if (ylabel[curve].trim().length() == 0)
                                        {
                                        for (int j=0; j < nn[curve]; j++)
                                                y[curve][j] = (double)(j+1);
                                        }
                                else	{
                                        ycolumn[curve] = table.getColumnIndex (ylabel[curve]);
                                        if (ycolumn[curve] == ResultsTable.COLUMN_NOT_FOUND)
                                                {
//                                                if (ylabel[curve].trim().length() != 0)
//                                                    IJ.log ("Cannot access Y"+curve+" table column "+ylabel[curve]+" !");
                                                ycolumn[curve] = 0;
                                                plotY[curve] = false;
                                                usecurvebox[curve].setSelected(false);
                                                }
                                        else if(operatorIndex[curve] == 5)  //calculate distance
                                                {
                                                xc1column[curve] = table.getColumnIndex (getPositionColumn(ylabel[curve],"X"));
                                                yc1column[curve] = table.getColumnIndex (getPositionColumn(ylabel[curve],"Y"));
                                                xyc1label[curve] = getSourceID(ylabel[curve]);
                                                if (xc1column[curve] == ResultsTable.COLUMN_NOT_FOUND )
                                                        {
                                                        if (getPositionColumn(ylabel[curve],"X").trim().length() != 0)
                                                            IJ.showMessage ("Cannot access "+getPositionColumn(ylabel[curve],"X")+" source position data column for distance calculation !");
                                                        operatorIndex[curve] = 0;
                                                        operatorselection[curve].setSelectedIndex(operatorIndex[curve]);
                                                        plotY[curve] = false;
                                                        usecurvebox[curve].setSelected(false);
                                                        }
                                                if (yc1column[curve] == ResultsTable.COLUMN_NOT_FOUND )
                                                        {
                                                        if (getPositionColumn(ylabel[curve],"Y").trim().length() != 0)
                                                            IJ.showMessage ("Cannot access "+getPositionColumn(ylabel[curve],"Y")+" source position data column for distance calculation !");
                                                        operatorIndex[curve] = 0;
                                                        operatorselection[curve].setSelectedIndex(operatorIndex[curve]);
                                                        plotY[curve] = false;
                                                        usecurvebox[curve].setSelected(false);  
                                                        }
                                                }
                                        if (showErrors[curve] == true || operatorIndex[curve] == 6)
                                                {
                                                errcolumn[curve] = ResultsTable.COLUMN_NOT_FOUND;
                                                if (operatorIndex[curve] == 6)   //custom error
                                                    {
                                                    errcolumn[curve] = table.getColumnIndex (oplabel[curve]);
                                                    showErrors[curve] = true;
                                                    }
                                                else if (ylabel[curve].startsWith("rel_flux_T") || ylabel[curve].startsWith("rel_flux_C"))
                                                    {
                                                    errcolumn[curve] = table.getColumnIndex ("rel_flux_err_"+ylabel[curve].substring(9));
                                                    }
                                                else if (ylabel[curve].startsWith("Source-Sky_"))
                                                    {
                                                    errcolumn[curve] = table.getColumnIndex ("Source_Error_"+ylabel[curve].substring(11));
                                                    }
                                                else if (ylabel[curve].startsWith("tot_C_cnts"))
                                                    {
                                                    errcolumn[curve] = table.getColumnIndex ("tot_C_err"+ylabel[curve].substring(10));
                                                    }                                                
                                               
                                                if (errcolumn[curve] == ResultsTable.COLUMN_NOT_FOUND)
                                                    {
                                                    showErrors[curve] = false;
                                                    }
                                                }

                                        
//                                        if (!errlabel[curve].equals(""))
//                                                {
//                                                errcolumn[curve] = table.getColumnIndex (errlabel[curve]);	
//                                                if (errcolumn[curve] == ResultsTable.COLUMN_NOT_FOUND)
//                                                        {
////                                                        if (errlabel[curve].trim().length() != 0)
////                                                            IJ.showMessage ("Cannot access Y"+curve+" error column "+errlabel[curve]+" !");
//                                                        showErrors[curve] = false;
////                                                        errorcolumnbox[curve].setSelected(false);
//                                                        }
//                                                else
//                                                        showErrors[curve] = true;
//                                                }
//                                        else
//                                                showErrors[curve] = false;                                        
                                        

                                        if (operatorIndex[curve] != 0)
                                                {
                                                opcolumn[curve] = table.getColumnIndex (oplabel[curve]);	
                                                if (opcolumn[curve] == ResultsTable.COLUMN_NOT_FOUND)
                                                        {
//                                                        if (oplabel[curve].trim().length() != 0)
//                                                            IJ.showMessage ("Cannot access Y"+curve+" operator data column "+oplabel[curve]+" !");
                                                        operatorIndex[curve] = 0;
                                                        operatorselection[curve].setSelectedIndex(operatorIndex[curve]);
                                                        }
                                                else if(operatorIndex[curve] == 5) //calculate distance
                                                        {
                                                        xc2column[curve] = table.getColumnIndex (getPositionColumn(oplabel[curve],"X"));
                                                        yc2column[curve] = table.getColumnIndex (getPositionColumn(oplabel[curve],"Y"));
                                                        xyc2label[curve] = getSourceID(oplabel[curve]);
                                                        if (xc2column[curve] == ResultsTable.COLUMN_NOT_FOUND )
                                                                {
                                                                if (getPositionColumn(oplabel[curve],"X").trim().length() != 0)
                                                                    IJ.showMessage ("Cannot access "+getPositionColumn(oplabel[curve],"X")+" source position data column for distance calculation !");
                                                                operatorIndex[curve] = 0;
                                                                operatorselection[curve].setSelectedIndex(operatorIndex[curve]);
                                                                }
                                                        if (yc2column[curve] == ResultsTable.COLUMN_NOT_FOUND )
                                                                {
                                                                if (getPositionColumn(oplabel[curve],"Y").trim().length() != 0)
                                                                    IJ.showMessage ("Cannot access "+getPositionColumn(oplabel[curve],"Y")+" source position data column for distance calculation !");
                                                                operatorIndex[curve] = 0;
                                                                operatorselection[curve].setSelectedIndex(operatorIndex[curve]);
                                                                }
                                                        }
                                                }
//                                        if (!operrlabel[curve].equals(""))
//                                                {
//                                                operrcolumn[curve] = table.getColumnIndex (operrlabel[curve]);	
//                                                if (operrcolumn[curve] == ResultsTable.COLUMN_NOT_FOUND)
//                                                        {
////                                                        if (operrlabel[curve].trim().length() != 0)
////                                                            IJ.showMessage ("Cannot access Y"+curve+" operator error column "+operrlabel[curve]+" !");
//                                                        showOpErrors[curve] = false;
////                                                        operrorcolumnbox[curve].setSelected(false);
//                                                        }
//                                                else
//                                                        showOpErrors[curve] = true;
//                                                }
//                                        else
//                                                showOpErrors[curve] = false;
                                        
                                        if (operatorIndex[curve] > 0 && operatorIndex[curve] < 5 && showErrors[curve] == true)
                                            {
                                            operrcolumn[curve] = ResultsTable.COLUMN_NOT_FOUND;
                                            if (oplabel[curve].startsWith("rel_flux_T") || oplabel[curve].startsWith("rel_flux_C"))
                                                {
                                                operrcolumn[curve] = table.getColumnIndex ("rel_flux_err_"+oplabel[curve].substring(9));
                                                }
                                            else if (oplabel[curve].startsWith("Source-Sky_"))
                                                {
                                                operrcolumn[curve] = table.getColumnIndex ("Source_Error_"+oplabel[curve].substring(11));
                                                }
                                            else if (oplabel[curve].startsWith("tot_C_cnts"))
                                                {
                                                operrcolumn[curve] = table.getColumnIndex ("tot_C_err"+oplabel[curve].substring(10));
                                                }
                                            if (operrcolumn[curve] == ResultsTable.COLUMN_NOT_FOUND)
                                                {
                                                showOpErrors[curve] = false;
                                                showErrors[curve] = false;
                                                }
                                            else
                                                {
                                                showOpErrors[curve] = true;
                                                }
                                            } 
                                        else
                                            showOpErrors[curve] = false;                                        

                                        int bucketSize = binSize[curve];
                                        for (int j=0; j < nn[curve]; j++)
                                                {
                                                y[curve][j] = 0;
                                                yerr[curve][j] = 0;
                                                yop[curve][j] = 0;
                                                xc1[curve][j] = 0;
                                                xc2[curve][j] = 0;
                                                yc1[curve][j] = 0;
                                                yc2[curve][j] = 0;
                                                // BIN DATA IF APPLICABLE
                                                if (nnr[curve] > 0 && j == nn[curve] - 1) bucketSize = nnr[curve];
                                                for (int k=excludedHeadSamples; k < (bucketSize+excludedHeadSamples); k++)
                                                        {
                                                        y[curve][j] += table.getValueAsDouble (ycolumn[curve],j*binSize[curve]+k);
                                                        if (showErrors[curve])
                                                                yerr[curve][j] += Math.pow(table.getValueAsDouble (errcolumn[curve],j*binSize[curve]+k),2);
                                                        if (operatorIndex[curve] != 0)
                                                                yop[curve][j] += table.getValueAsDouble (opcolumn[curve],j*binSize[curve]+k);
                                                        if (operatorIndex[curve] == 5) //calculate distance
                                                            {
                                                            xc1[curve][j] += table.getValueAsDouble (xc1column[curve],j*binSize[curve]+k);
                                                            xc2[curve][j] += table.getValueAsDouble (xc2column[curve],j*binSize[curve]+k);
                                                            yc1[curve][j] += table.getValueAsDouble (yc1column[curve],j*binSize[curve]+k);
                                                            yc2[curve][j] += table.getValueAsDouble (yc2column[curve],j*binSize[curve]+k);
                                                            }
                                                        if (showOpErrors[curve])
                                                                yoperr[curve][j] += Math.pow(table.getValueAsDouble (operrcolumn[curve],j*binSize[curve]+k),2);
                                                        }
                                                y[curve][j] = y[curve][j]*yMultiplierFactor/bucketSize;

                                                if (showErrors[curve])
                                                        yerr[curve][j] = yMultiplierFactor*Math.sqrt(yerr[curve][j])/bucketSize;
                                                else
                                                        yerr[curve][j] = 0.0;
                                                if (operatorIndex[curve] != 0)
                                                        {
                                                        yop[curve][j] = yop[curve][j]*yMultiplierFactor/bucketSize;
                                                        if (operatorIndex[curve] == 5)
                                                            {
                                                            xc1[curve][j] = xc1[curve][j]*yMultiplierFactor/bucketSize;
                                                            xc2[curve][j] = xc2[curve][j]*yMultiplierFactor/bucketSize;
                                                            yc1[curve][j] = yc1[curve][j]*yMultiplierFactor/bucketSize;
                                                            yc2[curve][j] = yc2[curve][j]*yMultiplierFactor/bucketSize;
                                                            }
                                                        }
                                                if (showOpErrors[curve])
                                                        yoperr[curve][j] = yMultiplierFactor*Math.sqrt(yoperr[curve][j])/bucketSize;
                                                else
                                                        yoperr[curve][j] = 0.0;

                                                //APPLY OPERATOR/OPERROR FUNCTIONS TO YDATA AND YERROR

                                                if (operatorIndex[curve] == 0)  //no operator
                                                        {

                                                        }
                                                else if (operatorIndex[curve] == 1)  //divide by
                                                        {
                                                        if (yop[curve][j]==0)
                                                            {
                                                            yerr[curve][j] = 1.0e+100;
                                                            y[curve][j] = 1.0e+100;
                                                            }
                                                        else
                                                            {
                                                            if (showErrors[curve] || showOpErrors[curve])
                                                                {
                                                                yerr[curve][j] = yMultiplierFactor*Math.sqrt(
                                                                ((yerr[curve][j]*yerr[curve][j])/(yop[curve][j]*yop[curve][j])) +
                                                                ((y[curve][j]*y[curve][j]*yoperr[curve][j]*yoperr[curve][j])/
                                                                 (yop[curve][j]*yop[curve][j]*yop[curve][j]*yop[curve][j])));
                                                                }
                                                            y[curve][j] = y[curve][j]*yMultiplierFactor/yop[curve][j];
                                                            }
                                                        }
                                                else if (operatorIndex[curve] == 2)  //multiply by
                                                        {
                                                        if (showErrors[curve] || showOpErrors[curve])
                                                                yerr[curve][j] = Math.sqrt(yop[curve][j]*yop[curve][j]*yerr[curve][j]*yerr[curve][j]+
                                                                    y[curve][j]*y[curve][j]*yoperr[curve][j]*yoperr[curve][j])/yMultiplierFactor;
                                                        y[curve][j] = y[curve][j]*yop[curve][j]/yMultiplierFactor;
                                                        }
                                                else if (operatorIndex[curve] == 3)  //subtract
                                                        {
                                                        if (showErrors[curve] || showOpErrors[curve])
                                                                yerr[curve][j] = Math.sqrt(yerr[curve][j]*yerr[curve][j]+yoperr[curve][j]*yoperr[curve][j]);
                                                        y[curve][j] = y[curve][j]-yop[curve][j];
                                                        }
                                                else if (operatorIndex[curve] == 4)  //add
                                                        {
                                                         if (showErrors[curve] || showOpErrors[curve])
                                                                yerr[curve][j] = Math.sqrt(yerr[curve][j]*yerr[curve][j]+yoperr[curve][j]*yoperr[curve][j]);
                                                        y[curve][j] = y[curve][j]+yop[curve][j];
                                                        }
                                                else if (operatorIndex[curve] == 5)  //distance from x1,y1 to x2,y2
                                                        {
                                                        y[curve][j] = (usePixelScale?pixelScale:1.0)*Math.sqrt(((xc1[curve][j]-xc2[curve][j])*(xc1[curve][j]-xc2[curve][j]))+
                                                                                ((yc1[curve][j]-yc2[curve][j])*(yc1[curve][j]-yc2[curve][j])));
                                                        }           
                                                }
                                        if (ylabel[curve].trim().startsWith("J.D.-2400000"))
                                            {
                                            for (int j=0; j < nn[curve]; j++)
                                                {
                                                y[curve][j] += 2400000;
                                                }
                                            }                                         
                                        //PERFORM MAGNITUDE CONVERSION IF APPLICABLE
                                        if (mmag[curve])
                                                {
                                                int cnt = 0;
                                                int numgood = 0;
                                                while (numgood<mmagrefs && cnt<nn[curve])  //Prepare mmag calulation basline reference level
                                                    {
                                                    if (!Double.isNaN(y[curve][cnt]))
                                                        {
                                                        baseline[curve] += y[curve][cnt];
                                                        numgood++;
                                                        }
                                                    cnt++;
                                                    }
                                                baseline[curve] /= (numgood == 0 ? 1 : numgood);
                                                baseline[curve] = (baseline[curve]==0 ? 1 : baseline[curve]);
                                                for (int j=0; j < nn[curve]; j++)
                                                        {
                                                        if (showErrors[curve] || showOpErrors[curve])
                                                            {
                                                            yerr[curve][j] = yMultiplierFactor*(2.5*Math.log10(1.0+yerr[curve][j]/y[curve][j]));
                                                            }
                                                        y[curve][j] = magSign*yMultiplierFactor*2.5*Math.log10(y[curve][j]/baseline[curve]);
                                                        }
                                                }

                                        }
                                }
                        }
                // PERFORM AUTOSCALING TO ONE OR MORE CURVES
                dx = 0.0;
                dy = 0.0;

                xPlotMin = xMin;
                xPlotMax = xMax;
                yPlotMin = yMin;
                yPlotMax = yMax;
                
                xautoscalemin = Double.POSITIVE_INFINITY;
                xautoscalemax = Double.NEGATIVE_INFINITY;
                yautoscalemin = Double.POSITIVE_INFINITY;
                yautoscalemax = Double.NEGATIVE_INFINITY;                

                firstCurve = -1;
                for (int curve = 0; curve<maxCurves; curve++)
                    {
                    if (plotY[curve])
                        {
                        xMinimum[curve] = minOf(x[curve],nn[curve]); //FIND MIN AND MAX X OF EACH SELECTED DATASET
                        xMaximum[curve] = maxOf(x[curve],nn[curve]);
                        yMn[curve] = showErrors[curve]?minOf(y[curve],yerr[curve],nn[curve]):minOf(y[curve],nn[curve]);      //FIND MIN AND MAX Y OF EACH SELECTED DATASET
                        yMx[curve] = showErrors[curve]?maxOf(y[curve],yerr[curve],nn[curve]):maxOf(y[curve],nn[curve]);
                        if (customScaleFactor[curve] >= 0)
                            {
                            yMinimum[curve] = force[curve] ? yMn[curve] : yMn[curve]*customScaleFactor[curve] + customShiftFactor[curve]; //FIND MIN AND MAX Y OF EACH SELECTED DATASET
                            yMaximum[curve] = force[curve] ? yMx[curve] : yMx[curve]*customScaleFactor[curve] + customShiftFactor[curve];
                            }
                        else
                            {
                            yMinimum[curve] = force[curve] ? yMn[curve] : yMx[curve]*customScaleFactor[curve] + customShiftFactor[curve]; //FIND MIN AND MAX Y OF EACH SELECTED DATASET
                            yMaximum[curve] = force[curve] ? yMx[curve] : yMn[curve]*customScaleFactor[curve] + customShiftFactor[curve];
                            }
                        if ((firstCurve == -1))
                                firstCurve = curve; //FIND THE FIRST CURVE TO DISPLAY - IT IS USED FOR THE Y-AXIS LABEL
                        if (ASInclude[curve] && !force[curve])
                            {
                            if (xMinimum[curve] < xautoscalemin) xautoscalemin = xMinimum[curve];
                            if (xMaximum[curve] > xautoscalemax) xautoscalemax = xMaximum[curve];
                            if (yMinimum[curve] < yautoscalemin) yautoscalemin = yMinimum[curve];
                            if (yMaximum[curve] > yautoscalemax) yautoscalemax = yMaximum[curve];
                            }
                        }
                    }

                if (Double.isInfinite(xautoscalemin)) xautoscalemin = Double.isInfinite(xautoscalemax) ? 0.0 : xautoscalemax - 1.0;
                if (Double.isInfinite(xautoscalemax)) xautoscalemax = xautoscalemin + 1.0;
                if (Double.isInfinite(yautoscalemin)) yautoscalemin = Double.isInfinite(yautoscalemax) ? 0.0 : yautoscalemax - 1.0;
                if (Double.isInfinite(yautoscalemax)) yautoscalemax = yautoscalemin + 1.0;                        

                if (firstCurve == -1)   //IF NO CURVES SELECTED FOR DISPLAY, USE THE FIRST CURVE
                        firstCurve = 0;

                 if (autoScaleX)
                    {
                    xPlotMin = xautoscalemin;
                    xPlotMax = (xautoscalemin == xautoscalemax) ? xautoscalemin+0.01 : xautoscalemax;
                    }
                else if (useFirstX)
                    {
                    firstXmin = xautoscalemin;                    
                    xPlotMin = xautoscalemin;
                    xPlotMax = (xWidth <= 0) ? xautoscalemin + 0.2 : xautoscalemin+xWidth;
                    }
                dx = (xPlotMax-xPlotMin)/99.0;

                if (autoScaleY)
                    {
                    yPlotMin = yautoscalemin;
                    yPlotMax = (yautoscalemin == yautoscalemax) ? yautoscalemin+0.01 : yautoscalemax;
                    }
                dy = (yPlotMax-yPlotMin)/99.0;


                // IF MJD, SHOW DECIMAL DATE DUE TO float API OF PlotWindow

                xOffset = 0.0;
                yOffset = 0.0;
                int jd=0;
                xlab ="";
                ylab = "";
                boolean xGoodLabel = false;
                boolean yGoodLabel = false;

                // MAKE X-AXIS LABEL

                for (int curve=0; curve<maxCurves; curve++)
                    {
                    if (plotY[curve])
                        {
                        if (xlabel2[curve].startsWith("J.D.-2400000") || xlabel2[curve].startsWith("JD") ||
                            xlabel2[curve].startsWith("HJD") || xlabel2[curve].startsWith("BJD"))
                                {
                                if (curve == firstCurve)
                                    {
                                    jd = (int)xPlotMin;
                                    xOffset = (double)jd;
                                    xPlotMin -= xOffset;
                                    xPlotMax -= xOffset;
                                    if (useXColumnName)
                                            {
                                            if (xlabel2[curve].trim().startsWith("J.D.-2400000"))
                                                xlab = "Geocentric Julian Date - "+jd;
                                            else if (xlabel2[curve].trim().startsWith("BJD_MOBS"))
                                                xlab = "Barycentric Julian Date - "+jd+" (mid-exposure)";
                                            else if (xlabel2[curve].trim().startsWith("BJD_SOBS"))
                                                xlab = "Barycentric Julian Date - "+jd+" (exposure start)";
                                            else if (xlabel2[curve].trim().startsWith("BJD"))
                                                xlab = "Barycentric Julian Date - "+jd;
                                            else if (xlabel2[curve].trim().startsWith("HJD_MOBS"))
                                                xlab = "Heliocentric Julian Date - "+jd+" (mid-exposure)";
                                            else if (xlabel2[curve].trim().startsWith("HJD_SOBS"))
                                                xlab = "Heliocentric Julian Date - "+jd+" (exposure start)";
                                            else if (xlabel2[curve].trim().startsWith("HJD"))
                                                xlab = "Heliocentric Julian Date - "+jd;
                                            else if (xlabel2[curve].trim().startsWith("JD_SOBS"))
                                                xlab = "Geocentric Julian Date - "+jd+" (exposure start)";
                                            else if (xlabel2[curve].trim().startsWith("JD_MOBS"))
                                                xlab = "Geocentric Julian Date - "+jd+" (mid-exposure)";
                                            else if (xlabel2[curve].trim().contains("JD"))
                                                xlab = "Julian Date - "+jd;
                                            else
                                                xlab = "Label Error";
                                            xGoodLabel = true;
                                            }
                                    else if (useXCustomName)
                                            {
                                            xlab = xLegend;
                                            xGoodLabel = true;
                                            }
                                    }
                                for (int j=0; j < nn[curve]; j++)
                                    x[curve][j] -= xOffset;
                                }
                        else if (useXColumnName && curve == firstCurve)
                                {
                                xlab = xlabel2[curve];
                                xGoodLabel = true;
                                }
                        else if (useXCustomName && curve == firstCurve)
                                {
                                xlab = xLegend;
                                xGoodLabel = true;
                                }
                        }
                     }

                if (xGoodLabel && (xExponent != 0) && showXScaleInfo)
                        xlab += " x 1E"+xExponent;

                // MAKE Y-AXIS LABEL

                for (int curve=0; curve<maxCurves; curve++)
                    {
                    if (plotY[curve])
                        {
                        if (ylabel[curve].trim().startsWith("J.D.-2400000") || ylabel[curve].trim().startsWith("JD") ||
                            ylabel[curve].trim().startsWith("HJD") || ylabel[curve].trim().startsWith("BJD"))
                                {
                                if (curve == firstCurve)
                                    {
                                    jd = (int)yPlotMin;
                                    yOffset = (double)jd;
                                    yPlotMin -= yOffset;
                                    yPlotMax -= yOffset;
                                    if (useYColumnName && operatorIndex[curve] != 5)
                                            {
                                            if (ylabel[curve].trim().startsWith("J.D.-2400000"))
                                                ylab = "Geocentric Julian Date - "+jd;
                                            else if (ylabel[curve].trim().startsWith("BJD_MOBS"))
                                                ylab = "Barycentric Julian Date - "+jd+" (mid-exposure)";
                                            else if (ylabel[curve].trim().startsWith("BJD_SOBS"))
                                                ylab = "Barycentric Julian Date - "+jd+" (exposure start)";
                                            else if (ylabel[curve].trim().startsWith("BJD"))
                                                ylab = "Barycentric Julian Date - "+jd;
                                            else if (ylabel[curve].trim().startsWith("HJD_MOBS"))
                                                ylab = "Heliocentric Julian Date - "+jd+" (mid-exposure)";
                                            else if (ylabel[curve].trim().startsWith("HJD_SOBS"))
                                                ylab = "Heliocentric Julian Date - "+jd+" (exposure start)";
                                            else if (ylabel[curve].trim().startsWith("HJD"))
                                                ylab = "Heliocentric Julian Date - "+jd;         
                                            else if (ylabel[curve].trim().startsWith("JD_MOBS"))
                                                ylab = "Geocentric Julian Date - "+jd+" (mid-exposure)";
                                            else if (ylabel[curve].trim().startsWith("JD_SOBS"))
                                                ylab = "Geocentric Julian Date - "+jd+" (exposure start)";
                                            else if (ylabel[curve].trim().contains("JD"))
                                                ylab = "Julian Date - "+jd;
                                            else
                                                ylab = "Label Error";
                                            yGoodLabel = true;
                                            }
                                    else if (useYCustomName)
                                            {
                                            ylab = yLegend;
                                            yGoodLabel = true;
                                            }
                                    else if(useYColumnName && operatorIndex[curve] == 5)
                                            {
                                            ylab = xyc1label[curve]+" - "+xyc2label[curve]+" centroid distance";
                                            ylab += usePixelScale ? " (arcsecs)" : " (pixels)";
                                            yGoodLabel = true;
                                            }
                                    }
                                for (int j=0; j < nn[curve]; j++)
                                    y[curve][j] -= yOffset;
                                }
                        else if (useYColumnName && curve == firstCurve)
                                {
                                if (operatorIndex[curve] != 5)
                                    {
                                    ylab=ylabel[firstCurve];
                                    if (operatorIndex[firstCurve] != 0)
                                            ylab += opSymbol[operatorIndex[firstCurve]]+oplabel[firstCurve];   
                                    }
                                else
                                    {
                                    ylab = xyc1label[firstCurve]+" - "+xyc2label[firstCurve]+" centroid distance";
                                    ylab += usePixelScale ? " (arcsecs)" : " (pixels)";
                                    }
                                yGoodLabel = true;
                                }
                        else if (useYCustomName && curve == firstCurve)
                                {
                                ylab = yLegend;
                                yGoodLabel = true;
                                }
                        }
                    }
                //FINISH Y-LABEL AFTER FINDING SCALING FACTORS BELOW

              
                //SCALE AND SHIFT SECONDARY CURVES TO FIT ON PLOT

                yRange = yPlotMax - yPlotMin;
                yMid = yPlotMin + yRange/2.;
                for (int curve=0; curve<maxCurves; curve++)
                    {
                    if (plotY[curve])
                        {
                        yMidpoint[curve] = (yMaximum[curve] + yMinimum[curve])/2.;
                        yWidth[curve] = yMaximum[curve] - yMinimum[curve];
                        if (yWidth[curve] == 0) yWidth[curve] = 1.0e-10;
                        if (force[curve])
                                {
                                subtotalScaleFactor[curve] = autoScaleFactor[curve]*yRange/yWidth[curve];
                                subtotalShiftFactor[curve] = yMid + yRange*autoShiftFactor[curve] - subtotalScaleFactor[curve]*yMidpoint[curve];
                                totalScaleFactor[curve] = 1.0/(subtotalScaleFactor[curve]*yMultiplierFactor);
                                totalShiftFactor[curve] = subtotalShiftFactor[curve];
                                }
                        else
                                {
                                subtotalScaleFactor[curve] = customScaleFactor[curve];
                                subtotalShiftFactor[curve] = customShiftFactor[curve];
                                totalScaleFactor[curve] = 1.0/(subtotalScaleFactor[curve]*yMultiplierFactor);
                                totalShiftFactor[curve] = subtotalShiftFactor[curve];
                                }
                        for (int j=0; j<nn[curve]; j++)
                            {
                            if (showErrors[curve] || showOpErrors[curve])
                                {
                                yerr[curve][j] = subtotalScaleFactor[curve]*yerr[curve][j];
                                }
                            y[curve][j] = subtotalScaleFactor[curve]*y[curve][j] + subtotalShiftFactor[curve];
                            }
                        }
                    }

                if (yGoodLabel)
                        {
                        if (showYScaleInfo || showYShiftInfo)
                                ylab += scaleShiftText(showYScaleInfo, showYShiftInfo, mmag[firstCurve], showYmmagInfo, totalScaleFactor[firstCurve], totalShiftFactor[firstCurve]);
                        if (mmag[firstCurve] && showYmmagInfo)
                                if (totalScaleFactor[firstCurve] == 0.001)
                                        ylab += " (mmag)";
                                else
                                        ylab += " (mag)";
                        if ((binSize[firstCurve]!=1) && showYBinInfo)
                                ylab += " (bin size = "+binSize[firstCurve]+")";
                        if (showYSymbolInfo)
                                ylab += " ("+colors[colorIndex[firstCurve]]+" "+markers[markerIndex[firstCurve]]+")";
                        }



 //----------------Set up plot options------------------------------------
                plotOptions = 0;
                if (xTics) plotOptions += ij.gui.Plot.X_TICKS;
                if (yTics) plotOptions += ij.gui.Plot.Y_TICKS;
                if (xGrid) plotOptions += ij.gui.Plot.X_GRID;
                if (yGrid) plotOptions += ij.gui.Plot.Y_GRID;
                if (xNumbers) plotOptions += ij.gui.Plot.X_NUMBERS;
                if (yNumbers) plotOptions += ij.gui.Plot.Y_NUMBERS;

                plot = new Plot ("Plot of "+tableName,xlab,ylab,nullX,nullY,plotOptions);
                plot.setSize(plotSizeX, plotSizeY);
                pltMinX = xPlotMin-5.*dx;
                pltMaxX = xPlotMax+5.*dx;
                pltMinY = yPlotMin-5.*dy;
                pltMaxY = yPlotMax+5.*dy;

                if (invertYAxis)
                    {
                    double yMaxTemp = pltMaxY;
                    pltMaxY = pltMinY;
                    pltMinY = yMaxTemp;
                    }

                if (plotImageCanvas != null) //zoom != 0.0 &&
                    {
                    plotMinX = totalPanOffsetX+newPanOffsetX+pltMinX+(pltMaxX-pltMinX)*(mouseX/(plotSizeX-(Plot.LEFT_MARGIN+Plot.RIGHT_MARGIN+1)))*zoom;
                    plotMaxX = totalPanOffsetX+newPanOffsetX+pltMaxX-(pltMaxX-pltMinX)*((plotSizeX-(Plot.LEFT_MARGIN+Plot.RIGHT_MARGIN+1)-mouseX)/(plotSizeX-(Plot.LEFT_MARGIN+Plot.RIGHT_MARGIN+1)))*zoom;
                    plotMinY = totalPanOffsetY+newPanOffsetY+pltMinY+(pltMaxY-pltMinY)*((plotSizeY-(Plot.TOP_MARGIN+Plot.BOTTOM_MARGIN)-mouseY)/(plotSizeY-(Plot.TOP_MARGIN+Plot.BOTTOM_MARGIN)))*zoom;
                    plotMaxY = totalPanOffsetY+newPanOffsetY+pltMaxY-(pltMaxY-pltMinY)*(mouseY/(plotSizeY-(Plot.TOP_MARGIN+Plot.BOTTOM_MARGIN)))*zoom;
                    }
                else
                    {
                    plotMinX = pltMinX;
                    plotMaxX = pltMaxX;
                    plotMinY = pltMinY;
                    plotMaxY = pltMaxY;
                    zoom = 0.0;
                    totalPanOffsetX=0.0;
                    totalPanOffsetY=0.0;
                    newPanOffsetX=0.0;
                    newPanOffsetY=0.0;
                    leftDragReleased=false;
                    }
                
                plot.setLimits (plotMinX, plotMaxX, plotMinY, plotMaxY);
                double legPosY = legendPosY;
                for (int curve=0; curve<maxCurves; curve++)
                    {
                    if (plotY[curve])
                        {
                        plot.setColor (color[curve]);
                        if (marker[curve] == ij.gui.Plot.DOT)
                            plot.setLineWidth(4);
                        else
                            plot.setLineWidth(1);

                        plot.addPoints (Arrays.copyOf(x[curve],nn[curve]),Arrays.copyOf(y[curve],nn[curve]),marker[curve]);

                        plot.setLineWidth(1);

                        if (showErrors[curve] || showOpErrors[curve])     //code to replace plot.addErrorBars
                            {               //since plot.addErrorBars only plots with lines enabled
                            for (int j=0; j < nn[curve]; j++)
                                {
                                plot.drawLine(x[curve][j], y[curve][j]-yerr[curve][j], x[curve][j], y[curve][j]+yerr[curve][j]);
                                }
                            }

                        if (lines[curve] && !(marker[curve] == ij.gui.Plot.LINE))
                            {
                             for (int j=0; j < nn[curve]-1; j++)
                                {
                                plot.drawLine(x[curve][j],y[curve][j],x[curve][j+1],y[curve][j+1]);
                                }
                            }

                        plot.setLineWidth(1);
                        if (legendLeft)
                            plot.setJustification(ij.process.ImageProcessor.LEFT_JUSTIFY);
                        else if (legendRight)
                            plot.setJustification(ij.process.ImageProcessor.RIGHT_JUSTIFY);
                        else
                            plot.setJustification(ij.process.ImageProcessor.CENTER_JUSTIFY);

                        // MAKE FULL LEGEND STRING FOR CURVE
                        String llab = "";

                        if (useColumnName[curve])
                            {
                            if (operatorIndex[curve] == 5)
                                {
                                llab = xyc1label[curve]+" - "+xyc2label[curve]+" centroid distance";
                                llab += usePixelScale ? " (arcsecs)" : " (pixels)";
                                }
                            else
                                {
                                llab=ylabel[curve];
                                if (operatorIndex[curve] != 0)
                                    llab += opSymbol[operatorIndex[curve]]+oplabel[curve];
                                }
                            }
                        else if (useLegend[curve])
                            llab = legend[curve];
                        if (useColumnName[curve] || useLegend[curve])
                            {
                            if (mmag[curve] && showLmmagInfo)
                                if (mmag[firstCurve] && totalScaleFactor[firstCurve] == 0.001)
                                    llab += " (mmag)";
                                else if (totalScaleFactor[curve] == 0.001)
                                    llab += " (mmag)";
                                else
                                    llab += " (mag)";
                            if (showLScaleInfo || showLShiftInfo)
                                llab += scaleShiftText(showLScaleInfo, showLShiftInfo, mmag[curve], showLmmagInfo, totalScaleFactor[curve], totalShiftFactor[curve]);
                            if (useTwoLineLegend)
                                {
                                plot.addLabel(legendPosX, legPosY, llab);
                                legPosY += 18./plotSizeY;
                                llab = " ";
                                }

                            if ((binSize[curve] != 1) && showLBinInfo)
                                llab += " (bin size = "+binSize[curve]+")";
                            if (showLSymbolInfo)
                                llab += " ("+markers[markerIndex[curve]]+")";
                            if (!llab.equals(" "))
                                {
                                plot.addLabel(legendPosX, legPosY, llab);
                                legPosY += 18./plotSizeY;
                                }
                            }
                        }
                    }

                if(useBoldedDatum && boldedDatum != -1)
                    {
                    for (int curve=0; curve<maxCurves; curve++)
                        {
                        if (plotY[curve])
                            {
                            plot.setColor (color[curve]);
                            if (marker[curve] == ij.gui.Plot.DOT)
                                plot.setLineWidth(8);
                            else
                                plot.setLineWidth(2);
                            double[] xx = {x[curve][boldedDatum]};
                            double[] yy = {y[curve][boldedDatum]};
                            plot.addPoints(xx,yy,marker[curve]);
                            }
                        }
                    }

                plot.setLineWidth(1);
                plot.setJustification(plot.CENTER);

                if (showVMarker1)
                    {
                    plot.setColor (Color.red);
                    double[] pre = {0.0,0.0,0.0,0.0,0.0};
                    double preVmark1Ref= invertYAxis?yPlotMin:yPlotMax;
                    int nI = 0;
                    for (int i=0; i<nn[firstCurve]; i++)
                        {
                        if (!Double.isNaN(x[firstCurve][i]) && !Double.isNaN(y[firstCurve][i]) && x[firstCurve][i] < vMarker1Value)
                            {
                            pre[4]=pre[3]; pre[3]=pre[2]; pre[2]=pre[1]; pre[1]=pre[0]; pre[0]=y[firstCurve][i];
                            nI++;
                            }
                        }
                    if (nI > 0)
                        {
                        nI = nI>5 ? 5 : nI;
                        preVmark1Ref = invertYAxis?minOf(pre,nI):maxOf(pre,nI);
                        preVmark1Ref = invertYAxis?Math.max(preVmark1Ref, plotMaxY):Math.min(preVmark1Ref, plotMaxY);
                        }
                    dashLength = 5*(plotMaxY - plotMinY)/plotSizeY;
                    numDashes = -10 + (preVmark1Ref - plotMinY)/dashLength;
                    for (int dashCount = 0; dashCount < numDashes; dashCount +=2)    //plot vMarker1
                        {
                        plot.drawLine(vMarker1Value, preVmark1Ref-dashLength*dashCount, vMarker1Value, preVmark1Ref-dashLength*(dashCount+1));
                        }
                    plot.setJustification(plot.CENTER);
                    plot.addLabel((vMarker1Value-plotMinX)/(plotMaxX-plotMinX), 1 - 25.0/plotSizeY, vMarker1TopText);
                    plot.addLabel((vMarker1Value-plotMinX)/(plotMaxX-plotMinX), 1 - 7.0/plotSizeY, vMarker1BotText);
                    }


                if (showVMarker2)
                    {
                    plot.setColor (Color.red);
                    double[] post = {0.0,0.0,0.0,0.0,0.0};
                    double postVMarker2Ref= invertYAxis?yPlotMin:yPlotMax;
                    int nE = 0;
                    for (int i=0; i<nn[firstCurve]; i++)
                        {
                        if (!Double.isNaN(x[firstCurve][i]) && !Double.isNaN(y[firstCurve][i]) && x[firstCurve][i] > vMarker2Value && nE < 5)
                            {
                            post[nE] = y[firstCurve][i];
                            nE++;
                            }
                        }
                    if (nE > 0)
                        {
                        postVMarker2Ref = invertYAxis?minOf(post,nE):maxOf(post,nE);
                        postVMarker2Ref =invertYAxis?Math.max(postVMarker2Ref, plotMaxY):Math.min(postVMarker2Ref, plotMaxY);
                        }
                    dashLength = 5*(plotMaxY - plotMinY)/plotSizeY;
                    numDashes = -10 + (postVMarker2Ref - plotMinY)/dashLength;
                    for (int dashCount = 0; dashCount < numDashes; dashCount +=2)     //plot vMarker2
                        {
                        plot.drawLine(vMarker2Value, postVMarker2Ref-dashLength*dashCount, vMarker2Value, postVMarker2Ref-dashLength*(dashCount+1));
                        }
                    plot.setJustification(plot.CENTER);
                    plot.addLabel((vMarker2Value-plotMinX)/(plotMaxX-plotMinX), 1 - 25.0/plotSizeY, vMarker2TopText);
                    plot.addLabel((vMarker2Value-plotMinX)/(plotMaxX-plotMinX), 1 - 7.0/plotSizeY, vMarker2BotText);
                    }


                plot.setColor (java.awt.Color.black);
                plot.setJustification(plot.CENTER);
                if (useTitle)
                        {
                        plot.changeFont(new java.awt.Font("Dialog",java.awt.Font.PLAIN,18));
                        plot.addTitle(titlePosX, titlePosY/plotSizeY>1?1:titlePosY/plotSizeY,title);
                        }
                if (useSubtitle)
                        {
                        plot.changeFont(new java.awt.Font("Dialog",java.awt.Font.PLAIN,14));
                        plot.addTitle(subtitlePosX, subtitlePosY/plotSizeY>1?1:subtitlePosY/plotSizeY, subtitle);
                        }

                plotImage = WindowManager.getImage("Plot of "+tableName);
                
                if (plotImage == null)
                    {
                    plotFrameLocationX=(int)Prefs.get("plot2.plotFrameLocationX", plotFrameLocationX);
                    plotFrameLocationY=(int)Prefs.get("plot2.plotFrameLocationY", plotFrameLocationY);
                    if (!Prefs.isLocationOnScreen(new Point(plotFrameLocationX,plotFrameLocationY)))
                        {
                        plotFrameLocationX = 10;
                        plotFrameLocationY = 10;
                        Prefs.set("plot2.plotFrameLocationX", plotFrameLocationX);
                        Prefs.set("plot2.plotFrameLocationY", plotFrameLocationY);
                        }
                    plotWindow = plot.show();
                    plotWindow.setIconImage(plotIcon.getImage());
                    plotImage = plotWindow.getImagePlus();//WindowManager.getImage("Plot of "+tableName);
                    
//                    plotWindow = plotImage.getWindow();
//                    plotWindow = new PlotWindow(plot);
//                    plotImage = plotWindow.getImagePlus();
//                    plotFrame = WindowManager.getFrame("Plot of "+tableName);
//                    if ((plotImage != null) && rememberWindowLocations)
//                        {
//                        plotFrameLocationX=(int)Prefs.get("plot2.plotFrameLocationX", plotFrameLocationX);
//                        plotFrameLocationY=(int)Prefs.get("plot2.plotFrameLocationY", plotFrameLocationY);
//                        if (!Prefs.isLocationOnScreen(new Point(plotFrameLocationX,plotFrameLocationY)))
//                            {
//                            plotFrameLocationX = 10;
//                            plotFrameLocationY = 10;
//                            }
//                        plotWindow.setLocation(plotFrameLocationX, plotFrameLocationY);
//                        }

//                    plotImage = WindowManager.getImage("Plot of "+tableName);
//                    plotFrame = WindowManager.getFrame("Plot of "+tableName);
//                    ImageProcessor ip = plot.getProcessor();
//                    plotImage.setProcessor("Plot of "+tableName,ip);
                    
                    plotImageCanvas = plotImage.getCanvas();
                    plotOverlayCanvas = new OverlayCanvas(plotImage);
                    list.clear();

                    MouseMotionListener[] mml = plotImageCanvas.getMouseMotionListeners();
                    if (mml.length>0)
                            for (int i=0; i<mml.length; i++)
                                    plotImageCanvas.removeMouseMotionListener(mml[i]);

                    MouseWheelListener[] mwl = plotImageCanvas.getMouseWheelListeners();
                    if (mwl.length>0)
                            for (int i=0; i<mwl.length; i++)
                                    plotImageCanvas.removeMouseWheelListener(mwl[i]);

                    MouseWheelListener[] mwl3 = plotWindow.getMouseWheelListeners();
                    if (mwl3.length>0)
                            for (int i=0; i<mwl3.length; i++)
                                    plotWindow.removeMouseWheelListener(mwl3[i]);

                    MouseListener[] ml = plotImageCanvas.getMouseListeners();
                    if (ml.length>0)
                            for (int i=0; i<ml.length; i++)
                                    plotImageCanvas.removeMouseListener(ml[i]);

                    plotImageCanvas.addMouseWheelListener(plotMouseWheelListener);
                    plotImageCanvas.addMouseMotionListener(plotMouseMotionListener);
                    plotImageCanvas.addMouseListener(plotMouseListener);
                    plotImageCanvas.setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
                    zoom = 0.0;
                    }
                else
                    {
                    plotWindow = plotImage.getWindow();
//                    plotFrame = (Frame)plotWindow;
                    ImageProcessor ip = plot.getProcessor();
                    plotImage.setProcessor("Plot of "+tableName,ip);
                    plotImageCanvas = plotImage.getCanvas();
                    plotImageCanvas.setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
                    }

                plotbottompanel = (Panel) plotWindow.getComponent(1);
                plotbottompanel.getComponentCount();
                plotbottompanel.setSize(600,30);
                plotcoordlabel = (Label) plotbottompanel.getComponent(plotbottompanel.getComponentCount()-1);
                plotcoordlabel.setSize(400,20);
                plotcoordlabel.setBackground(Color.white);

                //Replot to clean up blank areas

//                plotImage = WindowManager.getImage("Plot of "+tableName);
//                ImageProcessor ip = plot.getProcessor();
//                plotImage.setProcessor("Plot of "+tableName,ip);
                
                excludedHeadSamples = holdExcludedHeadSamples;
                excludedTailSamples = holdExcludedTailSamples;
                for (int i=0; i<maxCurves; i++)
                    {
                    showErrors[i] = errorcolumnbox[i].isSelected();
                    }
                    
                updatePlotRunning = false;
                }

    static String getPositionColumn(String label, String XorY)
        {
        String nums = "";
        for (int i = 0; i<label.length(); i++)
            {
            if (label.charAt(i) >= '0' && label.charAt(i) <= '9')
                {
                nums = nums.concat(label.substring(i, i+1));
                }
            }
        if (nums.trim().equalsIgnoreCase("") || nums.trim().equalsIgnoreCase("1"))
            return XorY+"_T1";
        return XorY+"_C"+(Integer.parseInt(nums));
        }

    static String getSourceID(String label)
        {
        String nums = "";
        for (int i = 0; i<label.length(); i++)
            {
            if (label.charAt(i) >= '0' && label.charAt(i) <= '9')
                {
                nums = nums.concat(label.substring(i, i+1));
                }
            }
        if (nums.trim().equalsIgnoreCase("") || nums.trim().equalsIgnoreCase("1")) return "T1";
        return "C"+(Integer.parseInt(nums));
        }

	static void addElement(Vector<Roi> list, Shape shape, Color color, int lineWidth) {
		roi = new ShapeRoi(shape);
		roi.setInstanceColor(color);
		roi.setLineWidth(lineWidth);
		list.addElement(roi);
                }
    
//    static MouseListener panelMouseListener = new MouseListener() {
//        public void mouseClicked(MouseEvent e) 
//            {
//            panelShiftDown = e.isShiftDown();
//            panelAltDown = e.isAltDown();
//            panelControlDown = e.isControlDown();
//            }
//
//        public void mousePressed(MouseEvent e) 
//            {
//            panelShiftDown = e.isShiftDown();
//            panelAltDown = e.isAltDown();
//            panelControlDown = e.isControlDown();
//            }
//
//        public void mouseReleased(MouseEvent e) 
//            {
//            panelShiftDown = e.isShiftDown();
//            panelAltDown = e.isAltDown();
//            panelControlDown = e.isControlDown();            
//            }
//
//        public void mouseEntered(MouseEvent e) 
//            {
//            panelShiftDown = e.isShiftDown();
//            panelAltDown = e.isAltDown();
//            panelControlDown = e.isControlDown();
//            }
//
//        public void mouseExited(MouseEvent e) 
//            {
//            panelShiftDown = e.isShiftDown();
//            panelAltDown = e.isAltDown();
//            panelControlDown = e.isControlDown();          
//            }
//    };    
    
    static MouseMotionListener panelMouseMotionListener = new MouseMotionListener() {
        public void mouseDragged(MouseEvent e) 
            {
            panelShiftDown = e.isShiftDown();
            panelAltDown = e.isAltDown();
            panelControlDown = e.isControlDown();            
            }
 
        public void mouseMoved(MouseEvent e)
            {
            panelShiftDown = e.isShiftDown();
            panelAltDown = e.isAltDown();
            panelControlDown = e.isControlDown();            
            }
    };
    
    

    static MouseListener plotMouseListener = new MouseListener() {
        public void mouseClicked(MouseEvent e) {

                // mouse clicked code is in mouseReleased() to allow drag/click thresholding
                }

        public void mousePressed(MouseEvent e) {
                startDragX = plotImageCanvas.offScreenX(e.getX());
                startDragY = plotImageCanvas.offScreenY(e.getY());
                startDragScreenX = e.getX();
                startDragScreenY = e.getY();
                startDragSubImageX =plotImageCanvas.getSrcRect().x;
                startDragSubImageY =plotImageCanvas.getSrcRect().y;
                button2Drag = false;
                }

        public void mouseReleased(MouseEvent e) {
                int screenX = e.getX();
                int screenY = e.getY();
                int imageX = plotImageCanvas.offScreenX(screenX);
                int imageY = plotImageCanvas.offScreenY(screenY);

                double xval = plotMinX + (double)(startDragX - Plot.LEFT_MARGIN) * (plotMaxX - plotMinX)/(double)(plotSizeX - (Plot.LEFT_MARGIN+Plot.RIGHT_MARGIN));
                double yval = plotMaxY - (double)(startDragY - Plot.TOP_MARGIN) * (plotMaxY - plotMinY)/(double)(plotSizeY - (Plot.TOP_MARGIN+Plot.BOTTOM_MARGIN));
                double dxval = (double)(imageX - startDragX) * (plotMaxX - plotMinX)/(double)(plotSizeX - (Plot.LEFT_MARGIN+Plot.RIGHT_MARGIN));
                double dyval = (double)(imageY - startDragY) * (-1.0) * (plotMaxY - plotMinY)/(double)(plotSizeY - (Plot.TOP_MARGIN+Plot.BOTTOM_MARGIN));


                if (((e.getModifiers() & MouseEvent.BUTTON2_MASK) != 0) && button2Drag)
                        {                                                          // measure distance and report in Results Table
                        button2Drag = false;
                        IJ.log("x="+fourPlaces.format(xval)+", y="+fourPlaces.format(yval)+
                                 ", dx="+fourPlaces.format(dxval)+", dy="+fourPlaces.format(dyval));
                        }
                if (Math.sqrt((screenX-startDragScreenX)*(screenX-startDragScreenX) +      //check mouse click/drag threshold
                        (screenY-startDragScreenY)*(screenY-startDragScreenY)) < 4.0)
                        {
                        if(e.getButton() == MouseEvent.BUTTON1 && e.isShiftDown())       //shift + left mouse click release (remove selected table row)
                                {                                                        //remove selected table row
                                selectedRowStart=tpanel.getSelectionStart();
                                selectedRowEnd=tpanel.getSelectionEnd();
                                for (int i=selectedRowEnd; i>=selectedRowStart;i--)
                                    {
                                    deletedRowList.add(tpanel.getLine(i));
                                    }
                                tpanel.clearSelection();
//                                table=MeasurementTable.getTable(tableName);
//                                if (Data_Processor.running) Data_Processor.setTable(table);
                                table.show();
                                tpanel=MeasurementTable.getTextPanel(MeasurementTable.longerName(tableName));
                                int lineCount = tpanel.getLineCount();
                                if (selectedRowEnd >= lineCount-excludedTailSamples)
                                    {
                                    selectedRowStart=binSize[firstCurve]*((lineCount-excludedHeadSamples-excludedTailSamples)/binSize[firstCurve])-binSize[firstCurve];
                                    selectedRowEnd=selectedRowStart+binSize[firstCurve]-1;
                                    }
                                tpanel.setSelection(selectedRowStart, selectedRowEnd);
                                boldedDatum=(selectedRowStart-excludedHeadSamples)/binSize[firstCurve];
                                plotcoordlabel.setText("DATA: x="+fourPlaces.format(x[firstCurve][boldedDatum])+", y="+fourPlaces.format(y[firstCurve][boldedDatum]));
                                IJ.showStatus("data values: x="+fourPlaces.format(x[firstCurve][boldedDatum])+", y="+fourPlaces.format(y[firstCurve][boldedDatum]));
                                updatePlot();
                                }
                        else if(e.getButton() == MouseEvent.BUTTON1) // shift left mouse click release
                                {
                                zoomControl(e.getX(), e.getY(), -5);
                                }

                        else if (e.getButton() == MouseEvent.BUTTON2)                     //middle mouse click release
                                {
                                IJ.log("x="+fourPlaces.format(xval)+", y="+fourPlaces.format(yval));
                                }
                        else if (e.getButton() == MouseEvent.BUTTON3 && e.isShiftDown() && !deletedRowList.isEmpty()) //shift + right mouse click release
                                {                                                                                     //undo delete selected table row
                                int lastRow = tpanel.getLineCount()-1;
                                if (selectedRowEnd<lastRow-binSize[firstCurve])
                                    {
                                    String[] lines = new String[lastRow-selectedRowStart+1];
                                    for (int i=0; i<lines.length; i++)
                                        {
                                        lines[i] = tpanel.getLine(selectedRowStart+i);
                                        }
                                    tpanel.setSelection(selectedRowStart, lastRow);
                                    tpanel.clearSelection();
                                    for (int i=0; i<binSize[firstCurve];i++)
                                        {
                                        if (!deletedRowList.isEmpty()) tpanel.appendWithoutUpdate(deletedRowList.remove(deletedRowList.size()-1));
                                        }
                                    for (int i=0; i<lines.length; i++)
                                        {
                                        tpanel.appendWithoutUpdate(lines[i]);
                                        }
                                    }
                                else if (selectedRowEnd<lastRow)
                                    {
                                    String[] lines = new String[lastRow-selectedRowEnd];
                                    for (int i=0; i<lines.length; i++)
                                        {
                                        lines[i] = tpanel.getLine(selectedRowEnd+1+i);
                                        }
//                                    IJ.log("made it");
                                    tpanel.setSelection(selectedRowEnd+1, lastRow);
                                    tpanel.clearSelection();
                                    for (int i=0; i<binSize[firstCurve];i++)
                                        {
                                        if (!deletedRowList.isEmpty()) tpanel.appendWithoutUpdate(deletedRowList.remove(deletedRowList.size()-1));
                                        }
                                    for (int i=0; i<lines.length; i++)
                                        {
                                        tpanel.appendWithoutUpdate(lines[i]);
                                        }
                                    selectedRowStart+=binSize[firstCurve];
                                    
                                    }
                                else
                                    {
                                    for (int i=0; i<binSize[firstCurve];i++)
                                        {
                                        if (!deletedRowList.isEmpty()) tpanel.appendWithoutUpdate(deletedRowList.remove(deletedRowList.size()-1));
                                        };
                                    selectedRowStart+=binSize[firstCurve];
                                    }
//                                table=MeasurementTable.getTable(tableName);
//                                if (Data_Processor.running) Data_Processor.setTable(table);
                                table.show();
                                tpanel=MeasurementTable.getTextPanel(MeasurementTable.longerName(tableName));
                                selectedRowEnd=selectedRowStart+binSize[firstCurve]-1;
                                tpanel.setSelection(selectedRowStart, selectedRowEnd);
                                boldedDatum=(selectedRowStart-excludedHeadSamples)/binSize[firstCurve];
                                plotcoordlabel.setText("DATA: x="+fourPlaces.format(x[firstCurve][boldedDatum])+", y="+fourPlaces.format(y[firstCurve][boldedDatum]));
                                IJ.showStatus("data values: x="+fourPlaces.format(x[firstCurve][boldedDatum])+", y="+fourPlaces.format(y[firstCurve][boldedDatum]));
                                updatePlot();
                                }
                        else if (e.getButton() == MouseEvent.BUTTON3)                     //right mouse click release
                                {
                                totalPanOffsetX=0.0;
                                totalPanOffsetY=0.0;
                                newPanOffsetX=0.0;
                                newPanOffsetY=0.0;
                                zoomControl(screenX, screenY, 10000);
                                }
                        }
                else                                                                      //complete drag operations
                    {
                    if (e.getButton() == MouseEvent.BUTTON1)                             //left mouse drag release
                        {
                        totalPanOffsetX += newPanOffsetX;
                        totalPanOffsetY += newPanOffsetY;
                        newPanOffsetX = 0.0;
                        newPanOffsetY = 0.0;
                        leftDragReleased = true;
                        }
                    }
                list.clear();
                path.reset();
                plotImageCanvas.repaint();
                }

        public void mouseEntered(MouseEvent e) {}

        public void mouseExited(MouseEvent e) {}
    };


    static MouseMotionListener plotMouseMotionListener = new MouseMotionListener() {
        public void mouseDragged(MouseEvent e) {
                int screenX = e.getX();
                int screenY = e.getY();
                int imageX = plotImageCanvas.offScreenX(e.getX());
                int imageY = plotImageCanvas.offScreenY(e.getY());
                if (Math.sqrt((screenX-startDragScreenX)*(screenX-startDragScreenX) +  //check mouse click/drag threshold
                        (screenY-startDragScreenY)*(screenY-startDragScreenY)) >= 4.0)
                        {
                        if ((e.getModifiers() & MouseEvent.BUTTON3_MASK) != 0)    // dragging with right mouse button
                                {                                                 // measure distance, show on plot and status bar and save to results window
                                list.clear();
                                path.reset();
                                path.moveTo(startDragX, startDragY);
                                path.lineTo(imageX, imageY);

                                addElement(list, path, Color.red, 2);
                                 plotImageCanvas.setDisplayList(list);
                                double x = plotMinX + (double)(imageX - (Plot.LEFT_MARGIN)) * (plotMaxX - plotMinX)/(double)(plotSizeX - (Plot.LEFT_MARGIN+Plot.RIGHT_MARGIN));
                                double y = plotMaxY - (double)(imageY - (Plot.TOP_MARGIN)) * (plotMaxY - plotMinY)/(double)(plotSizeY - (Plot.TOP_MARGIN+Plot.BOTTOM_MARGIN));
                                double dx = (double)(imageX - startDragX) * (plotMaxX - plotMinX)/(double)(plotSizeX - (Plot.LEFT_MARGIN+Plot.RIGHT_MARGIN));
                                double dy = (double)(imageY - startDragY) * (-1.0) * (plotMaxY - plotMinY)/(double)(plotSizeY - (Plot.TOP_MARGIN+Plot.BOTTOM_MARGIN));
                                plotbottompanel = (Panel) plotWindow.getComponent(1);
                                plotbottompanel.setSize(600,30);
                                plotcoordlabel = (Label) plotbottompanel.getComponent(4);
                                plotcoordlabel.setSize(400,20);
                                plotcoordlabel.setText("x="+fourPlaces.format(x)+", y="+fourPlaces.format(y)+
                                                       ", dx="+fourPlaces.format(dx)+", dy="+fourPlaces.format(dy));
                                IJ.showStatus("plot coordinates: x="+fourPlaces.format(x)+", y="+fourPlaces.format(y)+
                                              ", dx="+fourPlaces.format(dx)+", dy="+fourPlaces.format(dy));
                                }
                        else if ((e.getModifiers() & MouseEvent.BUTTON2_MASK) != 0)        // dragging with middle mouse button
                                {                                                          // measure distance, show on plot and status bar
                                button2Drag = true;                                        // and save to results window when mouse released
                                list.clear();
                                path.reset();
                                path.moveTo(startDragX, startDragY);
                                path.lineTo(imageX, imageY);

                                addElement(list, path, Color.red, 2);
                                 plotImageCanvas.setDisplayList(list);
                                double x = plotMinX + (double)(imageX - (Plot.LEFT_MARGIN)) * (plotMaxX - plotMinX)/(double)(plotSizeX - (Plot.LEFT_MARGIN+Plot.RIGHT_MARGIN));
                                double y = plotMaxY - (double)(imageY - (Plot.TOP_MARGIN)) * (plotMaxY - plotMinY)/(double)(plotSizeY - (Plot.TOP_MARGIN+Plot.BOTTOM_MARGIN));
                                double dx = (double)(imageX - startDragX) * (plotMaxX - plotMinX)/(double)(plotSizeX - (Plot.LEFT_MARGIN+Plot.RIGHT_MARGIN));
                                double dy = (double)(imageY - startDragY) * (-1.0) * (plotMaxY - plotMinY)/(double)(plotSizeY - (Plot.TOP_MARGIN+Plot.BOTTOM_MARGIN));
                                plotcoordlabel.setText("x="+fourPlaces.format(x)+", y="+fourPlaces.format(y)+
                                                       ", dx="+fourPlaces.format(dx)+", dy="+fourPlaces.format(dy));
                                IJ.showStatus("plot coordinates: x="+fourPlaces.format(x)+", y="+fourPlaces.format(y)+
                                              ", dx="+fourPlaces.format(dx)+", dy="+fourPlaces.format(dy));
                                }
                        else if ((e.getModifiers() & MouseEvent.BUTTON1_MASK) != 0)            // dragging with left mouse button (pan image)
                                {
                                Rectangle imageRect = plotImageCanvas.getSrcRect();
                                if (true)//(zoom != 0.0)
                                    {
                                    newPanOffsetX=-(plotMaxX - plotMinX)*(screenX-startDragScreenX)/(plotSizeX-(Plot.LEFT_MARGIN+Plot.RIGHT_MARGIN));
                                    newPanOffsetY=(plotMaxY - plotMinY)*(screenY-startDragScreenY)/(plotSizeY-(Plot.TOP_MARGIN+Plot.BOTTOM_MARGIN));
                                    updatePlot();
                                    }
                                else
                                    {
                                    totalPanOffsetX=0.0;
                                    totalPanOffsetY=0.0;
                                    newPanOffsetX=0.0;
                                    newPanOffsetY=0.0;
                                    leftDragReleased=false;
                                    }
                                if (e.isShiftDown())
                                    {
                                    plotcoordlabel.setText("DATA: x="+fourPlaces.format(x[firstCurve][boldedDatum])+", y="+fourPlaces.format(y[firstCurve][boldedDatum]));
                                    IJ.showStatus("data values: x="+fourPlaces.format(x[firstCurve][boldedDatum])+", y="+fourPlaces.format(y[firstCurve][boldedDatum]));
                                    }
                                else
                                    {
                                    double x = plotMinX + (double)(imageX - (Plot.LEFT_MARGIN)) * (plotMaxX - plotMinX)/(double)(plotSizeX - (Plot.LEFT_MARGIN+Plot.RIGHT_MARGIN));
                                    double y = plotMaxY - (double)(imageY - (Plot.TOP_MARGIN)) * (plotMaxY - plotMinY)/(double)(plotSizeY - (Plot.TOP_MARGIN+Plot.BOTTOM_MARGIN));  //was 58
                                    plotcoordlabel.setText("x="+fourPlaces.format(x)+", y="+fourPlaces.format(y));
                                    IJ.showStatus("plot coordinates: x="+fourPlaces.format(x)+", y="+fourPlaces.format(y));
                                    }
                                }
                        }
                }
 
        public void mouseMoved(MouseEvent e) {
                
                int imageX = plotImageCanvas.offScreenX(e.getX());
                int imageY = plotImageCanvas.offScreenY(e.getY());
                double xval = plotMinX + (double)(imageX - (Plot.LEFT_MARGIN)) * (plotMaxX - plotMinX)/(double)(plotSizeX - (Plot.LEFT_MARGIN+Plot.RIGHT_MARGIN));
                double yval = plotMaxY - (double)(imageY - (Plot.TOP_MARGIN)) * (plotMaxY - plotMinY)/(double)(plotSizeY - (Plot.TOP_MARGIN+Plot.BOTTOM_MARGIN));


                if (e.isShiftDown())                                     //select table line nearest x-val of first selected curve
                    {
                    nearestLine=0;
                    nearestX = Math.abs(xval-x[firstCurve][0]);
                    if (nn[firstCurve]>0)
                        {
                        for (int i=1; i<nn[firstCurve];i++)
                            {
                            if (Math.abs(xval-x[firstCurve][i]) < nearestX)
                                {
                                nearestX = Math.abs(xval-x[firstCurve][i]);
                                nearestLine = i;
                                }
                            }
                        }
                    tpanel.setSelection(excludedHeadSamples+nearestLine*binSize[firstCurve], excludedHeadSamples+(nearestLine+1)*binSize[firstCurve]-1);
                    boldedDatum=nearestLine;
                    plotcoordlabel.setText("DATA: x="+fourPlaces.format(x[firstCurve][nearestLine])+", y="+fourPlaces.format(y[firstCurve][nearestLine]));
                    IJ.showStatus("data values: x="+fourPlaces.format(x[firstCurve][nearestLine])+", y="+fourPlaces.format(y[firstCurve][nearestLine]));


                    if (useBoldedDatum) updatePlot();
                    }
                else
                    {
                    plotcoordlabel.setText("x="+fourPlaces.format(xval)+", y="+fourPlaces.format(yval));
                    IJ.showStatus("plot coordinates: x="+fourPlaces.format(xval)+", y="+fourPlaces.format(yval));
                    if (boldedDatum!=-1)
                        {
                        boldedDatum = -1;
                        if (useBoldedDatum)updatePlot();
                        }
                    }
                }
    };

    static MouseWheelListener plotMouseWheelListener = new MouseWheelListener() {
        public void mouseWheelMoved( MouseWheelEvent e )
                {
                int screenX = e.getX();
                int screenY = e.getY();
                int magChangeSteps = e.getWheelRotation();
                zoomControl(screenX, screenY, magChangeSteps);
                }
    };

        static public void zoomControl(int screenX, int screenY, int magChangeSteps)
                {
                if (screenX > Plot.LEFT_MARGIN && screenX <= plotSizeX - Plot.RIGHT_MARGIN &&
                    screenY >= Plot.TOP_MARGIN && screenY <= plotSizeY - Plot.BOTTOM_MARGIN)
                    {
                    if (zoom == 0.0)
                        {
                        mouseX = screenX - Plot.LEFT_MARGIN;
                        mouseY = screenY - Plot.TOP_MARGIN;
//                        totalPanOffsetX=0.0;
//                        totalPanOffsetY=0.0;
//                        newPanOffsetX=0.0;
//                        newPanOffsetY=0.0;
                        }
                    zoom -= (1-zoom)*magChangeSteps/25.0;
                    zoom = zoom > 0.99 ? 0.99 : zoom;
                    zoom = zoom < 0.0 ? 0.0 : zoom;
                    updatePlot();
                    }
                }

        static public void zoomIn(int x, int y) {
                double magnification = plotImageCanvas.getMagnification();
                if (magnification>=32) return;
                double newMag = plotImageCanvas.getHigherZoomLevel(magnification);

                        int w = (int)Math.round(plotSizeX/newMag);
                if (w*newMag<plotSizeX) w++;
                int h = (int)Math.round(plotSizeY/newMag);
                if (h*newMag<plotSizeY) h++;
                x = plotImageCanvas.offScreenX(x);
                y = plotImageCanvas.offScreenY(y);
                Rectangle r = new Rectangle(x-w/2, y-h/2, w, h);
                if (r.x<0) r.x = 0;
                if (r.y<0) r.y = 0;
                if (r.x+w>plotSizeX) r.x = plotSizeX-w;
                if (r.y+h>plotSizeY) r.y = plotSizeY-h;
                plotImageCanvas.setSourceRect(r);
                plotImageCanvas.setMagnification(newMag);
                plotImageCanvas.repaint();
                }

static void initializeVariables()
        {
        title = "title";
        titlePosX = .5;
        titlePosY = 0.06;
        subtitle ="subtitle";
        subtitlePosX = 0.5;
        subtitlePosY = 0.1;
        legendPosX = 0.8;
        legendPosY = 0.2;
        button2Drag = false;
        priorityColumns = "j.d.,jd,hjd,bjd,ratio,rel_flux_t,rel_flux_c,source-sky,tot_c_cnts,alt,air,ccd,saturated,source_radius,sky_radius,x,y,peak,sky/pixel";
//        priorityErrColumns = "ratio_err,rel_flux_err_t,rel_flux_err_c,source_error_t,source_error_c,tot_c_err";
        prioritizeColumns = true;
        xlabeldefault = "J.D.-2400000";
        plotSizeX = 800;
        plotSizeY = 800;
        xExponent = 0;
        yExponent = 0;
        mmagrefs = 10;
        xLegend = "";
        yLegend = "";
        invertYAxis = false;
        invertYAxisSign = 1;
        negateMag = true;
        selectAnotherTableCanceled = false;
        vMarker1TopText = "Predicted";
        vMarker1BotText = "Ingress";
        vMarker2TopText = "Predicted";
        vMarker2BotText = "Egress";
        plotSuffix = "_lightcurve";
        configSuffix = "_measurements";
        dataSuffix = "_measurements";
        appendDestinationSuffix = "_A";
        appendSourceSuffix = "_B";
        combinedTableName = "Measurements";
        legendLeft = false;
        legendRight = false;
        useTitle = false;
        useSubtitle = false;
        useXCustomName = false;
        useYCustomName = false;
        useXColumnName = true;
        useYColumnName = true;
        xTics = true;
        yTics = true;
        xGrid = false;
        yGrid = true;
        xNumbers = true;
        yNumbers = true;
        plotAutoMode = false;
        nextPanel = false;
        canceled = false;
        waitSecs = false;
        openDataSetWindow = true;
        rememberWindowLocations = true;
        autoScaleX = true;
        useFirstX = false;
        autoScaleY = true;
        showToolTips = true;
        showXScaleInfo = true;
        showYScaleInfo = true;
        showYShiftInfo = true;
        showLScaleInfo = true;
        showLShiftInfo = true;
        showYBinInfo = true;
        showLBinInfo = true;
        showYmmagInfo = true;
        showLmmagInfo = true;
        showYSymbolInfo = false;
        showLSymbolInfo = true;
        useTwoLineLegend = false;
        useWideDataPanel = true;
        subframeWasShowing = true;
        newPlotWindow = false;
        useDefaultSettings = false;
        tableHasText = false;
        setSubpanelVisible = false;
        usePixelScale = true;
        deletedRowList = new ArrayList<String>();
        useBoldedDatum = true;
        panelShiftDown = false;
        panelControlDown = false;
        panelAltDown = false;     
        keepFileNamesOnAppend = true;
        updatePlotRunning = false;
        boldedDatum=-1;
        pixelScale = 1.0;

        xMin = 0.0;
        xMax = 0.0;
        xWidth = 0.3;
        yMin = 0.0;
        yMax = 0.0;
        xautoscalemin = 0.0;
        xautoscalemax = 0.0;
        yautoscalemin = 0.0;
        yautoscalemax = 0.0;
        yRange = 0.0;
        zoom = 0.0;

        mainFrameLocationX = 40;
        mainFrameLocationY = 40;
        subFrameLocationX = 40;
        subFrameLocationY = 40;
        plotFrameLocationX = 40;
        plotFrameLocationY = 40;

        excludedHeadSamples = 0;
        excludedTailSamples = 0;
        holdExcludedHeadSamples = 0;
        holdExcludedTailSamples = 0;
        excludedHeadSamplesStep = 1;
        excludedTailSamplesStep = 1;

        xWidthStep = 0.01;
        xMaxStep = 0.01;
        xMinStep = 1.0;
        yMaxStep = 1.0;
        yMinStep = 1.0;
        yMid = 0.0;
        firstXmin = 0.0;
        plotSizeXStep = 10;
        plotSizeYStep = 10;
        vMarker1Step = 0.01;
        vMarker2Step = 0.01;

        mainmenubar = new JMenuBar();
        filemenu = new JMenu("File");
        preferencesmenu = new JMenu("Preferences");

        tableName = "Results";
        list = new Vector<Roi>();
        path = new GeneralPath();

        fourPlaces = new DecimalFormat("0.0000");
        
        delayedUpdateTask = null;
        delayedUpdateTimer = null;        

        markers = new String[]
                {
                "box",
                "circle",
                "cross",
                "dot",
                "line",
                "triangle",
                "X"
                };
        colors = new String[]
                {
                "black",
                "dark gray",
                "gray",
                "light gray",
                "green",
                "dark green",
                "light blue",
                "blue",
                "magenta",
                "pink",
                "red",
                "orange",
                "yellow",
                "brown",
                "purple",
                "teal"
                };

        operators = new String[]
                {
                "none",
                "divide by",
                "multiply by",
                "subtract",
                "add",
                "centroid distance",
                "custom error"
                };

        opSymbol = new String[]
                {
                "",
                " / ",
                " * ",
                " - ",
                " + ",
                " -> ",
                " (with error) "
                };

        spinnerscalelist = new String[]
                {
                "      0.0000000001",
                "        0.000000001",
                "          0.00000001",
                "            0.0000001",
                "              0.000001",
                "                0.00001",
                "                  0.0001",
                "                    0.001",
                "                    0.010",
                "                    0.100",
                "                    0.500",
                "                    1.000",
                "                  10.000",
                "                100.000",
                "              1000.000",
                "            10000.000",
                "          100000.000",
                "        1000000.000",
                "      10000000.000",
                "    100000000.000",
                "1000000000.000",
                };

        integerspinnerscalelist = new String[]
                {
                "                    1",
                "                    2",
                "                    3",
                "                    4",
                "                    5",
                "                    6",
                "                    7",
                "                    8",
                "                    9",
                "                  10",
                "                  25",
                "                  50",
                "                100",
                "                250",
                "                500",
                "              1000"
                };
            }

    static public void setupArrays()
                {
                maxCurves = (int)Prefs.get("plot.maxCurves", 8);
                if (maxCurves < 1) maxCurves = 1;
                nn = new int[maxCurves];
                nnr = new int[maxCurves];
                xlabel2 = new String[maxCurves] ;

                setupDataBuffers();

                xlabel = new String[maxCurves];
                ylabel = new String[maxCurves];
                xcolumn = new int[maxCurves];
                ycolumn = new int[maxCurves];
                errcolumn = new int[maxCurves];
                operrcolumn = new int[maxCurves];
                xc1column = new int[maxCurves];
                xc2column = new int[maxCurves];
                yc1column = new int[maxCurves];
                yc2column = new int[maxCurves];
                oplabel = new String[maxCurves];
                opcolumn = new int[maxCurves];
                errlabel = new String[maxCurves];
                operrlabel = new String[maxCurves];
                showErrors = new boolean[maxCurves];
                showOpErrors = new boolean[maxCurves];
                lines = new boolean[maxCurves];
                marker = new int[maxCurves];
                color = new Color[maxCurves];
                markerIndex = new int[maxCurves];
                colorIndex = new int[maxCurves];
                operatorIndex = new int[maxCurves];
                cblabels = new String[maxCurves];
                moreOptions = new boolean [maxCurves];
                autoScaleFactor = new double[maxCurves];
                autoScaleStep = new double[maxCurves];
                autoShiftFactor = new double[maxCurves];
                autoShiftStep = new double[maxCurves];
                customScaleFactor = new double[maxCurves];
                customScaleStep = new double[maxCurves];
                customShiftFactor = new double[maxCurves];
                customShiftStep = new double[maxCurves];
                baseline = new double[maxCurves];
                subtotalScaleFactor = new double[maxCurves];
                subtotalShiftFactor = new double[maxCurves];
                totalScaleFactor = new double[maxCurves];
                totalShiftFactor = new double[maxCurves];
                yWidth = new double[maxCurves];
                xMinimum = new double[maxCurves];
                xMaximum = new double[maxCurves];
                yMinimum = new double[maxCurves];
                yMn = new double[maxCurves];
                yMaximum = new double[maxCurves];
                yMx = new double[maxCurves];
                yMidpoint = new double[maxCurves];
                mmag = new boolean[maxCurves];
                plotY = new boolean[maxCurves];
                useColumnName = new boolean[maxCurves];
                useLegend = new boolean[maxCurves];
                force = new boolean[maxCurves];
                binSize = new int[maxCurves];
                legend = new String[maxCurves];
                xyc1label = new String[maxCurves];
                xyc2label = new String[maxCurves];
                ASInclude = new boolean[maxCurves];
                autoscaleincludebox = new JCheckBox[maxCurves];

                usecurvebox = new JCheckBox[maxCurves];
                errorcolumnbox = new JCheckBox[maxCurves];
//                operrorcolumnbox = new JCheckBox[maxCurves];
                usebinbox = new JCheckBox[maxCurves];
                usemmagbox = new JCheckBox[maxCurves];
                uselinesbox = new JCheckBox[maxCurves];
                forcebox = new JCheckBox[maxCurves];
                xdatacolumn = new JComboBox[maxCurves];
                ydatacolumn = new JComboBox[maxCurves];
                operatorselection = new JComboBox[maxCurves];
                operatorcolumn = new JComboBox[maxCurves];
                errorcolumn = new JComboBox[maxCurves];
                operrorcolumn = new JComboBox[maxCurves];
                markersymbolselection = new JComboBox[maxCurves];
                markercolorselection = new JComboBox[maxCurves];
                binsizespinnermodel = new SpinnerModel[maxCurves];
                binsizespinner = new JSpinner[maxCurves];
                morelegendradiopanelgroup = new JPanel[maxCurves];
                autoscalepanelgroup = new JPanel[maxCurves];
                customscalepanelgroup = new JPanel[maxCurves];
                legendnoneButton = new JRadioButton[maxCurves];
                legendcustomNameButton = new JRadioButton[maxCurves];
                legendcolumnNameButton = new JRadioButton[maxCurves];
                autoscaleButton = new JRadioButton[maxCurves];
                customscaleButton = new JRadioButton[maxCurves];
                morelegendRadioGroup = new ButtonGroup[maxCurves];
                autoscaleRadioGroup = new ButtonGroup[maxCurves];
                morelegendField = new JTextField[maxCurves];
                autoshiftmodel = new SpinnerModel[maxCurves];
                autoshiftstepmodel = new SpinnerModel[maxCurves];
                autoscalemodel  = new SpinnerModel[maxCurves];
                autoscalestepmodel  = new SpinnerModel[maxCurves];
                customshiftmodel = new SpinnerModel[maxCurves];
                customshiftstepmodel = new SpinnerModel[maxCurves];
                customscalemodel = new SpinnerModel[maxCurves];
                customscalestepmodel = new SpinnerModel[maxCurves];
                autoshiftspinner = new JSpinner[maxCurves];
                autoshiftstepspinner = new JSpinner[maxCurves];
                autoscalespinner  = new JSpinner[maxCurves];
                autoscalestepspinner  = new JSpinner[maxCurves];
                customshiftspinner = new JSpinner[maxCurves];
                customshiftstepspinner = new JSpinner[maxCurves];
                customscalespinner = new JSpinner[maxCurves];
                customscalestepspinner = new JSpinner[maxCurves];
                autoscalesteppopup = new JPopupMenu[maxCurves];
                autoshiftsteppopup = new JPopupMenu[maxCurves];
                customscalesteppopup = new JPopupMenu[maxCurves];
                customshiftsteppopup = new JPopupMenu[maxCurves];
                autoscalesteppanel = new JPanel[maxCurves];
                customscalesteppanel = new JPanel[maxCurves];
                autoshiftsteppanel = new JPanel[maxCurves];
                customshiftsteppanel = new JPanel[maxCurves];
                autoscalesteplabel = new JLabel[maxCurves];
                autoshiftsteplabel = new JLabel[maxCurves];
                customscalesteplabel = new JLabel[maxCurves];
                customshiftsteplabel = new JLabel[maxCurves];
                curvelabel = new JLabel[maxCurves];
                othercurvelabel = new JLabel[maxCurves];
                grabautopanel = new JPanel[maxCurves];
                grabautobutton = new JButton[maxCurves];

                customscaleeditor = new JSpinner.NumberEditor[maxCurves];
                customscaleformat = new DecimalFormat[maxCurves];
                customshifteditor = new JSpinner.NumberEditor[maxCurves];
                customshiftformat = new DecimalFormat[maxCurves];

                //INITIALIZE CERTAIN ARRAYS
                for (int i=0; i<maxCurves; i++)
                        {
                        xlabel[i] = "default";
                        ylabel[i] = "rel_flux_"+(i==0?"T":"C")+(i+1);
                        errlabel[i] = "rel_flux_err_"+(i==0?"T":"C")+(i+1);
                        oplabel[i] = "source-sky_"+(i==0?"T":"C")+(i+1);
                        operrlabel[i] = "source_error_"+(i==0?"T":"C")+(i+1);
                        colorIndex[i] = (7+i)%colors.length;  //start with blue
                        markerIndex[i] = 3;   //default to dot
                        force[i] = false;
                        showErrors[i] = false;
                        if (i==0)
                                {
                                plotY[i] = true;
                                ASInclude[i]= true;
                                }
                        else
                                {
                                plotY[i] = false;
                                ASInclude[i] = false;
                                }
                        autoScaleFactor[i] = 0.2;
                        autoScaleStep[i] = 1;
                        autoShiftFactor[i] = 0-i*10;
                        autoShiftStep[i] = 1;
                        customScaleFactor[i] = 1;
                        customScaleStep[i] = 1;
                        customShiftFactor[i] = 0-i*10;
                        customShiftStep[i] = 1;
                        binSize[i] = 1;
                        baseline[i] = 0;
                        legend[i] = "Legend"+(i+1);
                        }
                }

    static void setupDataBuffers()
        {
        x = new double[maxCurves][maxColumnLength];
        y = new double[maxCurves][maxColumnLength];
        yerr = new double[maxCurves][maxColumnLength];
        yop = new double[maxCurves][maxColumnLength];
        yoperr = new double[maxCurves][maxColumnLength];

        xc1 = new double[maxCurves][maxColumnLength];
        xc2 = new double[maxCurves][maxColumnLength];
        yc1 = new double[maxCurves][maxColumnLength];
        yc2 = new double[maxCurves][maxColumnLength];
        }

    /**
	 * Returns minimum of double array.
	 */
	static double minOf(double[] arr, int n)
		{
        double mn = Double.POSITIVE_INFINITY;
		for (int i=0; i < n; i++)
            {
            if (!Double.isNaN(arr[i]))
                {
                mn = (arr[i] < mn) ? arr[i] : mn;
                }
            }
		return mn;
		}

	/**
	 * Returns maximum of double array.
	 */
	static double maxOf(double[] arr, int n)
		{
		double mx = Double.NEGATIVE_INFINITY;
		for (int i=0; i < n; i++)
            {
            if (!Double.isNaN(arr[i]))
                {
                mx = (arr[i] > mx)? arr[i] : mx;
                }
            }
		return mx;
		}

    /**
	 * Returns minimum of double array including error.
	 */
	static double minOf(double[] arr, double[] err, int n)
		{
        double mn = Double.POSITIVE_INFINITY;
        double value = 0; 
		for (int i=0; i < n; i++)
            {
            if (!Double.isNaN(arr[i]) && !Double.isNaN(err[i]))
                {
                value = arr[i] - err[i];
                mn = (value < mn) ? value : mn;
                }
            }
		return mn;
		}

	/**
	 * Returns maximum of double array including.
	 */
	static double maxOf(double[] arr, double[] err, int n)
		{
		double mx = Double.NEGATIVE_INFINITY;
        double value = 0; 
		for (int i=0; i < n; i++)
            {
            if (!Double.isNaN(arr[i]) && !Double.isNaN(err[i]))
                {
                value = arr[i] + err[i];
                mx = (value > mx)? value : mx;
                }
            }
		return mx;
		}    
    
    
    static Color colorOf(int ci)
            {
            Color cmkr;
            switch (ci)
                {
                case  0: cmkr = java.awt.Color.black;       break;
                case  1: cmkr = java.awt.Color.darkGray;    break;
                case  2: cmkr = java.awt.Color.gray;        break;
                case  3: cmkr = java.awt.Color.lightGray;   break;
                case  4: cmkr = new Color(0,235,0);         break;   //green
                case  5: cmkr = new Color(0,155,0);         break;   //dark green
                case  6: cmkr = new Color(84,201,245);      break;   //light blue
                case  7: cmkr = java.awt.Color.blue;        break;
                case  8: cmkr = java.awt.Color.magenta;     break;
                case  9: cmkr = java.awt.Color.pink;        break;
                case 10: cmkr = java.awt.Color.red;         break;
                case 11: cmkr = java.awt.Color.orange;      break;
                case 12: cmkr = new Color(235,235,0);       break;   //yellow
                case 13: cmkr = new Color(167,131,96);      break;   //brown
                case 14: cmkr = new Color(124,0,255);       break;   //purple
                case 15: cmkr = new Color(3,148,163);       break;   //teal
                default: cmkr =  java.awt.Color.black;      break;
                }
            return cmkr;
                }

        static int markerOf(int mi)
                {
                int mkr = 3;
                if      (mi == 0) mkr = ij.gui.Plot.BOX;
                else if (mi == 1) mkr = ij.gui.Plot.CIRCLE;
                else if (mi == 2) mkr = ij.gui.Plot.CROSS;
                else if (mi == 3) mkr = ij.gui.Plot.DOT;
                else if (mi == 4) mkr = ij.gui.Plot.LINE;
                else if (mi == 5) mkr = ij.gui.Plot.TRIANGLE;
                else if (mi == 6) mkr = ij.gui.Plot.X;
                else              mkr = ij.gui.Plot.DOT;
                return mkr;
                }


        static String scaleShiftText(boolean showScale, boolean showShift, boolean usemmag, boolean showmmaginfo, double scale, double shift)
                {  
                String shiftText = "";
                String scaleText = "";
                NumberFormat Eformatter = new DecimalFormat("0.###E0");
                NumberFormat dformatter = new DecimalFormat("##0.###");
                if (showScale && showmmaginfo && mmag[firstCurve] && usemmag && totalScaleFactor[firstCurve]==0.001 && scale!=0.001)
                    {
                    scale = 1000*scale;
                    if (scale == 1.0)
                        scaleText +="";
                    else if(((scale > 0.001) && (scale < 1000.0)) || ((scale < -0.001) && (scale > -1000.0)))
                        scaleText += " x("+dformatter.format(1/scale)+")";
                    else
                        scaleText += " x("+Eformatter.format(1/scale)+")";
                    }
                else if(showScale && (scale != 1.0))
                    {
                    if (((scale > 0.001) && (scale < 1000.0)) || ((scale < -0.001) && (scale > -1000.0)))
                        scaleText += " x("+dformatter.format(1/scale)+")";
                    else
                        if (usemmag && showmmaginfo && (scale == 0.001))
                            scaleText +="";
                        else
                            scaleText += " x("+Eformatter.format(1/scale)+")";
                    }
                if (showShift)
                        {
                        if (shift < 0.0)
                                {
                                shift *= -1.0;
                                shiftText += " - ";
                                }
                        else if (shift != 0.0)
                                shiftText += " + ";
                        if ((shift >= 0.01) && (shift < 100.0))
                                shiftText += dformatter.format(shift);
                        else if (shift != 0.0)
                                shiftText += Eformatter.format(shift);
                        }
                return scaleText + shiftText;
                }

        static String convertToText(double increment)
                {
                if      (increment == 0.0000000001) return "      0.0000000001";
                else if (increment == 0.000000001)  return "        0.000000001";
                else if (increment == 0.00000001)   return "          0.00000001";
                else if (increment == 0.0000001)    return "            0.0000001";
                else if (increment == 0.000001)     return "              0.000001";
                else if (increment == 0.00001)      return "                0.00001";
                else if (increment == 0.0001)       return "                  0.0001";
                else if (increment == 0.001)        return "                    0.001";
                else if (increment == 0.01)         return "                    0.010";
                else if (increment == 0.1)          return "                    0.100";
                else if (increment == 0.5)          return "                    0.500";
                else if (increment == 1.0)          return "                    1.000";
                else if (increment == 10.0)         return "                  10.000";
                else if (increment == 100.0)        return "                100.000";
                else if (increment == 1000.0)       return "              1000.000";
                else if (increment == 10000.0)      return "            10000.000";
                else if (increment == 100000.0)     return "          100000.000";
                else if (increment == 1000000.0)    return "        1000000.000";
                else if (increment == 10000000.0)   return "      10000000.000";
                else if (increment == 100000000.0)  return "    100000000.000";
                else if (increment == 1000000000.0) return "1000000000.000";
                else return "                    1.000";
                }

        static String intConvertToText(int increment)
                {
                if      (increment == 1)          return "                    1";
                else if (increment == 2)          return "                    2";
                else if (increment == 3)          return "                    3";
                else if (increment == 4)          return "                    4";
                else if (increment == 5)          return "                    5";
                else if (increment == 6)          return "                    6";
                else if (increment == 7)          return "                    7";
                else if (increment == 8)          return "                    8";
                else if (increment == 9)          return "                    9";
                else if (increment == 10)         return "                  10";
                else if (increment == 25)         return "                  25";
                else if (increment == 50)         return "                  50";
                else if (increment == 100)        return "                100";
                else if (increment == 250)        return "                250";
                else if (increment == 500)        return "                500";
                else if (increment == 1000)       return "              1000";
//                else if (increment == 10000)      return "            10000";
//                else if (increment == 100000)     return "          100000";
//                else if (increment == 1000000)    return "        1000000";
//                else if (increment == 10000000)   return "      10000000";
//                else if (increment == 100000000)  return "    100000000";
//                else if (increment == 1000000000) return "1000000000";
                else return "                    1";
                }

       static void checkForUT(JSpinner spinner)
                {
                JSpinner.NumberEditor ed = (JSpinner.NumberEditor)spinner.getEditor();
                String text = ed.getTextField().getText();
                String[] hms;
                if (Double.isNaN(Tools.parseDouble(text)))
                    {
                    Double value = 0.5;
                    String[] pieces = text.replaceAll("[^0-9\\.:]{1,}", " ").trim().split("[ :]{1,}");
                    try {if (pieces.length > 0) value += Double.parseDouble(pieces[0])/24.0;}
                    catch (NumberFormatException e) { }
                    try {if (pieces.length > 1) value += Double.parseDouble(pieces[1])/1440.0;}
                    catch (NumberFormatException e) { }
                    try {if (pieces.length > 2) value += Double.parseDouble(pieces[2])/86400.0;}
                    catch (NumberFormatException e) { }
                    spinner.setValue(value%1);
                    }
                }


       static void showMainJPanel()
                {
// DISPLAY MAIN PANEL

                mainFrame = new JFrame ("Multi-plot Main");
                plotIcon = createImageIcon("astroj/images/plot.png", "Plot Icon");
                mainFrame.setIconImage(plotIcon.getImage());
                mainFrame.setDefaultCloseOperation(javax.swing.WindowConstants.DISPOSE_ON_CLOSE);
                mainFrame.addWindowListener(new WindowAdapter(){
                    @Override
                    public void windowClosing(WindowEvent e){
                    saveAndClose();}});
                
                mainpanel = new JPanel (new SpringLayout());
//                mainpanel.addMouseListener(panelMouseListener);
                mainpanel.addMouseMotionListener(panelMouseMotionListener);
                FileDrop fileDrop = new FileDrop(mainpanel, BorderFactory.createEmptyBorder(),new FileDrop.Listener()
                    {   public void filesDropped( java.io.File[] files )
                        {
                        openDragAndDropFiles(files);
                        }
                    });
               
                
                mainscrollpane = new JScrollPane(mainpanel);
                
                mainmenubar = new JMenuBar();
                filemenu = new JMenu("  File      ");
                preferencesmenu = new JMenu("Preferences    ");
                
                mainpanelgroupa = new JPanel (new SpringLayout());
                
                mainpanelgroupa.setBorder(dataSectionBorder);
                JPanel mainpanelgroupb = new JPanel (new SpringLayout());
                mainpanelgroupb.setBorder(BorderFactory.createTitledBorder("Data Options"));
                JPanel mainpanelgroupc = new JPanel (new SpringLayout());   //TITLE - SUBTITLE - LEGEND GROUP
                JPanel mainpanelgroupd = new JPanel (new SpringLayout());   // X & Y AXIS LEGEND GROUP
                JPanel mainpanelgroupe = new JPanel (new SpringLayout());
                mainpanelgroupe.setBorder(BorderFactory.createTitledBorder("Scaling Options"));
                JPanel mainpanelgroupf = new JPanel (new SpringLayout());

    // BUILD MENUBAR

                opentablemenuitem = new JMenuItem("Select open table...");
                opentablemenuitem.setToolTipText("<html>"+"select a table that is already open for plotting"+"</html>");
                opentablemenuitem.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                        selectAnotherTableCanceled = false;
                        MeasurementTable newTable = selectAnotherTable("");
                        if (newTable == null)
                            {
                            if (!selectAnotherTableCanceled)
                                {
                                makeDummyTable();
                                IJ.beep();
                                IJ.showMessage("No tables open");
                                }
                            selectAnotherTableCanceled = false;
                            return;
                            }
                        newTable.show();
                        setTable(newTable, true);
                        plotWindow.setVisible(true);
                        }});
                filemenu.add(opentablemenuitem);

                JMenuItem appendtablerowsmenuitem = new JMenuItem("Append open table as new rows...");
                appendtablerowsmenuitem.setToolTipText("<html>"+"select a table that is already open to append to current table"+"</html>");
                appendtablerowsmenuitem.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                            appendDataAsRows(false, null); }});
                filemenu.add(appendtablerowsmenuitem);                

                JMenuItem appendtablecolumnssmenuitem = new JMenuItem("Append open table as new columns...");
                appendtablecolumnssmenuitem.setToolTipText("<html>"+"select a table that is already open to append to current table"+"</html>");
                appendtablecolumnssmenuitem.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                            appendDataAsColumns(false, null); }});
                filemenu.add(appendtablecolumnssmenuitem); 
                filemenu.addSeparator();

                opendatamenuitem = new JMenuItem("Open table from file...");
                opendatamenuitem.setToolTipText("<html>"+"default input format = tab delimited, or use"+"<br>"+
                    "filename.csv = comma delimited"+"<br>"+
                    "filename.prn or filename.spc = space delimted"+"<br>"+
                    "---------------------------------------------"+"<br>"+
                    "first line should be column headings delimited as stated above"+"<br>"+
                    "lines starting with # are considered comments and ignored, except"+"<br>"+
                    "the last comment before the first data line can be headings"+"</html>");
                opendatamenuitem.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                            openData(); }});
                filemenu.add(opendatamenuitem);
                
                JMenuItem appenddatarowsmenuitem = new JMenuItem("Append table from file as new rows...");
                appenddatarowsmenuitem.setToolTipText("<html>"+"default input format = tab delimited, or use"+"<br>"+
                    "filename.csv = comma delimited"+"<br>"+
                    "filename.prn or filename.spc = space delimted"+"<br>"+
                    "---------------------------------------------"+"<br>"+
                    "first line should be column headings delimited as stated above"+"<br>"+
                    "lines starting with # are considered comments and ignored, except"+"<br>"+
                    "the last comment before the first data line can be headings"+"</html>");
                appenddatarowsmenuitem.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                            appendDataAsRows(true, null); }});
                filemenu.add(appenddatarowsmenuitem);
                
                JMenuItem appenddatacolumnsmenuitem = new JMenuItem("Append table from file as new columns...");
                appenddatacolumnsmenuitem.setToolTipText("<html>"+"default input format = tab delimited, or use"+"<br>"+
                    "filename.csv = comma delimited"+"<br>"+
                    "filename.prn or filename.spc = space delimted"+"<br>"+
                    "---------------------------------------------"+"<br>"+
                    "first line should be column headings delimited as stated above"+"<br>"+
                    "lines starting with # are considered comments and ignored, except"+"<br>"+
                    "the last comment before the first data line can be headings"+"</html>");
                appenddatacolumnsmenuitem.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                            appendDataAsColumns(true, null); }});
                filemenu.add(appenddatacolumnsmenuitem);
                
                openplotconfigmenuitem = new JMenuItem("Open plot configuration from file...");
                openplotconfigmenuitem.setToolTipText("<html>"+"opens a previously saved plot configuration<br>"+
                                                               "from a user selected file</html>");
                openplotconfigmenuitem.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                            openConfig(); }});
                filemenu.add(openplotconfigmenuitem);

                opendataconfigmenuitem = new JMenuItem("Open table and plot configuration from file...");
                opendataconfigmenuitem.setToolTipText("<html>"+ "opens a previously saved measurement table"+"<br>"+
                                                                "from a user selected file and attempts to"+"<br>"+
                                                                "open a plot configuration file with the"+"<br>"+
                                                                "same name but ending in .plotcfg"+"</html>");
                opendataconfigmenuitem.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                            openDataAndConfig(null); }});
                filemenu.add(opendataconfigmenuitem);
               
                filemenu.addSeparator();

                savedatamenuitem = new JMenuItem("Save data to file...");
                savedatamenuitem.setToolTipText("<html>"+"saves measurement table data to a user selected file"+"</html>");
                savedatamenuitem.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                            saveData(); }});
                filemenu.add(savedatamenuitem);

                JMenuItem savedatasubsetmenuitem = new JMenuItem("Save data subset to file...");
                savedatasubsetmenuitem.setToolTipText("<html>"+"saves a subset of measurement table data to a user selected file"+"</html>");
                savedatasubsetmenuitem.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                            saveDataSubsetDialog(); }});
                filemenu.add(savedatasubsetmenuitem);
                
                JMenuItem saveimagepngmenuitem = new JMenuItem("Save plot image as PNG...");
                saveimagepngmenuitem.setToolTipText("<html>"+"saves plot image as a .png file"+"</html>");
                saveimagepngmenuitem.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                            savePlotImageAsPng(); }});
                filemenu.add(saveimagepngmenuitem);

                JMenuItem saveimagejpgmenuitem = new JMenuItem("Save plot image as JPG...");
                saveimagejpgmenuitem.setToolTipText("<html>"+"saves plot image as a .jpg file"+"</html>");
                saveimagejpgmenuitem.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                            savePlotImageAsJpg(); }});
                filemenu.add(saveimagejpgmenuitem);

                saveplotconfigmenuitem = new JMenuItem("Save plot configuration...");
                saveplotconfigmenuitem.setToolTipText("<html>"+"saves plot configuration to a user selected file</html>");
                saveplotconfigmenuitem.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                            saveConfig(); }});
                filemenu.add(saveplotconfigmenuitem);

                filemenu.addSeparator();

                savedataconfigmenuitem = new JMenuItem("Save data and plot configuration...");
                savedataconfigmenuitem.setToolTipText("<html>"+"Saves measurement table data to a user selected file<br>"+
                                                               "using the suffix defined in suffix settings.<br>"+
                                                               "Saves the plot configuration to a file with the<br>"+
                                                               "same name but ending with the suffix defined plus .plotcfg"+"</html>");
                savedataconfigmenuitem.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                            saveDataImageConfig(false, true, true, false, "", null, null, null); }});
                filemenu.add(savedataconfigmenuitem);


                JMenuItem saveimagepngconfigmenuitem = new JMenuItem("Save plot image (PNG), and configuration...");
                saveimagepngconfigmenuitem.setToolTipText("<html>"+"Saves the plot image to a user selected file<br>"+
                                                               "using the suffix defined in suffix settings plus .png.<br>"+
                                                               "Saves the plot configuration to a file with the<br>"+
                                                               "same name but ending with the suffix defined plus .plotcfg"+"</html>");
                saveimagepngconfigmenuitem.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                            saveDataImageConfig(true, true, false, false, "png", null, null, null); }});
                filemenu.add(saveimagepngconfigmenuitem);

                JMenuItem saveimagejpgconfigmenuitem = new JMenuItem("Save plot image (JPG), and configuration...");
                saveimagejpgconfigmenuitem.setToolTipText("<html>"+"Saves the plot image to a user selected file<br>"+
                                                               "using the suffix defined in suffix settings plus .jpg.<br>"+
                                                               "Saves the plot configuration to a file with the<br>"+
                                                               "same name but ending with the suffix defined plus .plotcfg"+"</html>");
                saveimagejpgconfigmenuitem.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                            saveDataImageConfig(true, true, false, false, "jpg", null, null, null); }});
                filemenu.add(saveimagejpgconfigmenuitem);

                JMenuItem savedataimagepngconfigmenuitem = new JMenuItem("Save data, plot image (PNG), and configuration...");
                savedataimagepngconfigmenuitem.setToolTipText("<html>"+"Saves measurement table data to a user selected file<br>"+
                                                               "using the suffix defined in suffix settings.<br>"+
                                                               "Saves the plot image to a file with the<br>"+
                                                               "same name but ending with the suffix defined plus .png.<br>"+
                                                               "Saves the plot configuration to a file with the<br>"+
                                                               "same name but ending with the suffix defined plus .plotcfg"+"</html>");
                savedataimagepngconfigmenuitem.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                            saveDataImageConfig(true, true, true, false, "png", null, null, null); }});
                filemenu.add(savedataimagepngconfigmenuitem);

                JMenuItem savedataimagejpgconfigmenuitem = new JMenuItem("Save data, plot image (JPG), and configuration...");
                savedataimagejpgconfigmenuitem.setToolTipText("<html>"+"Saves measurement table data to a user selected file<br>"+
                                                               "using the suffix defined in suffix settings.<br>"+
                                                               "Saves the plot image to a file with the<br>"+
                                                               "same name but ending with the suffix defined plus .jpg.<br>"+
                                                               "Saves the plot configuration to a file with the<br>"+
                                                               "same name but ending with the suffix defined plus .plotcfg"+"</html>");
                savedataimagejpgconfigmenuitem.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                            saveDataImageConfig(true, true, true, false, "jpg", null, null, null); }});
                filemenu.add(savedataimagejpgconfigmenuitem);

                JMenuItem setSaveStateMenuItem = new JMenuItem("    suffix settings...");
                setSaveStateMenuItem.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                            setSaveStateDialog(); }});
                filemenu.add(setSaveStateMenuItem);

                filemenu.addSeparator();

                JMenuItem exitmenuitem = new JMenuItem("Exit");
                exitmenuitem.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                            saveAndClose(); }});
                filemenu.add(exitmenuitem);
                
                mainmenubar.add(filemenu);

                opendatasetCB = new JCheckBoxMenuItem("Open Y-data columns window at startup",openDataSetWindow);
                opendatasetCB.addItemListener(new ItemListener(){
                    public void itemStateChanged(ItemEvent e) {
                        if (e.getStateChange() == ItemEvent.DESELECTED)
                                openDataSetWindow = false;
                        else if (e.getStateChange() == ItemEvent.SELECTED)
                                openDataSetWindow = true;
                        Prefs.set("plot.openDataSetWindow",openDataSetWindow); }});
                preferencesmenu.add(opendatasetCB);

                usewidedataCB = new JCheckBoxMenuItem("Use --wide-- Y-data columns window",useWideDataPanel);
                usewidedataCB.addItemListener(new ItemListener(){
                    public void itemStateChanged(ItemEvent e) {
                        if (e.getStateChange() == ItemEvent.DESELECTED)
                                useWideDataPanel = false;
                        else if (e.getStateChange() == ItemEvent.SELECTED)
                                useWideDataPanel = true;
                        Prefs.set("plot.useWideDataPanel",useWideDataPanel);
                        subframeWasShowing = false;
                        if (subFrame.isShowing())
                            {
                            subFrameLocationX = subFrame.getLocation().x;
                            subFrameLocationY = subFrame.getLocation().y;
                            Prefs.set("plot2.subFrameLocationX",subFrameLocationX);
                            Prefs.set("plot2.subFrameLocationY",subFrameLocationY);
                            subFrame.dispose();
                            subframeWasShowing = true;
                            }

                        if (subframeWasShowing)
                                showMoreCurvesJPanel();
                        }});

                preferencesmenu.add(usewidedataCB);  

                rememberwindowlocationsCB = new JCheckBoxMenuItem("Open windows at previous locations",rememberWindowLocations);
                rememberwindowlocationsCB.addItemListener(new ItemListener(){
                    public void itemStateChanged(ItemEvent e) {
                        if (e.getStateChange() == ItemEvent.DESELECTED)
                                rememberWindowLocations = false;
                        else if (e.getStateChange() == ItemEvent.SELECTED)
                                rememberWindowLocations = true;
                        Prefs.set("plot.rememberWindowLocations",rememberWindowLocations); }});
                preferencesmenu.add(rememberwindowlocationsCB);


                showtooltipsCB = new JCheckBoxMenuItem("Show tooltips help",showToolTips);
                showtooltipsCB.addItemListener(new ItemListener(){
                    public void itemStateChanged(ItemEvent e) {
                        if (e.getStateChange() == ItemEvent.DESELECTED)
                            {
                            showToolTips=false;
                            }
                        else if (e.getStateChange() == ItemEvent.SELECTED)
                            {
                            showToolTips=true;
                            }
                        ToolTipManager.sharedInstance().setEnabled(showToolTips);
                        Prefs.set ("astroIJ.showToolTips",showToolTips);                        
                        }});
                preferencesmenu.add(showtooltipsCB);

                preferencesmenu.addSeparator();

                useBoldedDatumCB = new JCheckBoxMenuItem("Bold data point nearest mouse while holding shift",useBoldedDatum);
                useBoldedDatumCB.addItemListener(new ItemListener(){
                    public void itemStateChanged(ItemEvent e) {
                        if (e.getStateChange() == ItemEvent.DESELECTED)
                                useBoldedDatum = false;
                        else if (e.getStateChange() == ItemEvent.SELECTED)
                                useBoldedDatum = true;
                        Prefs.set("plot.useBoldedDatum",useBoldedDatum);
                        updatePlot();}});
                preferencesmenu.add(useBoldedDatumCB);

                setverticalmarkertextmenuitem = new JMenuItem("Change vertical marker text...");
                setverticalmarkertextmenuitem.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                            setVMarkerText();
                            updatePlot();}});
                preferencesmenu.add(setverticalmarkertextmenuitem);
                
                JCheckBoxMenuItem invertyaxisCB = new JCheckBoxMenuItem ("Invert Y-axis",invertYAxis);
                invertyaxisCB.addItemListener(new ItemListener(){
                    public void itemStateChanged(ItemEvent e) {
                        if (e.getStateChange() == ItemEvent.DESELECTED)
                            {
                            invertYAxis = false;
                            invertYAxisSign = 1;
                            }
                        else if (e.getStateChange() == ItemEvent.SELECTED)
                            {
                            invertYAxis = true;
                            invertYAxisSign = -1;
                            }
                        for(int c=0; c<maxCurves;c++)
                            {
                            customshiftspinner[c].setModel(new SpinnerNumberModel(new Double(force[c]?autoShiftFactor[c]*100:customShiftFactor[c]), null, null,
                                                         new Double(force[c]?invertYAxisSign*autoShiftStep[c]:invertYAxisSign*customShiftStep[c])));
                            }
                        Prefs.set("plot.invertYAxis",invertYAxis);
                        updatePlot();}});
                preferencesmenu.add(invertyaxisCB);

                JCheckBoxMenuItem negateMagCB = new JCheckBoxMenuItem ("Negate relative magnitude calculations",negateMag);
                negateMagCB.addItemListener(new ItemListener(){
                    public void itemStateChanged(ItemEvent e) {
                        if (e.getStateChange() == ItemEvent.DESELECTED)
                                negateMag = false;
                        else if (e.getStateChange() == ItemEvent.SELECTED)
                                negateMag = true;
                        Prefs.set("plot.negateMag",negateMag);
                        updatePlot();}});
                preferencesmenu.add(negateMagCB);

                preferencesmenu.addSeparator();

                usePriorityColumnsCB = new JCheckBoxMenuItem("Show specified priority data names first in pulldown lists",prioritizeColumns);
                usePriorityColumnsCB.addItemListener(new ItemListener(){
                    public void itemStateChanged(ItemEvent e) {
                        if (e.getStateChange() == ItemEvent.DESELECTED)
                                prioritizeColumns = false;
                        else if (e.getStateChange() == ItemEvent.SELECTED)
                                prioritizeColumns = true;
                        Prefs.set("plot.prioritizeColumns",prioritizeColumns);
                        oldUnfilteredColumns = null;
                        updatePlot();}});
                preferencesmenu.add(usePriorityColumnsCB);

                changePriorityColumnsmenuitem = new JMenuItem("Change priority data names list...");
                changePriorityColumnsmenuitem.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                            changePriorityColumns();
                            if (prioritizeColumns)
                                {
                                oldUnfilteredColumns = null;
                                updatePlot();
                                } }});
                preferencesmenu.add(changePriorityColumnsmenuitem);

                preferencesmenu.addSeparator();

                usepixelscaleCB = new JCheckBoxMenuItem("Use pixel scale when calculating distance",usePixelScale);
                usepixelscaleCB.addItemListener(new ItemListener(){
                    public void itemStateChanged(ItemEvent e) {
                        if (e.getStateChange() == ItemEvent.DESELECTED)
                                usePixelScale = false;
                        else if (e.getStateChange() == ItemEvent.SELECTED)
                                usePixelScale = true;
                        Prefs.set("plot.usePixelScale",usePixelScale);
                        updatePlot();}});
                preferencesmenu.add(usepixelscaleCB);

                changepixelscalemenuitem = new JMenuItem("Set pixel scale ("+pixelScale+")...");
                changepixelscalemenuitem.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                            changePixelScale();
                            updatePlot();}});
                preferencesmenu.add(changepixelscalemenuitem);

                preferencesmenu.addSeparator();

                changemaxdatalengthmenuitem = new JMenuItem("Set minimum data column buffer length...");
                changemaxdatalengthmenuitem.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                            changeMaxDataLength();
                            updatePlot();}});
                preferencesmenu.add(changemaxdatalengthmenuitem);

                usedefaultsettingsCB = new JCheckBoxMenuItem("Reset preferences to default settings (restart required)",useDefaultSettings);
                usedefaultsettingsCB.addItemListener(new ItemListener(){
                    public void itemStateChanged(ItemEvent e) {
                        if (e.getStateChange() == ItemEvent.DESELECTED)
                                useDefaultSettings = false;
                        else if (e.getStateChange() == ItemEvent.SELECTED)
                                useDefaultSettings = true;
                        Prefs.set("plot.useDefaultSettings",useDefaultSettings); }});
                preferencesmenu.add(usedefaultsettingsCB);

                mainmenubar.add(preferencesmenu);
                                
                JMenu helpmenu = new JMenu("Help");
                JMenuItem helpmenuitem = new JMenuItem("General help...");
                helpmenuitem.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                            openHelpPanel(); }});
                helpmenu.add(helpmenuitem);
                
                JMenuItem helpdatamenuitem = new JMenuItem("Data naming convention...");
                helpdatamenuitem.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                            openDataHelpPanel(); }});
                helpmenu.add(helpdatamenuitem);                 

                mainmenubar.add(helpmenu);                



    // MAIN PANEL GROUP A START
                xcolumnlabel = new JLabel ("Default X-data");
                xcolumnlabel.setHorizontalAlignment(JLabel.CENTER);
                mainpanelgroupa.add (xcolumnlabel);

                ycolumnslabel = new JLabel ("Y-data");
                ycolumnslabel.setHorizontalAlignment(JLabel.CENTER);
                mainpanelgroupa.add (ycolumnslabel);

                JLabel dummy50label = new JLabel ("Number of Y-datasets");
                dummy50label.setHorizontalAlignment(JLabel.CENTER);
                mainpanelgroupa.add (dummy50label);

                JLabel mmagsampleslabel = new JLabel ("Rel. Mag. Reference");
                mmagsampleslabel.setHorizontalAlignment(JLabel.CENTER);
                mmagsampleslabel.setToolTipText("The number of starting samples averaged to calculate relative magnitude reference level");
                mainpanelgroupa.add (mmagsampleslabel);

                JLabel vmarker1toplabel = new JLabel ("V. Marker 1");
                vmarker1toplabel.setHorizontalAlignment(JLabel.CENTER);
                vmarker1toplabel.setToolTipText("Enter vertical marker 1 x-axis location");
                mainpanelgroupa.add (vmarker1toplabel);

                JLabel vmarker2toplabel = new JLabel ("V. Marker 2");
                vmarker2toplabel.setHorizontalAlignment(JLabel.CENTER);
                vmarker2toplabel.setToolTipText("Enter vertical marker 2 x-axis location");
                mainpanelgroupa.add (vmarker2toplabel);

                JPanel xdatacolumnpanel = new JPanel(new SpringLayout());
                xdatacolumndefault = new JComboBox(columns);
                xdatacolumndefault.setSelectedItem(xlabeldefault);
                xdatacolumndefault.setPrototypeDisplayValue("123456789012345");
                xdatacolumndefault.addActionListener(new ActionListener(){
                    public void actionPerformed(ActionEvent ae) {
                        xlabeldefault = (String) xdatacolumndefault.getSelectedItem();
                        updatePlot();}});
                        
                xdatacolumnpanel.add (xdatacolumndefault);
                xdatacolumnpanel.setPreferredSize(new Dimension(125, 25));
                SpringUtil.makeCompactGrid (xdatacolumnpanel, 1, 1, 0,0,0,0);
                mainpanelgroupa.add (xdatacolumnpanel);

                JPanel moreybuttonpanel = new JPanel(new SpringLayout());
                moreybutton = new JButton("     Open     ");
                moreybutton.setPreferredSize(new Dimension(150, 20));
                moreybutton.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                        if (subFrame.isShowing())
                            subFrame.setVisible(true);
                        else
                            {
                            setSubpanelVisible = true;
                            showMoreCurvesJPanel();
                            }
                    }
                });

                moreybuttonpanel.add (moreybutton);
                moreybuttonpanel.setPreferredSize(new Dimension(125, 25));
                SpringUtil.makeCompactGrid (moreybuttonpanel, 1, 1, 0,0,0,0);
                mainpanelgroupa.add (moreybuttonpanel);

                JPanel maxcurvespanel = new JPanel(new SpringLayout());
                maxcurvespanel.setBorder(BorderFactory.createLineBorder(Color.lightGray,1));
                maxcurvesmodel = new SpinnerNumberModel(new Integer(maxCurves),new Integer(1), null, new Integer(1));
                maxcurvesspinner = new JSpinner(maxcurvesmodel);
                maxcurvesspinner.setPreferredSize(new Dimension(60,25));
                maxcurvesspinner.addChangeListener (new ChangeListener()
                    {
                    public void stateChanged(ChangeEvent ev)
                        {
                        maxCurves = ((Integer)maxcurvesspinner.getValue()).intValue();
                        Prefs.set("plot.maxCurves", maxCurves);
                        subframeWasShowing = false;
                        if (subFrame.isShowing())
                            {
                            subFrameLocationX = subFrame.getLocation().x;
                            subFrameLocationY = subFrame.getLocation().y;
                            Prefs.set("plot2.subFrameLocationX",subFrameLocationX);
                            Prefs.set("plot2.subFrameLocationY",subFrameLocationY);
                            subFrame.dispose();
                            subframeWasShowing = true;
                            }
                        setupArrays();
                        getPreferences();
                        if (subframeWasShowing)
                                showMoreCurvesJPanel();
                        updatePlot();
                        } });
                maxcurvesspinner.addMouseWheelListener( new MouseWheelListener()
                    {
                    public void mouseWheelMoved( MouseWheelEvent e )
                        {
                        int newValue =  new Integer(((Integer)maxcurvesspinner.getValue()).intValue() - e.getWheelRotation());
                        if (newValue > 0) maxcurvesspinner.setValue(newValue);
                        }
                    });
                maxcurvespanel.add (maxcurvesspinner);

                JLabel maxcurveslabel2 = new JLabel("sets");
                maxcurveslabel2.setHorizontalAlignment(JLabel.LEFT);
                maxcurvespanel.add (maxcurveslabel2);
                SpringUtil.makeCompactGrid (maxcurvespanel, 1, 2, 0,0,0,0);
                mainpanelgroupa.add(maxcurvespanel);

                JPanel mmagsubpanel = new JPanel (new SpringLayout());
                mmagsubpanel.setBorder(BorderFactory.createLineBorder(Color.lightGray,1));
                mmagsubpanel.setToolTipText("The number of starting samples averaged to calculate relative magnitude reference level");
                mmagrefsmodel = new SpinnerNumberModel(new Integer(mmagrefs), new Integer (1), null, new Integer(1));
                mmagrefsspinner = new JSpinner(mmagrefsmodel);
                mmagrefsspinner.setPreferredSize(new Dimension(60, 25));
                mmagrefsspinner.setToolTipText("The number of starting samples averaged to calculate relative magnitude reference level");
                mmagrefsspinner.addChangeListener (new ChangeListener()
                    {
                    public void stateChanged(ChangeEvent ev)
                        {
                        mmagrefs = ((Integer)mmagrefsspinner.getValue()).intValue();
                        updatePlot();
                        }
                    });
                mmagrefsspinner.addMouseWheelListener( new MouseWheelListener()
                    {
                    public void mouseWheelMoved( MouseWheelEvent e )
                        {
                        int newValue =  new Integer(((Integer)mmagrefsspinner.getValue()).intValue() - e.getWheelRotation());
                        if (newValue > 0) mmagrefsspinner.setValue(newValue);
                        }
                    });
                mmagsubpanel.add (mmagrefsspinner);

                mmagrefsspinnerlabel = new JLabel ("samples");
                mmagrefsspinnerlabel.setHorizontalAlignment(JLabel.LEFT);
                mmagsubpanel.add (mmagrefsspinnerlabel);
                SpringUtil.makeCompactGrid (mmagsubpanel, 1, 2, 0,0,0,0);
                mainpanelgroupa.add (mmagsubpanel);

                //VERTICAL MARKER 1 TIME STEPSIZE POPUP

                vmarker1steppopup = new JPopupMenu();
                JPanel vmarker1steppanel = new JPanel();
                vmarker1stepmodel = new SpinnerListModel(spinnerscalelist);
                vmarker1stepspinner = new JSpinner(vmarker1stepmodel);
                vmarker1stepspinner.setValue(convertToText(vMarker1Step));
                vmarker1stepspinner.addChangeListener (new ChangeListener()
                    {
                    public void stateChanged(ChangeEvent ev)
                        {
                        vMarker1Step = Double.parseDouble((String)vmarker1stepspinner.getValue());
                        vmarker1spinner.setModel(new SpinnerNumberModel(new Double(vMarker1Value), null, null, new Double(vMarker1Step)));
                        Prefs.set("plot.vmarker1Step",vMarker1Step);
                        }
                    });

                JLabel vmarker1steplabel = new JLabel ("Stepsize:");
                vmarker1steppanel.add(vmarker1steplabel);
                vmarker1steppanel.add(vmarker1stepspinner);
                vmarker1steppopup.add(vmarker1steppanel);
                vmarker1steppopup.setLightWeightPopupEnabled(false);
                vmarker1steppopup.addPopupMenuListener (new PopupMenuListener()
                    {
                    public void popupMenuCanceled(PopupMenuEvent popupMenuEvent) {
                            vmarker1steppopup.setVisible(false);
                            }

                    public void popupMenuWillBecomeInvisible(PopupMenuEvent popupMenuEvent) {
                            vmarker1steppopup.setVisible(true);
                            }

                    public void popupMenuWillBecomeVisible(PopupMenuEvent popupMenuEvent) {
                            vmarker1steppopup.setVisible(false);
                            }
                    });

                JPanel vmarker1panel = new JPanel(new SpringLayout());
                vmarker1panel.setBorder(BorderFactory.createLineBorder(Color.lightGray,1));
                showVMarker1CB = new JCheckBox ("", showVMarker1);
                showVMarker1CB.setToolTipText("Select to show vertical marker 1");
                showVMarker1CB.addItemListener(new ItemListener(){
                    public void itemStateChanged(ItemEvent e) {
                        if (e.getStateChange() == ItemEvent.DESELECTED)
                                showVMarker1 = false;
                        else if (e.getStateChange() == ItemEvent.SELECTED)
                                showVMarker1 = true;
                        updatePlot();}});
                showVMarker1CB.setHorizontalAlignment(JLabel.CENTER);
                vmarker1panel.add (showVMarker1CB);

                vmarker1spinnermodel = new SpinnerNumberModel(new Double(vMarker1Value),  null, null, new Double(vMarker1Step));

                vmarker1spinner = new JSpinner(vmarker1spinnermodel);
//                JComponent editor = new JSpinner.NumberEditor(vmarker1spinner);
//                vmarker1spinner.setEditor(editor);
                vmarker1spinner.setPreferredSize(new Dimension(75, 25));
                vmarker1spinner.setComponentPopupMenu(vmarker1steppopup);
                vmarker1spinner.setToolTipText("<html>"+"Enter vertical marker 1 x-axis location"+"<br>"+
                    "or enter UT time in HH:MM or HH:MM:SS format and press 'Enter'"+"<br>"+
                    "---------------------------------------------"+"<br>"+
                    "Right click to set spinner stepsize"+"</html>");
                vmarker1spinner.addChangeListener (new ChangeListener()
                    {
                    public void stateChanged(ChangeEvent ev)
                        {
                        checkForUT(vmarker1spinner);
                        vMarker1Value = ((Double)vmarker1spinner.getValue()).doubleValue();
                        updatePlot();
                        }
                    });
                vmarker1spinner.addMouseWheelListener( new MouseWheelListener()
                    {
                    public void mouseWheelMoved( MouseWheelEvent e )
                        {
                        vmarker1spinner.setValue(((Double)vmarker1spinner.getValue()).doubleValue()
                                - e.getWheelRotation()*vMarker1Step );
                        }
                    });
                vmarker1panel.add (vmarker1spinner);

                SpringUtil.makeCompactGrid (vmarker1panel, 1, vmarker1panel.getComponentCount(), 2,2,0,0);
                mainpanelgroupa.add (vmarker1panel);



                //VMARKER2 TIME STEPSIZE POPUP

                vmarker2steppopup = new JPopupMenu();
                JPanel vmarker2steppanel = new JPanel();
                vmarker2stepmodel = new SpinnerListModel(spinnerscalelist);
                vmarker2stepspinner = new JSpinner(vmarker2stepmodel);
                vmarker2stepspinner.setValue(convertToText(vMarker2Step));
                vmarker2stepspinner.addChangeListener (new ChangeListener()
                    {
                    public void stateChanged(ChangeEvent ev)
                        {
                        vMarker2Step = Double.parseDouble((String)vmarker2stepspinner.getValue());
                        vmarker2spinner.setModel(new SpinnerNumberModel(new Double(vMarker2Value), null, null, new Double(vMarker2Step)));
                        Prefs.set("plot.vMarker2Step",vMarker2Step);
                        }
                    });

                JLabel vmarker2steplabel = new JLabel ("Stepsize:");
                vmarker2steppanel.add(vmarker2steplabel);
                vmarker2steppanel.add(vmarker2stepspinner);
                vmarker2steppopup.add(vmarker2steppanel);
                vmarker2steppopup.setLightWeightPopupEnabled(false);
                vmarker2steppopup.addPopupMenuListener (new PopupMenuListener()
                    {
                    public void popupMenuCanceled(PopupMenuEvent popupMenuEvent) {
                            vmarker2steppopup.setVisible(false);
                            }

                    public void popupMenuWillBecomeInvisible(PopupMenuEvent popupMenuEvent) {
                            vmarker2steppopup.setVisible(true);
                            }

                    public void popupMenuWillBecomeVisible(PopupMenuEvent popupMenuEvent) {
                            vmarker2steppopup.setVisible(false);
                            }
                    });

                JPanel vmarker2panel = new JPanel(new SpringLayout());
                vmarker2panel.setBorder(BorderFactory.createLineBorder(Color.lightGray,1));
                showVMarker2CB = new JCheckBox ("", showVMarker2);
                showVMarker2CB.setToolTipText("Select to show vertical marker 2");
                showVMarker2CB.addItemListener(new ItemListener(){
                    public void itemStateChanged(ItemEvent e) {
                        if (e.getStateChange() == ItemEvent.DESELECTED)
                                showVMarker2 = false;
                        else if (e.getStateChange() == ItemEvent.SELECTED)
                                showVMarker2 = true;
                        updatePlot();}});
                showVMarker2CB.setHorizontalAlignment(JLabel.CENTER);
                vmarker2panel.add (showVMarker2CB);

                vmarker2spinnermodel = new SpinnerNumberModel(new Double(vMarker2Value),  null, null, new Double(vMarker2Step));
                vmarker2spinner = new JSpinner(vmarker2spinnermodel);
                vmarker2spinner.setPreferredSize(new Dimension(75, 25));
                vmarker2spinner.setComponentPopupMenu(vmarker2steppopup);
                vmarker2spinner.setToolTipText("<html>"+"Enter vertical Marker 2 x-axis location"+"<br>"+
                    "or enter UT time in HH:MM or HH:MM:SS format and press 'Enter'"+"<br>"+
                    "---------------------------------------------"+"<br>"+
                    "Right click to set spinner stepsize"+"</html>");
                vmarker2spinner.addChangeListener (new ChangeListener()
                    {
                    public void stateChanged(ChangeEvent ev)
                        {
                        checkForUT(vmarker2spinner);
                        vMarker2Value = ((Double)vmarker2spinner.getValue()).doubleValue();
                        updatePlot();
                        }
                    });
                vmarker2spinner.addMouseWheelListener( new MouseWheelListener()
                    {
                    public void mouseWheelMoved( MouseWheelEvent e )
                        {
                        vmarker2spinner.setValue(((Double)vmarker2spinner.getValue()).doubleValue()
                                - e.getWheelRotation()*vMarker2Step );
                        }
                    });
                vmarker2panel.add (vmarker2spinner);

                SpringUtil.makeCompactGrid (vmarker2panel, 1, vmarker2panel.getComponentCount(), 2,2,0,0);
                mainpanelgroupa.add (vmarker2panel);

                SpringUtil.makeCompactGrid (mainpanelgroupa, 2, mainpanelgroupa.getComponentCount()/2, 5,5,5,5);
                mainpanel.add(mainpanelgroupa);

    // MAIN PANEL GROUP C START

                // TITLE GROUP
                        JPanel titlegroup = new JPanel (new SpringLayout());
                        titlegroup.setBorder (BorderFactory.createEmptyBorder(10,10,10,10));
                        titlegroup.setBorder(BorderFactory.createTitledBorder("Title"));

                                // TITLE RADIO BUTTON GROUP
                                JPanel titleradiopanelgroup = new JPanel (new SpringLayout());
                                        JRadioButton noTitleButton = new JRadioButton("None");
                                        if (!useTitle) noTitleButton.setSelected(true);
                                        else noTitleButton.setSelected(false);
                                titleradiopanelgroup.add (noTitleButton);
                                        JRadioButton useTitleButton = new JRadioButton("Custom");
                                        if (useTitle) useTitleButton.setSelected(true);
                                        else useTitleButton.setSelected(false);
                                titleradiopanelgroup.add (useTitleButton);
                                        ButtonGroup titleRadioGroup = new ButtonGroup();
                                        titleRadioGroup.add(noTitleButton);
                                        titleRadioGroup.add(useTitleButton);
                                        noTitleButton.addActionListener(new ActionListener(){
                                            public void actionPerformed(ActionEvent ae) {
                                                useTitle = false;
                                                updatePlot();}});
                                        useTitleButton.addActionListener(new ActionListener(){
                                            public void actionPerformed(ActionEvent ae) {
                                                useTitle =  true;
                                                updatePlot();}});
                                SpringUtil.makeCompactGrid (titleradiopanelgroup, 1, 2, 0,0,0,0);
                        titlegroup.add (titleradiopanelgroup);

                                // TITLE TEXT FIELD
                                titleField = new JTextField (title);
                                titleField.setBorder (BorderFactory.createEmptyBorder(0,4,0,4));
                                titleField.getDocument().addDocumentListener (new DocumentListener()
                                    {
                                    public void insertUpdate (DocumentEvent ev)
                                        {
                                        title = titleField.getText();
                                        updatePlot();
                                        }
                                    public void removeUpdate (DocumentEvent ev)
                                        {
                                        title = titleField.getText();
                                        updatePlot();
                                        }
                                    public void changedUpdate (DocumentEvent ev)
                                        {
                                        title = titleField.getText();
                                        updatePlot();
                                        }
                                    });
                                titleField.setPreferredSize (new Dimension(250,20));
                                titleField.setHorizontalAlignment (JTextField.LEFT);
                        titlegroup.add (titleField);

                                //MAKE TITLE Y-POSITION SLIDER
                                
                                titlePosYSlider = new JSlider(JSlider.HORIZONTAL, 0,1000, (int)(titlePosY));
                                titlePosYSlider.setBorder (BorderFactory.createEmptyBorder(10,0,0,0));
                                titlePosYSlider.addChangeListener (new ChangeListener()
                                    {
                                    public void stateChanged(ChangeEvent ev)
                                        {
                                        JSlider slider=(JSlider)ev.getSource();
                                        titlePosY=slider.getValue();
                                        updatePlot();
                                        }
                                    });
                                titlePosYSlider.addMouseWheelListener( new MouseWheelListener()
                                    {
                                    public void mouseWheelMoved( MouseWheelEvent e )
                                        {
                                        int newValue =  new Integer(((Integer)titlePosYSlider.getValue()).intValue() + e.getWheelRotation());
                                        if ((newValue >= 0) && (newValue <=1000)) titlePosYSlider.setValue(newValue);
                                        }
                                    });
                                titlePosYSlider.setPreferredSize (new Dimension(250,50));
                                titlePosYSlider.setInverted(false);
                                titlePosYSlider.setMajorTickSpacing(500);
                                titlePosYSlider.setMinorTickSpacing(50);
                                titlePosYSlider.setPaintTicks(true);
                                Hashtable<Integer, JLabel> titlePosYLabel = new Hashtable<Integer, JLabel>();
                                titlePosYLabel.put(0, new JLabel("Top"));
                                titlePosYLabel.put(500, new JLabel("Middle") );
                                titlePosYLabel.put(1000, new JLabel("Bottom") );
                                titlePosYSlider.setLabelTable( titlePosYLabel );
                                titlePosYSlider.setPaintTrack(true);
                                titlePosYSlider.setPaintLabels(true);
                        titlegroup.add (titlePosYSlider);

                                //MAKE TITLE X-POSITION SLIDER
                                
                                titlePosXSlider = new JSlider(0,500, (int) (titlePosX*500.0));
                                titlePosXSlider.setBorder (BorderFactory.createEmptyBorder(10,5,0,5));
                                titlePosXSlider.addChangeListener (new ChangeListener()
                                    {
                                    public void stateChanged(ChangeEvent ev)
                                        {
                                        JSlider slider=(JSlider)ev.getSource();
                                        titlePosX=slider.getValue()/500.0;
                                        updatePlot();
                                        }
                                    });
                                titlePosXSlider.addMouseWheelListener( new MouseWheelListener()
                                    {
                                    public void mouseWheelMoved( MouseWheelEvent e )
                                        {
                                        int newValue =  new Integer(((Integer)titlePosXSlider.getValue()).intValue() + e.getWheelRotation());
                                        if ((newValue >= 0) && (newValue <=500)) titlePosXSlider.setValue(newValue);
                                        }
                                    });
                                titlePosXSlider.setPreferredSize (new Dimension(250,50));
                                titlePosXSlider.setMajorTickSpacing(250);
                                titlePosXSlider.setMinorTickSpacing(25);
                                titlePosXSlider.setPaintTicks(true);
                                Hashtable<Integer, JLabel> titlePosXLabel = new Hashtable<Integer, JLabel>();
                                titlePosXLabel.put( 0, new JLabel("Left"));
                                titlePosXLabel.put( 250, new JLabel("Center") );
                                titlePosXLabel.put( 500, new JLabel("Right") );
                                titlePosXSlider.setLabelTable( titlePosXLabel );
                                titlePosXSlider.setPaintTrack(true);
                                titlePosXSlider.setPaintLabels(true);
                        titlegroup.add (titlePosXSlider);
                        SpringUtil.makeCompactGrid (titlegroup, 4, 1, 0,0,0,0);
		mainpanelgroupc.add (titlegroup);

                        //SUBTITLE GROUP
                        JPanel subtitlegroup = new JPanel (new SpringLayout());
                        subtitlegroup.setBorder(BorderFactory.createTitledBorder("Subtitle"));

                                // SUBTITLE RADIO BUTTON GROUP
                                JPanel subtitleradiopanelgroup = new JPanel (new SpringLayout());
                                        JRadioButton noSubtitleButton = new JRadioButton("None");
                                        if (!useSubtitle) noSubtitleButton.setSelected(true);
                                        else noSubtitleButton.setSelected(false);
                                subtitleradiopanelgroup.add (noSubtitleButton);
                                        JRadioButton useSubtitleButton = new JRadioButton("Custom");
                                        if (useSubtitle) useSubtitleButton.setSelected(true);
                                        else useSubtitleButton.setSelected(false);
                                subtitleradiopanelgroup.add (useSubtitleButton);
                                        ButtonGroup subtitleRadioGroup = new ButtonGroup();
                                        subtitleRadioGroup.add(noSubtitleButton);
                                        subtitleRadioGroup.add(useSubtitleButton);
                                        noSubtitleButton.addActionListener(new ActionListener(){
                                            public void actionPerformed(ActionEvent ae) {
                                                useSubtitle = false;
                                                updatePlot();}});
                                        useSubtitleButton.addActionListener(new ActionListener(){
                                            public void actionPerformed(ActionEvent ae) {
                                                useSubtitle =  true;
                                                updatePlot();}});
                                SpringUtil.makeCompactGrid (subtitleradiopanelgroup, 1, 2, 0,0,0,0);
                        subtitlegroup.add (subtitleradiopanelgroup);

                                // SUBTITLE TEXT FIELD
                                subtitleField = new JTextField (subtitle);
                                subtitleField.setBorder (BorderFactory.createEmptyBorder(0,4,0,4));
                                subtitleField.getDocument().addDocumentListener (new DocumentListener()
                                    {
                                    public void insertUpdate (DocumentEvent ev)
                                        {
                                        subtitle = subtitleField.getText();
                                        updatePlot();
                                        }
                                    public void removeUpdate (DocumentEvent ev)
                                        {
                                        subtitle = subtitleField.getText();
                                        updatePlot();
                                        }
                                    public void changedUpdate (DocumentEvent ev)
                                        {
                                        subtitle = subtitleField.getText();
                                        updatePlot();
                                        }
                                    });
                                subtitleField.setPreferredSize (new Dimension(250,20));
                                subtitleField.setHorizontalAlignment (JTextField.LEFT);

                        subtitlegroup.add (subtitleField);

                                //SUBTITLE Y-POSITION SLIDER
                                
                                subtitlePosYSlider = new JSlider(JSlider.HORIZONTAL, 0,1000, (int)(subtitlePosY));
                                subtitlePosYSlider.setBorder (BorderFactory.createEmptyBorder(10,0,0,0));
                                subtitlePosYSlider.addChangeListener (new ChangeListener()
                                    {
                                    public void stateChanged(ChangeEvent ev)
                                        {
                                        JSlider slider=(JSlider)ev.getSource();
                                        subtitlePosY=slider.getValue();
                                        updatePlot();
                                        }
                                    });
                                subtitlePosYSlider.addMouseWheelListener( new MouseWheelListener()
                                    {
                                    public void mouseWheelMoved( MouseWheelEvent e )
                                        {
                                        int newValue =  new Integer(((Integer)subtitlePosYSlider.getValue()).intValue() + e.getWheelRotation());
                                        if ((newValue >= 0) && (newValue <=1000)) subtitlePosYSlider.setValue(newValue);
                                        }
                                    });
                                subtitlePosYSlider.setPreferredSize (new Dimension(250,50));
                                subtitlePosYSlider.setInverted(false);
                                subtitlePosYSlider.setMajorTickSpacing(500);
                                subtitlePosYSlider.setMinorTickSpacing(50);
                                subtitlePosYSlider.setPaintTicks(true);
                                Hashtable<Integer, JLabel> subtitlePosYLabel = new Hashtable<Integer, JLabel>();

                                subtitlePosYLabel.put(0, new JLabel("Top"));
                                subtitlePosYLabel.put(500, new JLabel("Middle") );
                                subtitlePosYLabel.put(1000, new JLabel("Bottom") );
                                subtitlePosYSlider.setLabelTable( titlePosYLabel );
                                subtitlePosYSlider.setPaintTrack(true);
                                subtitlePosYSlider.setPaintLabels(true);
                        subtitlegroup.add (subtitlePosYSlider);

                                //SUBTITLE X-POSITION SLIDER
                                
                                subtitlePosXSlider = new JSlider(0,500, (int) (subtitlePosX*500.0));
                                subtitlePosXSlider.setBorder (BorderFactory.createEmptyBorder(10,5,0,5));
                                subtitlePosXSlider.addChangeListener (new ChangeListener()
                                    {
                                    public void stateChanged(ChangeEvent ev)
                                        {
                                        JSlider slider=(JSlider)ev.getSource();
                                        subtitlePosX=slider.getValue()/500.0;
                                        updatePlot();
                                        }
                                    });
                                subtitlePosXSlider.addMouseWheelListener( new MouseWheelListener()
                                    {
                                    public void mouseWheelMoved( MouseWheelEvent e )
                                        {
                                        int newValue =  new Integer(((Integer)subtitlePosXSlider.getValue()).intValue() + e.getWheelRotation());
                                        if ((newValue >= 0) && (newValue <=500)) subtitlePosXSlider.setValue(newValue);
                                        }
                                    });
                                subtitlePosXSlider.setPreferredSize (new Dimension(250,50));
                                subtitlePosXSlider.setMajorTickSpacing(250);
                                subtitlePosXSlider.setMinorTickSpacing(25);
                                subtitlePosXSlider.setPaintTicks(true);
                                Hashtable<Integer, JLabel> subtitlePosXLabel = new Hashtable<Integer, JLabel>();
                                subtitlePosXLabel.put( 0, new JLabel("Left"));
                                subtitlePosXLabel.put( 250, new JLabel("Center") );
                                subtitlePosXLabel.put( 500, new JLabel("Right") );
                                subtitlePosXSlider.setLabelTable( titlePosXLabel );
                                subtitlePosXSlider.setPaintTrack(true);
                                subtitlePosXSlider.setPaintLabels(true);
                        subtitlegroup.add (subtitlePosXSlider);
                        SpringUtil.makeCompactGrid (subtitlegroup, 4, 1, 0,0,0,0);
                mainpanelgroupc.add (subtitlegroup);


                        // LEGEND GROUP
                        legendgroup = new JPanel (new SpringLayout());
                        legendgroup.setBorder(BorderFactory.createTitledBorder("Legend"));

                        // LEGEND LABEL POPUP

                        legendpopup = new JPopupMenu();
                        JMenuItem legendshowscalingCB = new JCheckBoxMenuItem("Append scaling factor to legend",showLScaleInfo);
                        legendshowscalingCB.addItemListener(new ItemListener(){
                            public void itemStateChanged(ItemEvent e) {
                                if (e.getStateChange() == ItemEvent.DESELECTED)
                                        showLScaleInfo = false;
                                else if (e.getStateChange() == ItemEvent.SELECTED)
                                        showLScaleInfo = true;
                                updatePlot();}});
                        legendpopup.add(legendshowscalingCB);
                        JMenuItem legendshowshiftCB = new JCheckBoxMenuItem("Append shift factor to legend",showLShiftInfo);
                        legendshowshiftCB.addItemListener(new ItemListener(){
                            public void itemStateChanged(ItemEvent e) {
                                if (e.getStateChange() == ItemEvent.DESELECTED)
                                        showLShiftInfo = false;
                                else if (e.getStateChange() == ItemEvent.SELECTED)
                                        showLShiftInfo = true;
                                updatePlot();}});
                        legendpopup.add(legendshowshiftCB);
                        JMenuItem legendshowbinCB = new JCheckBoxMenuItem("Append bin size to legend",showLBinInfo);
                        legendshowbinCB.addItemListener(new ItemListener(){
                            public void itemStateChanged(ItemEvent e) {
                                if (e.getStateChange() == ItemEvent.DESELECTED)
                                        showLBinInfo = false;
                                else if (e.getStateChange() == ItemEvent.SELECTED)
                                        showLBinInfo = true;
                                updatePlot();}});
                        legendpopup.add(legendshowbinCB);
                        JMenuItem legendshowmmagCB = new JCheckBoxMenuItem("Append magnitude usage to legend",showLmmagInfo);
                        legendshowmmagCB.addItemListener(new ItemListener(){
                            public void itemStateChanged(ItemEvent e) {
                                if (e.getStateChange() == ItemEvent.DESELECTED)
                                        showLmmagInfo = false;
                                else if (e.getStateChange() == ItemEvent.SELECTED)
                                        showLmmagInfo = true;
                                updatePlot();}});
                        legendpopup.add(legendshowmmagCB);
                        JMenuItem legendshowsymbolCB = new JCheckBoxMenuItem("Append symbol description to legend",showLSymbolInfo);
                        legendshowsymbolCB.addItemListener(new ItemListener(){
                            public void itemStateChanged(ItemEvent e) {
                                if (e.getStateChange() == ItemEvent.DESELECTED)
                                        showLSymbolInfo = false;
                                else if (e.getStateChange() == ItemEvent.SELECTED)
                                        showLSymbolInfo = true;
                                updatePlot();}});
                        legendpopup.add(legendshowsymbolCB);
                        JMenuItem twolinelegendCB = new JCheckBoxMenuItem("Use two lines per legend",useTwoLineLegend);
                        twolinelegendCB.addItemListener(new ItemListener(){
                            public void itemStateChanged(ItemEvent e) {
                                if (e.getStateChange() == ItemEvent.DESELECTED)
                                        useTwoLineLegend = false;
                                else if (e.getStateChange() == ItemEvent.SELECTED)
                                        useTwoLineLegend = true;
                                updatePlot();}});
                        legendpopup.add(twolinelegendCB);


                        legendpopup.addPopupMenuListener (new PopupMenuListener() {
                                public void popupMenuCanceled(PopupMenuEvent popupMenuEvent) { }
                                public void popupMenuWillBecomeInvisible(PopupMenuEvent popupMenuEvent) { }
                                public void popupMenuWillBecomeVisible(PopupMenuEvent popupMenuEvent) { } });


                                // LEGEND RADIO BUTTON GROUP
                                JPanel legendradiopanelgroup = new JPanel ();
                                legendradiopanelgroup.setLayout(new BoxLayout(legendradiopanelgroup,BoxLayout.X_AXIS));
                                JLabel legendradiolabel = new JLabel("Align:");
                                legendradiolabel.setToolTipText("Right click to set legend preferences");
                                legendradiopanelgroup.add (legendradiolabel);
                                legendradiopanelgroup.add(Box.createHorizontalStrut(2));
                                        JRadioButton leftButton = new JRadioButton("Left");
                                        if (legendLeft)
                                                leftButton.setSelected(true);
                                        else
                                                leftButton.setSelected(false);
                                legendradiopanelgroup.add (leftButton);
                                legendradiopanelgroup.add(Box.createHorizontalStrut(2));
                                        JRadioButton centerButton = new JRadioButton("Center");
                                        if (!legendLeft && !legendRight)
                                                centerButton.setSelected(true);
                                        else
                                                centerButton.setSelected(false);
                                legendradiopanelgroup.add (centerButton);
                                legendradiopanelgroup.add(Box.createHorizontalStrut(2));
                                        JRadioButton rightButton = new JRadioButton("Right");
                                        if (legendRight)
                                                rightButton.setSelected(true);
                                        else
                                                rightButton.setSelected(false);
                                legendradiopanelgroup.add (rightButton);
                                legendradiopanelgroup.add(Box.createHorizontalStrut(2));
                                legendradiopanelgroup.add(Box.createHorizontalGlue());

                                        ButtonGroup legendRadioGroup = new ButtonGroup();
                                        legendRadioGroup.add(leftButton);
                                        legendRadioGroup.add(centerButton);
                                        legendRadioGroup.add(rightButton);
                                        leftButton.addActionListener(new ActionListener(){
                                            public void actionPerformed(ActionEvent ae) {
                                                legendLeft = true;
                                                legendRight = false;
                                                updatePlot();}});
                                        centerButton.addActionListener(new ActionListener(){
                                            public void actionPerformed(ActionEvent ae) {
                                                legendLeft = false;
                                                legendRight = false;
                                                updatePlot();}});
                                        rightButton.addActionListener(new ActionListener(){
                                            public void actionPerformed(ActionEvent ae) {
                                                legendLeft = false;
                                                legendRight = true;
                                                updatePlot();}});
                                configureIcon = createImageIcon("astroj/images/configure.png", "Configure legend options");
                                
                                legendconfigureButton = new JButton(configureIcon);
                                legendconfigureButton.setToolTipText("Configure legend options");
                                legendconfigureButton.setMargin(new Insets(0,0,0,0));
                                
                                legendconfigureButton.addActionListener(new ActionListener(){
                                            public void actionPerformed(ActionEvent ae) {
                                                legendpopup.show(legendconfigureButton, mainpanel.getX() , mainpanel.getY()+25);}});
                                legendradiopanelgroup.add(legendconfigureButton);
                                legendradiopanelgroup.add(Box.createHorizontalStrut(2));
//                                SpringUtil.makeCompactGrid (legendradiopanelgroup, 1, legendradiopanelgroup.getComponentCount(), 1,1,1,1);
                                leftButton.setComponentPopupMenu(legendpopup);
                                leftButton.setToolTipText("Right click to set legend preferences");
                                rightButton.setComponentPopupMenu(legendpopup);
                                rightButton.setToolTipText("Right click to set legend preferences");
                                centerButton.setComponentPopupMenu(legendpopup);
                                centerButton.setToolTipText("Right click to set legend preferences");
                        legendgroup.add (legendradiopanelgroup);

                                // LEGEND SLIDER LABEL
                                JLabel legendsliderlabel = new JLabel("Position");
                                legendsliderlabel.setBorder (BorderFactory.createEmptyBorder(5,0,0,0));
                                legendsliderlabel.setToolTipText("Right click to set legend preferences");
                                legendsliderlabel.setHorizontalAlignment (JLabel.CENTER);
                                legendsliderlabel.setComponentPopupMenu(legendpopup);

                        legendgroup.add (legendsliderlabel);

                                //LEGEND Y-POSITION SLIDER
                                
                                legendPosYSlider = new JSlider(JSlider.HORIZONTAL, 0,500, (int) (legendPosY*500.0));
                                legendPosYSlider.setBorder (BorderFactory.createEmptyBorder(10,0,0,0));
                                legendPosYSlider.addChangeListener (new ChangeListener()
                                    {
                                    public void stateChanged(ChangeEvent ev)
                                        {
                                        JSlider slider=(JSlider)ev.getSource();
                                        legendPosY=slider.getValue()/500.0;
                                        updatePlot();
                                        }
                                    });
                                legendPosYSlider.addMouseWheelListener( new MouseWheelListener()
                                    {
                                    public void mouseWheelMoved( MouseWheelEvent e )
                                        {
                                        int newValue =  new Integer(((Integer)legendPosYSlider.getValue()).intValue() + e.getWheelRotation());
                                        if ((newValue >= 0) && (newValue <=500)) legendPosYSlider.setValue(newValue);
                                        }
                                    });
                                legendPosYSlider.setPreferredSize (new Dimension(250,50));
                                legendPosYSlider.setInverted(false);
                                legendPosYSlider.setMajorTickSpacing(250);
                                legendPosYSlider.setMinorTickSpacing(25);
                                legendPosYSlider.setPaintTicks(true);
                                Hashtable<Integer, JLabel> legendPosYLabel = new Hashtable<Integer, JLabel>();
                                legendPosYLabel.put( 0, new JLabel("Top"));
                                legendPosYLabel.put( 250, new JLabel("Middle") );
                                legendPosYLabel.put( 500, new JLabel("Bottom") );
                                legendPosYSlider.setLabelTable( titlePosYLabel );
                                legendPosYSlider.setPaintTrack(true);
                                legendPosYSlider.setPaintLabels(true);
                        legendgroup.add (legendPosYSlider);

                                //LEGEND X-POSITION SLIDER
                                
                                legendPosXSlider = new JSlider(0,500, (int) (legendPosX*500.0));
                                legendPosXSlider.setBorder (BorderFactory.createEmptyBorder(10,5,0,5));
                                legendPosXSlider.addChangeListener (new ChangeListener()
                                    {
                                    public void stateChanged(ChangeEvent ev)
                                        {
                                        JSlider slider=(JSlider)ev.getSource();
                                        legendPosX=slider.getValue()/500.0;
                                        updatePlot();
                                        }
                                    });
                                legendPosXSlider.addMouseWheelListener( new MouseWheelListener()
                                    {
                                    public void mouseWheelMoved( MouseWheelEvent e )
                                        {
                                        int newValue =  new Integer(((Integer)legendPosXSlider.getValue()).intValue() + e.getWheelRotation());
                                        if ((newValue >= 0) && (newValue <=500)) legendPosXSlider.setValue(newValue);
                                        }
                                    });
                                legendPosXSlider.setPreferredSize (new Dimension(250,50));
                                legendPosXSlider.setMajorTickSpacing(250);
                                legendPosXSlider.setMinorTickSpacing(25);
                                legendPosXSlider.setPaintTicks(true);
                                Hashtable<Integer, JLabel> legendPosXLabel = new Hashtable<Integer, JLabel>();
                                legendPosXLabel.put( 0, new JLabel("Left"));
                                legendPosXLabel.put( 250, new JLabel("Center") );
                                legendPosXLabel.put( 500, new JLabel("Right") );
                                legendPosXSlider.setLabelTable( titlePosXLabel );
                                legendPosXSlider.setPaintTrack(true);
                                legendPosXSlider.setPaintLabels(true);
                        legendgroup.add (legendPosXSlider);
                        SpringUtil.makeCompactGrid (legendgroup, 4, 1, 0,0,0,0);

                mainpanelgroupc.add (legendgroup);

                SpringUtil.makeCompactGrid (mainpanelgroupc, 1, 3, 0,0,0,0);
                mainpanel.add(mainpanelgroupc);

    // X & Y AXIS LEGEND GROUP (GROUP D)

            //X-AXIS LEGEND GROUP
                        JPanel xlegendgroup = new JPanel (new SpringLayout());
                        xlegendgroup.setBorder(BorderFactory.createTitledBorder("X-Axis Label"));

                        // X-AXIS LABEL POPUP

                        xaxispopup = new JPopupMenu();
                        JMenuItem showxscalingCB = new JCheckBoxMenuItem("Append scaling factor to X-axis label",showXScaleInfo);

                        showxscalingCB.addItemListener(new ItemListener(){
                            public void itemStateChanged(ItemEvent e) {
                                if (e.getStateChange() == ItemEvent.DESELECTED)
                                        showXScaleInfo = false;
                                else if (e.getStateChange() == ItemEvent.SELECTED)
                                        showXScaleInfo = true;
                                updatePlot();}});
                        xaxispopup.add(showxscalingCB);
                        xaxispopup.addPopupMenuListener (new PopupMenuListener() {
                                public void popupMenuCanceled(PopupMenuEvent popupMenuEvent) { }
                                public void popupMenuWillBecomeInvisible(PopupMenuEvent popupMenuEvent) { }
                                public void popupMenuWillBecomeVisible(PopupMenuEvent popupMenuEvent) { } });

                        // X-AXIS LEGEND RADIO BUTTON GROUP
                                JPanel xlegendradiopanelgroup = new JPanel ();
                                xlegendradiopanelgroup.setLayout(new BoxLayout(xlegendradiopanelgroup,BoxLayout.X_AXIS));
                                JRadioButton xnoneButton = new JRadioButton("None");
                                if (!useXCustomName && !useXColumnName)
                                        xnoneButton.setSelected(true);
                                else
                                        xnoneButton.setSelected(false);
                                xlegendradiopanelgroup.add(Box.createHorizontalStrut(2));
                                xlegendradiopanelgroup.add(xnoneButton);
                                xlegendradiopanelgroup.add(Box.createHorizontalStrut(2));
                                JRadioButton xcolumnNameButton = new JRadioButton("Column Label");
                                if (useXColumnName)
                                        xcolumnNameButton.setSelected(true);
                                else
                                        xcolumnNameButton.setSelected(false);
                                xlegendradiopanelgroup.add(xcolumnNameButton);
                                xlegendradiopanelgroup.add(Box.createHorizontalStrut(2));
                                JRadioButton xcustomNameButton = new JRadioButton("Custom Label");
                                if (!useXColumnName && useXCustomName)
                                        xcustomNameButton.setSelected(true);
                                else
                                        xcustomNameButton.setSelected(false);
                                xlegendradiopanelgroup.add(xcustomNameButton);
                                xlegendradiopanelgroup.add(Box.createHorizontalStrut(2));
                                xlegendradiopanelgroup.add(Box.createGlue());
                                ButtonGroup xlegendRadioGroup = new ButtonGroup();
                                xlegendRadioGroup.add(xnoneButton);
                                xlegendRadioGroup.add(xcolumnNameButton);
                                xlegendRadioGroup.add(xcustomNameButton);
                                xnoneButton.addActionListener(new ActionListener(){
                                    public void actionPerformed(ActionEvent ae) {
                                        useXColumnName = false;
                                        useXCustomName = false;
                                        updatePlot();}});
                                xcustomNameButton.addActionListener(new ActionListener(){
                                    public void actionPerformed(ActionEvent ae) {
                                        useXColumnName = false;
                                        useXCustomName = true;
                                        updatePlot();}});
                                xcolumnNameButton.addActionListener(new ActionListener(){
                                    public void actionPerformed(ActionEvent ae) {
                                        useXColumnName = true;
                                        useXCustomName = false;
                                        updatePlot();}});
                                xlegendconfigureButton = new JButton(configureIcon);
                                xlegendconfigureButton.setToolTipText("Set X-axis label preferences");
                                xlegendconfigureButton.setMargin(new Insets(0,0,0,0));

                                xlegendconfigureButton.addActionListener(new ActionListener(){
                                            public void actionPerformed(ActionEvent ae) {
                                                xaxispopup.show(xlegendconfigureButton, mainpanel.getX() , mainpanel.getY()+25);}});
                                xlegendradiopanelgroup.add(xlegendconfigureButton);
                                xlegendradiopanelgroup.add(Box.createHorizontalStrut(2));
//                                SpringUtil.makeCompactGrid (xlegendradiopanelgroup, 1, xlegendradiopanelgroup.getComponentCount(), 1,1,1,1);
                                xnoneButton.setComponentPopupMenu(xaxispopup);
                                xnoneButton.setToolTipText("Right click to set X-axis label preferences");
                                xcustomNameButton.setComponentPopupMenu(xaxispopup);
                                xcustomNameButton.setToolTipText("Right click to set X-axis label preferences");
                                xcolumnNameButton.setComponentPopupMenu(xaxispopup);
                                xcolumnNameButton.setToolTipText("Right click to set X-axis label preferences");
                        xlegendgroup.add (xlegendradiopanelgroup);

                                // X LEGEND TEXT FIELD
                                xlegendField = new JTextField (xLegend);
                                xlegendField.setBorder (BorderFactory.createEmptyBorder(0,4,0,4));
                                xlegendField.getDocument().addDocumentListener (new DocumentListener()
                                    {
                                    public void insertUpdate (DocumentEvent ev)
                                        {
                                        xLegend = xlegendField.getText();
                                        updatePlot();
                                        }
                                    public void removeUpdate (DocumentEvent ev)
                                        {
                                        xLegend = xlegendField.getText();
                                        updatePlot();
                                        }
                                    public void changedUpdate (DocumentEvent ev)
                                        {
                                        xLegend = xlegendField.getText();
                                        updatePlot();
                                        }
                                    });
                                xlegendField.setPreferredSize (new Dimension(250,20));
                                xlegendField.setHorizontalAlignment (JTextField.LEFT);
                                xlegendField.setComponentPopupMenu(xaxispopup);
                                xlegendField.setToolTipText("Right click to set X-axis label preferences");

                        xlegendgroup.add (xlegendField);
                        SpringUtil.makeCompactGrid (xlegendgroup, 2, 1, 0,0,0,0);
                mainpanelgroupd.add (xlegendgroup);

            //Y-AXIS LEGEND GROUP
                        JPanel ylegendgroup = new JPanel (new SpringLayout());
                        ylegendgroup.setBorder(BorderFactory.createTitledBorder("Y-Axis Label"));

                        // Y-AXIS LABEL POPUP

                        yaxispopup = new JPopupMenu();
                        JMenuItem showyscalingCB = new JCheckBoxMenuItem("Append scaling factor to Y-axis label",showYScaleInfo);
                        showyscalingCB.addItemListener(new ItemListener(){
                            public void itemStateChanged(ItemEvent e) {
                                if (e.getStateChange() == ItemEvent.DESELECTED)
                                        showYScaleInfo = false;
                                else if (e.getStateChange() == ItemEvent.SELECTED)
                                        showYScaleInfo = true;
                                updatePlot();}});
                        yaxispopup.add(showyscalingCB);
                        JMenuItem showyshiftCB = new JCheckBoxMenuItem("Append shift factor to Y-axis label",showYShiftInfo);
                        showyshiftCB.addItemListener(new ItemListener(){
                            public void itemStateChanged(ItemEvent e) {
                                if (e.getStateChange() == ItemEvent.DESELECTED)
                                        showYShiftInfo = false;
                                else if (e.getStateChange() == ItemEvent.SELECTED)
                                        showYShiftInfo = true;
                                updatePlot();}});
                        yaxispopup.add(showyshiftCB);
                        JMenuItem showybinCB = new JCheckBoxMenuItem("Append bin size to Y-axis label",showYBinInfo);
                        showybinCB.addItemListener(new ItemListener(){
                            public void itemStateChanged(ItemEvent e) {
                                if (e.getStateChange() == ItemEvent.DESELECTED)
                                        showYBinInfo = false;
                                else if (e.getStateChange() == ItemEvent.SELECTED)
                                        showYBinInfo = true;
                                updatePlot();}});
                        yaxispopup.add(showybinCB);
                        JMenuItem showymmagCB = new JCheckBoxMenuItem("Append magnitude usage to Y-axis label",showYmmagInfo);
                        showymmagCB.addItemListener(new ItemListener(){
                            public void itemStateChanged(ItemEvent e) {
                                if (e.getStateChange() == ItemEvent.DESELECTED)
                                        showYmmagInfo = false;
                                else if (e.getStateChange() == ItemEvent.SELECTED)
                                        showYmmagInfo = true;
                                updatePlot();}});
                        yaxispopup.add(showymmagCB);
                        JMenuItem showysymbolCB = new JCheckBoxMenuItem("Append symbol description to Y-axis label",showYSymbolInfo);
                        showysymbolCB.addItemListener(new ItemListener(){
                            public void itemStateChanged(ItemEvent e) {
                                if (e.getStateChange() == ItemEvent.DESELECTED)
                                        showYSymbolInfo = false;
                                else if (e.getStateChange() == ItemEvent.SELECTED)
                                        showYSymbolInfo = true;
                                updatePlot();}});
                        yaxispopup.add(showysymbolCB);
                        yaxispopup.addPopupMenuListener (new PopupMenuListener() {
                                public void popupMenuCanceled(PopupMenuEvent popupMenuEvent) { }
                                public void popupMenuWillBecomeInvisible(PopupMenuEvent popupMenuEvent) { }
                                public void popupMenuWillBecomeVisible(PopupMenuEvent popupMenuEvent) { } });

                        // Y-AXIS LEGEND RADIO BUTTON GROUP
                                JPanel ylegendradiopanelgroup = new JPanel ();
                                ylegendradiopanelgroup.setLayout(new BoxLayout(ylegendradiopanelgroup,BoxLayout.X_AXIS));
                                        JRadioButton ynoneButton = new JRadioButton("None");
                                        if (!useYCustomName && !useYColumnName)
                                                ynoneButton.setSelected(true);
                                        else
                                                ynoneButton.setSelected(false);
                                ylegendradiopanelgroup.add(ynoneButton);
                                ylegendradiopanelgroup.add(Box.createHorizontalStrut(2));
                                        JRadioButton ycolumnNameButton = new JRadioButton("Column Label");
                                        if (useYColumnName)
                                                ycolumnNameButton.setSelected(true);
                                        else
                                                ycolumnNameButton.setSelected(false);
                                ylegendradiopanelgroup.add(ycolumnNameButton);
                                ylegendradiopanelgroup.add(Box.createHorizontalStrut(2));
                                        JRadioButton ycustomNameButton = new JRadioButton("Custom Label");
                                        if (!useYColumnName && useYCustomName)
                                                ycustomNameButton.setSelected(true);
                                        else
                                                ycustomNameButton.setSelected(false);
                                ylegendradiopanelgroup.add(ycustomNameButton);
                                ylegendradiopanelgroup.add(Box.createHorizontalStrut(2));
                                ylegendradiopanelgroup.add(Box.createGlue());

                                ylegendconfigureButton = new JButton(configureIcon);
                                ylegendconfigureButton.setToolTipText("Set Y-axis label preferences");
                                ylegendconfigureButton.setMargin(new Insets(0,0,0,0));

                                ylegendconfigureButton.addActionListener(new ActionListener(){
                                            public void actionPerformed(ActionEvent ae) {
                                                yaxispopup.show(ylegendconfigureButton, mainpanel.getX() , mainpanel.getY()+25);}});
                                ylegendradiopanelgroup.add(ylegendconfigureButton);
                                ylegendradiopanelgroup.add(Box.createHorizontalStrut(2));

                                        ButtonGroup ylegendRadioGroup = new ButtonGroup();
                                        ylegendRadioGroup.add(ynoneButton);
                                        ylegendRadioGroup.add(ycustomNameButton);
                                        ylegendRadioGroup.add(ycolumnNameButton);
                                        ynoneButton.addActionListener(new ActionListener(){
                                            public void actionPerformed(ActionEvent ae) {
                                                useYColumnName = false;
                                                useYCustomName = false;
                                                updatePlot();}});
                                        ycustomNameButton.addActionListener(new ActionListener(){
                                            public void actionPerformed(ActionEvent ae) {
                                                useYColumnName = false;
                                                useYCustomName = true;
                                                updatePlot();}});
                                        ycolumnNameButton.addActionListener(new ActionListener(){
                                            public void actionPerformed(ActionEvent ae) {
                                                useYColumnName = true;
                                                useYCustomName = false;
                                                updatePlot();}});
//                                SpringUtil.makeCompactGrid (ylegendradiopanelgroup, 1, 3, 0,0,0,0);
                                ynoneButton.setComponentPopupMenu(yaxispopup);
                                ynoneButton.setToolTipText("Right click to set Y-axis label preferences");
                                ycustomNameButton.setComponentPopupMenu(yaxispopup);
                                ycustomNameButton.setToolTipText("Right click to set Y-axis label preferences");
                                ycolumnNameButton.setComponentPopupMenu(yaxispopup);
                                ycolumnNameButton.setToolTipText("Right click to set Y-axis label preferences");
                        ylegendgroup.add (ylegendradiopanelgroup);

                                // Y-LEGEND TEXT FIELD
                                ylegendField = new JTextField (yLegend);
                                ylegendField.setBorder (BorderFactory.createEmptyBorder(0,4,0,4));
                                ylegendField.getDocument().addDocumentListener (new DocumentListener()
                                    {
                                    public void insertUpdate (DocumentEvent ev)
                                        {
                                        yLegend = ylegendField.getText();
                                        updatePlot();
                                        }
                                    public void removeUpdate (DocumentEvent ev)
                                        {
                                        yLegend = ylegendField.getText();
                                        updatePlot();
                                        }
                                    public void changedUpdate (DocumentEvent ev)
                                        {
                                        yLegend = ylegendField.getText();
                                        updatePlot();
                                        }
                                    });
                                ylegendField.setPreferredSize (new Dimension(250,20));
                                ylegendField.setHorizontalAlignment (JTextField.LEFT);
                                ylegendField.setComponentPopupMenu(yaxispopup);
                                ylegendField.setToolTipText("Right click to set Y-axis label preferences");

                        ylegendgroup.add (ylegendField);
                        SpringUtil.makeCompactGrid (ylegendgroup, 2, 1, 0,0,0,0);
                mainpanelgroupd.add (ylegendgroup);

            //TRIM DATASET SUBPANEL

                JPanel trimdatapanel = new JPanel (new SpringLayout());
                trimdatapanel.setBorder(BorderFactory.createTitledBorder("Trim Data Samples"));

                //TRIM DATA FROM HEAD STEPSIZE POPUP

                trimdataheadsteppopup = new JPopupMenu();
                JPanel trimdataheadsteppanel = new JPanel();
                trimdataheadstepmodel = new SpinnerListModel(integerspinnerscalelist);
                trimdataheadstepspinner = new JSpinner(trimdataheadstepmodel);
                trimdataheadstepspinner.setValue(intConvertToText(excludedHeadSamplesStep));
                trimdataheadstepspinner.addChangeListener (new ChangeListener()
                    {
                    public void stateChanged(ChangeEvent ev)
                        {
                        excludedHeadSamplesStep = Integer.parseInt(((String)trimdataheadstepspinner.getValue()).trim());
                        trimdataheadspinner.setModel(new SpinnerNumberModel(new Integer(excludedHeadSamples), new Integer(0), null, new Integer(excludedHeadSamplesStep)));
                        Prefs.set("plot.excludedHeadSamplesStep",excludedHeadSamplesStep);
                        }
                    });

                JLabel trimdataheadsteplabel = new JLabel ("Stepsize:");
                trimdataheadsteppanel.add(trimdataheadsteplabel);
                trimdataheadsteppanel.add(trimdataheadstepspinner);
                trimdataheadsteppopup.add(trimdataheadsteppanel);
                trimdataheadsteppopup.setLightWeightPopupEnabled(false);
                trimdataheadsteppopup.addPopupMenuListener (new PopupMenuListener()
                    {
                    public void popupMenuCanceled(PopupMenuEvent popupMenuEvent) {
                            trimdataheadsteppopup.setVisible(false);
                            }

                    public void popupMenuWillBecomeInvisible(PopupMenuEvent popupMenuEvent) {
                            trimdataheadsteppopup.setVisible(true);
                            }

                    public void popupMenuWillBecomeVisible(PopupMenuEvent popupMenuEvent) {
                            trimdataheadsteppopup.setVisible(false);
                            }
                    });

                JLabel trimdataheadlabel = new JLabel ("Head");
                trimdataheadlabel.setHorizontalAlignment(JLabel.RIGHT);
                trimdataheadlabel.setPreferredSize(new Dimension(50, 25));
                trimdataheadlabel.setToolTipText("Right click to set spinner stepsize");
                trimdataheadlabel.setComponentPopupMenu(trimdataheadsteppopup);
                trimdatapanel.add (trimdataheadlabel);

                trimdataheadmodel = new SpinnerNumberModel(new Integer(excludedHeadSamples), new Integer(0), null, new Integer(excludedHeadSamplesStep));
                trimdataheadspinner = new JSpinner(trimdataheadmodel);
                trimdataheadspinner.setComponentPopupMenu(trimdataheadsteppopup);
                trimdataheadspinner.setToolTipText("Right click to set spinner stepsize");
                trimdataheadspinner.addChangeListener (new ChangeListener()
                    {
                    public void stateChanged(ChangeEvent ev)
                        {
                        excludedHeadSamples = ((Integer)trimdataheadspinner.getValue()).intValue();
                        Prefs.set("plot.excludedHeadSamples", excludedHeadSamples);
                        updatePlot();
                        }
                    });
                trimdataheadspinner.addMouseWheelListener( new MouseWheelListener()
                    {
                    public void mouseWheelMoved( MouseWheelEvent e )
                        {
                        int newValue =  new Integer(((Integer)trimdataheadspinner.getValue()).intValue() - e.getWheelRotation());
                        if (newValue >= 0) trimdataheadspinner.setValue(newValue);
                        }
                    });
                trimdatapanel.add (trimdataheadspinner);

                //TRIM DATA FROM TAIL STEPSIZE POPUP

                trimdatatailsteppopup = new JPopupMenu();
                JPanel trimdatatailsteppanel = new JPanel();
                trimdatatailstepmodel = new SpinnerListModel(integerspinnerscalelist);
                trimdatatailstepspinner = new JSpinner(trimdatatailstepmodel);
                trimdatatailstepspinner.setValue(intConvertToText(excludedTailSamplesStep));
                trimdatatailstepspinner.addChangeListener (new ChangeListener()
                    {
                    public void stateChanged(ChangeEvent ev)
                        {
                        excludedTailSamplesStep = Integer.parseInt(((String)trimdatatailstepspinner.getValue()).trim());
                        trimdatatailspinner.setModel(new SpinnerNumberModel(new Integer(excludedTailSamples), new Integer(0), null, new Integer(excludedTailSamplesStep)));
                        Prefs.set("plot.excludedTailSamplesStep",excludedTailSamplesStep);
                        }
                    });

                JLabel trimdatatailsteplabel = new JLabel ("Stepsize:");
                trimdatatailsteppanel.add(trimdatatailsteplabel);
                trimdatatailsteppanel.add(trimdatatailstepspinner);
                trimdatatailsteppopup.add(trimdatatailsteppanel);
                trimdatatailsteppopup.setLightWeightPopupEnabled(false);
                trimdatatailsteppopup.addPopupMenuListener (new PopupMenuListener()
                    {
                    public void popupMenuCanceled(PopupMenuEvent popupMenuEvent) {
                            trimdatatailsteppopup.setVisible(false);
                            }

                    public void popupMenuWillBecomeInvisible(PopupMenuEvent popupMenuEvent) {
                            trimdatatailsteppopup.setVisible(true);
                            }

                    public void popupMenuWillBecomeVisible(PopupMenuEvent popupMenuEvent) {
                            trimdatatailsteppopup.setVisible(false);
                            }
                    });

                JLabel trimdatataillabel = new JLabel ("Tail");
                trimdatataillabel.setHorizontalAlignment(JLabel.RIGHT);
                trimdatataillabel.setPreferredSize(new Dimension(50, 25));
                trimdatataillabel.setToolTipText("Right click to set spinner stepsize");
                trimdatataillabel.setComponentPopupMenu(trimdatatailsteppopup);
                trimdatapanel.add (trimdatataillabel);

                trimdatatailmodel = new SpinnerNumberModel(new Integer(excludedTailSamples), new Integer(0), null, new Integer(excludedTailSamplesStep));
                trimdatatailspinner = new JSpinner(trimdatatailmodel);
                trimdatatailspinner.setComponentPopupMenu(trimdatatailsteppopup);
                trimdatatailspinner.setToolTipText("Right click to set spinner stepsize");
                trimdatatailspinner.addChangeListener (new ChangeListener()
                    {
                    public void stateChanged(ChangeEvent ev)
                        {
                        excludedTailSamples = ((Integer)trimdatatailspinner.getValue()).intValue();
                        Prefs.set("plot.excludedTailSamples", excludedTailSamples);
                        updatePlot();
                        }
                    });
                trimdatatailspinner.addMouseWheelListener( new MouseWheelListener()
                    {
                    public void mouseWheelMoved( MouseWheelEvent e )
                        {
                        int newValue =  new Integer(((Integer)trimdatatailspinner.getValue()).intValue() - e.getWheelRotation());
                        if (newValue >= 0) trimdatatailspinner.setValue(newValue);
                        }
                    });
                trimdatapanel.add (trimdatatailspinner);
                SpringUtil.makeCompactGrid (trimdatapanel, 2, 2, 0,0,0,0);
                mainpanelgroupd.add (trimdatapanel);

                SpringUtil.makeCompactGrid (mainpanelgroupd, 1, 3, 0,0,0,0);
                mainpanel.add(mainpanelgroupd);


    // MAIN PANEL SCALING GROUP START (GROUP E)

            //X-SCALING SUBPANEL

                JPanel xscalingpanel = new JPanel (new SpringLayout());
                xscalingpanel.setBorder(BorderFactory.createTitledBorder("X-axis"));
                autoxButton = new JRadioButton("Auto X-range");
                autoxButton.setHorizontalAlignment(JLabel.CENTER);
                firstxButton = new JRadioButton("First X-value as min");
                firstxButton.setHorizontalAlignment(JLabel.CENTER);
                customxButton = new JRadioButton("Custom X-range");
                customxButton.setHorizontalAlignment(JLabel.CENTER);
                if (autoScaleX)
                        {
                        autoxButton.setSelected(true);
                        firstxButton.setSelected(false);
                        customxButton.setSelected(false);
                        }
                else if (useFirstX)
                        {
                        autoxButton.setSelected(false);
                        firstxButton.setSelected(true);
                        customxButton.setSelected(false);
                        }
                else
                        {
                        autoxButton.setSelected(false);
                        firstxButton.setSelected(false);
                        customxButton.setSelected(true);
                        }

                xscalingpanel.add (autoxButton);
                xscalingpanel.add (firstxButton);
                xscalingpanel.add (customxButton);

                ButtonGroup xaxisradiogroup = new ButtonGroup();
                        xaxisradiogroup.add(autoxButton);
                        xaxisradiogroup.add(firstxButton);
                        xaxisradiogroup.add(customxButton);
                autoxButton.addActionListener(new ActionListener(){
                    public void actionPerformed(ActionEvent ae) {
                        autoScaleX = true;
                        useFirstX = false;
                        updatePlot();}});
                firstxButton.addActionListener(new ActionListener(){
                    public void actionPerformed(ActionEvent ae) {
                        autoScaleX = false;
                        useFirstX = true;
                        updatePlot();}});
                customxButton.addActionListener(new ActionListener(){
                    public void actionPerformed(ActionEvent ae) {
                        autoScaleX = false;
                        useFirstX = false;
                        updatePlot();}});

                ImageIcon grabautoIcon = createImageIcon("astroj/images/grabautoscale.png", "Transfer autoscale values to custom scale values");

                JPanel grabautoxpanel = new JPanel (new SpringLayout());
                grabautoxbutton = new JButton(grabautoIcon);
                grabautoxbutton.setMargin(new Insets(0,0,0,0));
                Dimension d = new Dimension(60,15);
                grabautoxbutton.setMaximumSize(d);
                grabautoxpanel.setBorder(BorderFactory.createEmptyBorder(0, 0, 0, 0));
                grabautoxbutton.setToolTipText("Transfer 'Auto X-range' min and max values to 'Custom X-range'");

                grabautoxbutton.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                        xMin = xautoscalemin;
                        xminspinner.setValue(xMin);
                        xMax = xautoscalemax;
                        xmaxspinner.setValue(xMax);
                        autoxButton.setSelected(false);
                        firstxButton.setSelected(false);
                        customxButton.setSelected(true);
                        autoScaleX = false;
                        useFirstX = false;
                        updatePlot();
                    }
                });
                grabautoxpanel.add (grabautoxbutton);

                SpringUtil.makeCompactGrid (grabautoxpanel,1,1, 30,0,10,2);
                xscalingpanel.add (grabautoxpanel);

                //XWIDTH STEPSIZE POPUP

                xwidthsteppopup = new JPopupMenu();
                JPanel xwidthsteppanel = new JPanel();
                xwidthstepmodel = new SpinnerListModel(spinnerscalelist);
                xwidthstepspinner = new JSpinner(xwidthstepmodel);
                xwidthstepspinner.setValue(convertToText(xWidthStep));
                xwidthstepspinner.addChangeListener (new ChangeListener()
                    {
                    public void stateChanged(ChangeEvent ev)
                        {
                        xWidthStep = Double.parseDouble((String)xwidthstepspinner.getValue());
                        xwidthspinner.setModel(new SpinnerNumberModel(new Double(xWidth), new Double(0), null, new Double(xWidthStep)));
                        Prefs.set("plot.xWidthStep",xWidthStep);
                        }
                    });

                JLabel xwidthsteplabel = new JLabel ("Stepsize:");
                xwidthsteppanel.add(xwidthsteplabel);
                xwidthsteppanel.add(xwidthstepspinner);
                xwidthsteppopup.add(xwidthsteppanel);
                xwidthsteppopup.setLightWeightPopupEnabled(false);
                xwidthsteppopup.setLocation(500, 500);
                xwidthsteppopup.addPopupMenuListener (new PopupMenuListener()
                    {
                    public void popupMenuCanceled(PopupMenuEvent popupMenuEvent) {
                            xwidthsteppopup.setVisible(false);
                            }

                    public void popupMenuWillBecomeInvisible(PopupMenuEvent popupMenuEvent) {
                            xwidthsteppopup.setVisible(true);
                            }

                    public void popupMenuWillBecomeVisible(PopupMenuEvent popupMenuEvent) {
                            xwidthsteppopup.setVisible(false);
                            }
                    });

                JPanel xwidthpanel = new JPanel (new SpringLayout());
                JLabel xwidthlabel = new JLabel ("X-width");
                xwidthlabel.setHorizontalAlignment(JLabel.RIGHT);
                xwidthlabel.setPreferredSize(new Dimension(60, 25));
                xwidthlabel.setComponentPopupMenu(xwidthsteppopup);
                xwidthlabel.setToolTipText("Right click to set spinner stepsize");
                xwidthpanel.add (xwidthlabel);
                xwidthspinnermodel = new SpinnerNumberModel(new Double(xWidth), new Double(0.000001), null, new Double(xWidthStep));
                xwidthspinner = new JSpinner(xwidthspinnermodel);
                xwidthspinner.setToolTipText("Right click to set spinner stepsize");
                xwidthspinner.addChangeListener (new ChangeListener()
                    {
                    public void stateChanged(ChangeEvent ev)
                        {
                        xWidth = ((Double)xwidthspinner.getValue()).doubleValue();
                        updatePlot();
                        }
                    });
                xwidthspinner.addMouseWheelListener( new MouseWheelListener()
                    {
                    public void mouseWheelMoved( MouseWheelEvent e )
                        {
                        double newvalue = ((Double)xwidthspinner.getValue()).doubleValue() - e.getWheelRotation()*xWidthStep ;
                        if (newvalue > 0)
                            xwidthspinner.setValue(newvalue);
                        else
                            xwidthspinner.setValue(0.000001);
                        }
                    });
                xwidthpanel.add (xwidthspinner);
                xwidthspinner.setComponentPopupMenu(xwidthsteppopup);
                SpringUtil.makeCompactGrid (xwidthpanel, 1, 2, 2,2,0,0);
                xscalingpanel.add (xwidthpanel);

                //XMAX STEPSIZE POPUP

                xmaxsteppopup = new JPopupMenu();
                JPanel xmaxsteppanel = new JPanel();
                xmaxstepmodel = new SpinnerListModel(spinnerscalelist);
                xmaxstepspinner = new JSpinner(xmaxstepmodel);
                xmaxstepspinner.setValue(convertToText(xMaxStep));
                xmaxstepspinner.addChangeListener (new ChangeListener()
                    {
                    public void stateChanged(ChangeEvent ev)
                        {
                        xMaxStep = Double.parseDouble((String)xmaxstepspinner.getValue());
                        xmaxspinner.setModel(new SpinnerNumberModel(new Double(xMax), null, null, new Double(xMaxStep)));
                        Prefs.set("plot.xMaxStep",xMaxStep);
                        }
                    });

                JLabel xmaxsteplabel = new JLabel ("Stepsize:");
                xmaxsteppanel.add(xmaxsteplabel);
                xmaxsteppanel.add(xmaxstepspinner);
                xmaxsteppopup.add(xmaxsteppanel);
                xmaxsteppopup.setLightWeightPopupEnabled(false);
                xmaxsteppopup.addPopupMenuListener (new PopupMenuListener()
                    {
                    public void popupMenuCanceled(PopupMenuEvent popupMenuEvent) {
                            xmaxsteppopup.setVisible(false);
                            }

                    public void popupMenuWillBecomeInvisible(PopupMenuEvent popupMenuEvent) {
                            xmaxsteppopup.setVisible(true);
                            }

                    public void popupMenuWillBecomeVisible(PopupMenuEvent popupMenuEvent) {
                            xmaxsteppopup.setVisible(false);
                            }
                    });

                JPanel customxmaxpanel = new JPanel(new SpringLayout());
                JLabel xmaxlabel = new JLabel ("X-max");
                xmaxlabel.setHorizontalAlignment(JLabel.RIGHT);
                xmaxlabel.setPreferredSize(new Dimension(50, 25));
                xmaxlabel.setToolTipText("Right click to set spinner stepsize");
                xmaxlabel.setComponentPopupMenu(xmaxsteppopup);
                customxmaxpanel.add (xmaxlabel);

                xmaxspinnermodel = new SpinnerNumberModel(new Double(xMax),null, null, new Double(xMaxStep));
                xmaxspinner = new JSpinner(xmaxspinnermodel);
                xmaxspinner.setComponentPopupMenu(xmaxsteppopup);
                xmaxspinner.setToolTipText("Right click to set spinner stepsize");
                xmaxspinner.addChangeListener (new ChangeListener()
                    {
                    public void stateChanged(ChangeEvent ev)
                        {
                        xMax = ((Double)xmaxspinner.getValue()).doubleValue();
                        updatePlot();
                        }
                    });
                xmaxspinner.addMouseWheelListener( new MouseWheelListener()
                    {
                    public void mouseWheelMoved( MouseWheelEvent e )
                        {
                        xmaxspinner.setValue(((Double)xmaxspinner.getValue()).doubleValue()
                                - e.getWheelRotation()*xMaxStep );
                        }
                    });
                customxmaxpanel.add (xmaxspinner);
                SpringUtil.makeCompactGrid (customxmaxpanel, 1, 2, 2,2,0,0);
                xscalingpanel.add (customxmaxpanel);

                JPanel xmultiplierpanel = new JPanel(new SpringLayout());
                xmultiplierpanel.setBorder(BorderFactory.createLineBorder(Color.lightGray));
                JLabel xmultiplierlabel = new JLabel ("X x 1E");
                xmultiplierlabel.setHorizontalAlignment(JLabel.RIGHT);
                xmultiplierpanel.setMaximumSize(new Dimension(250, 30));
                xmultiplierpanel.add (xmultiplierlabel);
                xmultiplierlabel.setToolTipText("Sets the X-axis multiplication factor");
                xmultipliermodel = new SpinnerNumberModel(new Integer(xExponent),null, null, new Integer(1));
                xmultiplierspinner = new JSpinner(xmultipliermodel);
                xmultiplierspinner.setMaximumSize(new Dimension (35,25));
                xmultiplierspinner.setToolTipText("Sets the X-axis multiplication factor");
                xmultiplierspinner.addChangeListener (new ChangeListener()
                    {
                    public void stateChanged(ChangeEvent ev)
                        {
                        xExponent = ((Integer)xmultiplierspinner.getValue()).intValue();
                        updatePlot();
                        }
                    });
                xmultiplierspinner.addMouseWheelListener( new MouseWheelListener()
                    {
                    public void mouseWheelMoved( MouseWheelEvent e )
                        {
                        xmultiplierspinner.setValue( new Integer(((Integer)xmultiplierspinner.getValue()).intValue()
                                - e.getWheelRotation()) );
                        }
                    });
                xmultiplierpanel.add (xmultiplierspinner);
                SpringUtil.makeCompactGrid (xmultiplierpanel, 1,2, 2,2,2,0);
                xscalingpanel.add (xmultiplierpanel);

                //XMIN STEPSIZE POPUP

                xminsteppopup = new JPopupMenu();
                JPanel xminsteppanel = new JPanel();
                xminstepmodel = new SpinnerListModel(spinnerscalelist);
                xminstepspinner = new JSpinner(xminstepmodel);
                xminstepspinner.setValue(convertToText(xMinStep));
                xminstepspinner.addChangeListener (new ChangeListener()
                    {
                    public void stateChanged(ChangeEvent ev)
                        {
                        xMinStep = Double.parseDouble((String)xminstepspinner.getValue());
                        xminspinner.setModel(new SpinnerNumberModel(new Double(xMin), null, null, new Double(xMinStep)));
                        Prefs.set("plot.xMinStep",xMinStep);
                        }
                    });

                JLabel xminsteplabel = new JLabel ("Stepsize:");
                xminsteppanel.add(xminsteplabel);
                xminsteppanel.add(xminstepspinner);
                xminsteppopup.add(xminsteppanel);
                xminsteppopup.setLightWeightPopupEnabled(false);
                xminsteppopup.addPopupMenuListener (new PopupMenuListener()
                    {
                    public void popupMenuCanceled(PopupMenuEvent popupMenuEvent) {
                            xminsteppopup.setVisible(false);
                            }

                    public void popupMenuWillBecomeInvisible(PopupMenuEvent popupMenuEvent) {
                            xminsteppopup.setVisible(true);
                            }

                    public void popupMenuWillBecomeVisible(PopupMenuEvent popupMenuEvent) {
                            xminsteppopup.setVisible(false);
                            }
                    });

                ImageIcon grabfirstxIcon = createImageIcon("astroj/images/grabautoscale.png", "Transfer autoscale values to custom scale values");
                JPanel grabfirstxpanel = new JPanel (new SpringLayout());
                JButton grabfirstxbutton = new JButton(grabfirstxIcon);
                grabfirstxbutton.setMargin(new Insets(0,0,0,0));
                grabfirstxbutton.setMaximumSize(new Dimension(60,15));
                grabfirstxpanel.setBorder(BorderFactory.createEmptyBorder(0, 0, 0, 0));
                grabfirstxbutton.setToolTipText("Transfer 'First-X' min and max values to 'Custom X-range'");

                grabfirstxbutton.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                        xMin = firstXmin;
                        xminspinner.setValue(firstXmin);
                        xMax = firstXmin + xWidth;
                        xmaxspinner.setValue(xMax);
                        autoxButton.setSelected(false);
                        firstxButton.setSelected(false);
                        customxButton.setSelected(true);
                        autoScaleX = false;
                        useFirstX = false;
                        updatePlot();
                    }
                });
                grabfirstxpanel.add (grabfirstxbutton);

                SpringUtil.makeCompactGrid (grabfirstxpanel,1,1, 72,2,22,0);                
                xscalingpanel.add (grabfirstxpanel);


                JPanel customxminpanel = new JPanel(new SpringLayout());
                JLabel xminlabel = new JLabel ("X-min");
                xminlabel.setHorizontalAlignment(JLabel.RIGHT);
                xminlabel.setPreferredSize(new Dimension(50, 25));
                xminlabel.setToolTipText("Right click to set spinner stepsize");
                xminlabel.setComponentPopupMenu(xminsteppopup);
                customxminpanel.add (xminlabel);

                xminspinnermodel = new SpinnerNumberModel(new Double(xMin),null, null, new Double(xMinStep));
                xminspinner = new JSpinner(xminspinnermodel);
                xminspinner.setComponentPopupMenu(xminsteppopup);
                xminspinner.setToolTipText("Right click to set spinner stepsize");
                xminspinner.addChangeListener (new ChangeListener()
                    {
                    public void stateChanged(ChangeEvent ev)
                        {
                        xMin = ((Double)xminspinner.getValue()).doubleValue();
                        updatePlot();
                        }
                    });
                xminspinner.addMouseWheelListener( new MouseWheelListener()
                    {
                    public void mouseWheelMoved( MouseWheelEvent e )
                        {
                        xminspinner.setValue(((Double)xminspinner.getValue()).doubleValue()
                                - e.getWheelRotation()*xMinStep );
                        }
                    });
                customxminpanel.add (xminspinner);
                SpringUtil.makeCompactGrid (customxminpanel, 1, 2, 2,2,0,0);
                xscalingpanel.add (customxminpanel);

                SpringUtil.makeCompactGrid (xscalingpanel, 3, 3, 0,0,0,0);

                mainpanelgroupe.add (xscalingpanel);

            //Y-SCALING SUBPANEL

                JPanel yscalingpanel = new JPanel (new SpringLayout());
                yscalingpanel.setBorder(BorderFactory.createTitledBorder("Y-axis"));
                autoyButton = new JRadioButton("Auto Y-range");
                autoyButton.setHorizontalAlignment(JLabel.CENTER);
                customyButton = new JRadioButton("Custom Y-range");
                customyButton.setHorizontalAlignment(JLabel.CENTER);
                if (autoScaleY)
                        {
                        autoyButton.setSelected(true);
                        customyButton.setSelected(false);
                        }
                else
                        {
                        autoyButton.setSelected(false);
                        customyButton.setSelected(true);
                        }

                yscalingpanel.add (autoyButton);
                yscalingpanel.add (customyButton);

                ButtonGroup yaxisradiogroup = new ButtonGroup();
                        yaxisradiogroup.add(autoyButton);
                        yaxisradiogroup.add(customyButton);
                autoyButton.addActionListener(new ActionListener(){
                    public void actionPerformed(ActionEvent ae) {
                        autoScaleY = true;
                        updatePlot();}});
                customyButton.addActionListener(new ActionListener(){
                    public void actionPerformed(ActionEvent ae) {
                        autoScaleY = false;
                        updatePlot();}});

                JPanel grabautoypanel = new JPanel (new SpringLayout());
                grabautoybutton = new JButton(grabautoIcon);
                grabautoybutton.setMargin(new Insets(0,0,0,0));
                grabautoybutton.setMaximumSize(d);
                grabautoypanel.setBorder(BorderFactory.createEmptyBorder(0, 0, 0, 0));
                grabautoybutton.setToolTipText("Transfer auto scale Y-range values to custom scale Y-range values");
                grabautoybutton.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                        yMin = yautoscalemin;
                        yminspinner.setValue(yMin);
                        yMax = yautoscalemax;
                        ymaxspinner.setValue(yMax);
                        autoyButton.setSelected(false);
                        customyButton.setSelected(true);
                        autoScaleY = false;
                        updatePlot();
                    }
                    });
                grabautoypanel.add (grabautoybutton);
                SpringUtil.makeCompactGrid (grabautoypanel,1,1, 30,0,10,2);
                yscalingpanel.add (grabautoypanel);

                //YMAX STEPSIZE POPUP

                ymaxsteppopup = new JPopupMenu();
                JPanel ymaxsteppanel = new JPanel();
                ymaxstepmodel = new SpinnerListModel(spinnerscalelist);
                ymaxstepspinner = new JSpinner(ymaxstepmodel);
                ymaxstepspinner.setValue(convertToText(yMaxStep));
                ymaxstepspinner.addChangeListener (new ChangeListener()
                    {
                    public void stateChanged(ChangeEvent ev)
                        {
                        yMaxStep = Double.parseDouble((String)ymaxstepspinner.getValue());
                        ymaxspinner.setModel(new SpinnerNumberModel(new Double(yMax), null, null, new Double(yMaxStep)));
                        Prefs.set("plot.yMaxStep",yMaxStep);
                        }
                    });

                JLabel ymaxsteplabel = new JLabel ("Stepsize:");
                ymaxsteppanel.add(ymaxsteplabel);
                ymaxsteppanel.add(ymaxstepspinner);
                ymaxsteppopup.add(ymaxsteppanel);
                ymaxsteppopup.setLightWeightPopupEnabled(false);
                ymaxsteppopup.addPopupMenuListener (new PopupMenuListener()
                    {
                    public void popupMenuCanceled(PopupMenuEvent popupMenuEvent) {
                            ymaxsteppopup.setVisible(false);
                            }

                    public void popupMenuWillBecomeInvisible(PopupMenuEvent popupMenuEvent) {
                            ymaxsteppopup.setVisible(true);
                            }

                    public void popupMenuWillBecomeVisible(PopupMenuEvent popupMenuEvent) {
                            ymaxsteppopup.setVisible(false);
                            }
                    });

                JPanel customymaxpanel = new JPanel(new SpringLayout());
                JLabel ymaxlabel = new JLabel ("Y-max");
                ymaxlabel.setHorizontalAlignment(JLabel.RIGHT);
                ymaxlabel.setPreferredSize(new Dimension(50, 25));
                ymaxlabel.setToolTipText("Right click to set spinner stepsize");
                ymaxlabel.setComponentPopupMenu(ymaxsteppopup);
                customymaxpanel.add (ymaxlabel);

                ymaxspinnermodel = new SpinnerNumberModel(new Double(yMax), null, null, new Double(yMaxStep));
                ymaxspinner = new JSpinner(ymaxspinnermodel);
                ymaxspinner.setComponentPopupMenu(ymaxsteppopup);
                ymaxspinner.setToolTipText("Right click to set spinner stepsize");
                ymaxspinner.addChangeListener (new ChangeListener()
                    {
                    public void stateChanged(ChangeEvent ev)
                        {
                        yMax = ((Double)ymaxspinner.getValue()).doubleValue();
                        updatePlot();
                        }
                    });
                ymaxspinner.addMouseWheelListener( new MouseWheelListener()
                    {
                    public void mouseWheelMoved( MouseWheelEvent e )
                        {
                        ymaxspinner.setValue(((Double)ymaxspinner.getValue()).doubleValue()
                                - e.getWheelRotation()*yMaxStep );
                        }
                    });
                customymaxpanel.add (ymaxspinner);
                SpringUtil.makeCompactGrid (customymaxpanel, 1, 2, 2,2,0,0);
                yscalingpanel.add (customymaxpanel);

                JPanel ymultiplierpanel = new JPanel(new SpringLayout());
                ymultiplierpanel.setBorder(BorderFactory.createLineBorder(Color.lightGray));
                JLabel ymultiplierlabel = new JLabel ("Y x 1E");
                ymultiplierlabel.setHorizontalAlignment(JLabel.RIGHT);
                ymultiplierpanel.setMaximumSize(new Dimension(250, 30));
                ymultiplierpanel.add (ymultiplierlabel);
                ymultiplierlabel.setToolTipText("Sets the Y-axis multiplication factor");
                ymultipliermodel = new SpinnerNumberModel(new Integer(yExponent),null, null, new Integer(1));
                ymultiplierspinner = new JSpinner(ymultipliermodel);
                ymultiplierspinner.setMaximumSize(new Dimension (35,25));
                ymultiplierspinner.setToolTipText("Sets the Y-axis multiplication factor");
                ymultiplierspinner.addChangeListener (new ChangeListener()
                    {
                    public void stateChanged(ChangeEvent ev)
                        {
                        yExponent = ((Integer)ymultiplierspinner.getValue()).intValue();
                        updatePlot();
                        }
                    });
                ymultiplierspinner.addMouseWheelListener( new MouseWheelListener()
                    {
                    public void mouseWheelMoved( MouseWheelEvent e )
                        {
                        ymultiplierspinner.setValue( new Integer(((Integer)ymultiplierspinner.getValue()).intValue()
                                - e.getWheelRotation()) );
                        }
                    });
                ymultiplierpanel.add (ymultiplierspinner);
                SpringUtil.makeCompactGrid (ymultiplierpanel, 1,2, 2,2,2,0);
                yscalingpanel.add (ymultiplierpanel);

                //YMIN STEPSIZE POPUP

                yminsteppopup = new JPopupMenu();
                JPanel yminsteppanel = new JPanel();
                yminstepmodel = new SpinnerListModel(spinnerscalelist);
                yminstepspinner = new JSpinner(yminstepmodel);
                yminstepspinner.setValue(convertToText(yMinStep));
                yminstepspinner.addChangeListener (new ChangeListener()
                    {
                    public void stateChanged(ChangeEvent ev)
                        {
                        yMinStep = Double.parseDouble((String)yminstepspinner.getValue());
                        yminspinner.setModel(new SpinnerNumberModel(new Double(yMin), null, null, new Double(yMinStep)));
                        Prefs.set("plot.yMinStep",yMinStep);
                        }
                    });

                JLabel yminsteplabel = new JLabel ("Stepsize:");
                yminsteppanel.add(yminsteplabel);
                yminsteppanel.add(yminstepspinner);
                yminsteppopup.add(yminsteppanel);
                yminsteppopup.setLightWeightPopupEnabled(false);
                yminsteppopup.addPopupMenuListener (new PopupMenuListener()
                    {
                    public void popupMenuCanceled(PopupMenuEvent popupMenuEvent) {
                            yminsteppopup.setVisible(false);
                            }

                    public void popupMenuWillBecomeInvisible(PopupMenuEvent popupMenuEvent) {
                            yminsteppopup.setVisible(true);
                            }

                    public void popupMenuWillBecomeVisible(PopupMenuEvent popupMenuEvent) {
                            yminsteppopup.setVisible(false);
                            }
                    });


                JPanel customyminpanel = new JPanel(new SpringLayout());
                JLabel yminlabel = new JLabel ("Y-min");
                yminlabel.setHorizontalAlignment(JLabel.RIGHT);
                yminlabel.setComponentPopupMenu(yminsteppopup);
                yminlabel.setPreferredSize(new Dimension(50, 25));
                yminlabel.setToolTipText("Right click to set spinner stepsize");
                customyminpanel.add (yminlabel);

                yminspinnermodel = new SpinnerNumberModel(new Double(yMin),  null, null, new Double(yMinStep));
                yminspinner = new JSpinner(yminspinnermodel);
                yminspinner.setComponentPopupMenu(yminsteppopup);
                yminspinner.setToolTipText("Right click to set spinner stepsize");
                yminspinner.addChangeListener (new ChangeListener()
                    {
                    public void stateChanged(ChangeEvent ev)
                        {
                        yMin = ((Double)yminspinner.getValue()).doubleValue();
                        updatePlot();
                        }
                    });
                yminspinner.addMouseWheelListener( new MouseWheelListener()
                    {
                    public void mouseWheelMoved( MouseWheelEvent e )
                        {
                        yminspinner.setValue(((Double)yminspinner.getValue()).doubleValue()
                                - e.getWheelRotation()*yMinStep );
                        }
                    });
                customyminpanel.add (yminspinner);
                SpringUtil.makeCompactGrid (customyminpanel, 1, 2, 2,2,0,0);
                yscalingpanel.add (customyminpanel);

                SpringUtil.makeCompactGrid (yscalingpanel, 3, 2, 0,0,0,0);
                mainpanelgroupe.add (yscalingpanel);


            //PLOTSIZE SUBPANEL

                JPanel plotsizepanel = new JPanel (new SpringLayout());
                plotsizepanel.setBorder(BorderFactory.createTitledBorder("Plot Size"));


                JLabel dummylabel10 = new JLabel ("");
                dummylabel10.setHorizontalAlignment(JLabel.CENTER);
                dummylabel10.setPreferredSize(new Dimension(50, 25));
                plotsizepanel.add (dummylabel10);

                JLabel dummylabel11 = new JLabel ("");
                dummylabel11.setHorizontalAlignment(JLabel.CENTER);
                dummylabel11.setPreferredSize(new Dimension(50, 25));
                plotsizepanel.add (dummylabel11);

                //PLOT HEIGHT STEPSIZE POPUP

                plotheightsteppopup = new JPopupMenu();
                JPanel plotheightsteppanel = new JPanel();
                plotheightstepmodel = new SpinnerListModel(integerspinnerscalelist);
                plotheightstepspinner = new JSpinner(plotheightstepmodel);
                plotheightstepspinner.setValue(intConvertToText(plotSizeYStep));
                plotheightstepspinner.addChangeListener (new ChangeListener()
                    {
                    public void stateChanged(ChangeEvent ev)
                        {
                        plotSizeYStep = Integer.parseInt(((String)plotheightstepspinner.getValue()).trim());
                        plotheightspinner.setModel(new SpinnerNumberModel(new Integer(plotSizeY), new Integer(1), null, new Integer(plotSizeYStep)));
                        Prefs.set("plot.plotSizeYStep",plotSizeYStep);
                        }
                    });

                JLabel plotheightsteplabel = new JLabel ("Stepsize:");
                plotheightsteppanel.add(plotheightsteplabel);
                plotheightsteppanel.add(plotheightstepspinner);
                plotheightsteppopup.add(plotheightsteppanel);
                plotheightsteppopup.setLightWeightPopupEnabled(false);
                plotheightsteppopup.addPopupMenuListener (new PopupMenuListener()
                    {
                    public void popupMenuCanceled(PopupMenuEvent popupMenuEvent) {
                            plotheightsteppopup.setVisible(false);
                            }

                    public void popupMenuWillBecomeInvisible(PopupMenuEvent popupMenuEvent) {
                            plotheightsteppopup.setVisible(true);
                            }

                    public void popupMenuWillBecomeVisible(PopupMenuEvent popupMenuEvent) {
                            plotheightsteppopup.setVisible(false);
                            }
                    });

                JLabel plotheightlabel = new JLabel ("Height");
                plotheightlabel.setHorizontalAlignment(JLabel.RIGHT);
                plotheightlabel.setPreferredSize(new Dimension(50, 25));
                plotheightlabel.setToolTipText("Right click to set spinner stepsize");
                plotheightlabel.setComponentPopupMenu(plotheightsteppopup);
                plotsizepanel.add (plotheightlabel);

                plotheightspinnermodel = new SpinnerNumberModel(new Integer(plotSizeY), new Integer(1), null, new Integer(plotSizeYStep));
                plotheightspinner = new JSpinner(plotheightspinnermodel);
                plotheightspinner.setComponentPopupMenu(plotheightsteppopup);
                plotheightspinner.setToolTipText("Right click to set spinner stepsize");
                plotheightspinner.addChangeListener (new ChangeListener()
                    {
                    public void stateChanged(ChangeEvent ev)
                        {
                        plotSizeY = ((Integer)plotheightspinner.getValue()).intValue();
                        updatePlot();
                        }
                    });
                plotheightspinner.addMouseWheelListener( new MouseWheelListener()
                    {
                    public void mouseWheelMoved( MouseWheelEvent e )
                        {
                        int newValue =  new Integer(((Integer)plotheightspinner.getValue()).intValue() - e.getWheelRotation()*plotSizeYStep);
                        if (newValue > 0) plotheightspinner.setValue(newValue);
                        }
                    });
                plotsizepanel.add (plotheightspinner);

                //PLOT WIDTH STEPSIZE POPUP

                plotwidthsteppopup = new JPopupMenu();
                JPanel plotwidthsteppanel = new JPanel();
                plotwidthstepmodel = new SpinnerListModel(integerspinnerscalelist);
                plotwidthstepspinner = new JSpinner(plotwidthstepmodel);
                plotwidthstepspinner.setValue(intConvertToText(plotSizeXStep));
                plotwidthstepspinner.addChangeListener (new ChangeListener()
                    {
                    public void stateChanged(ChangeEvent ev)
                        {
                        plotSizeXStep = Integer.parseInt(((String)plotwidthstepspinner.getValue()).trim());
                        plotwidthspinner.setModel(new SpinnerNumberModel(new Integer(plotSizeX), new Integer(1), null, new Integer(plotSizeXStep)));
                        Prefs.set("plot.plotSizeXStep",plotSizeXStep);
                        }
                    });
                JLabel plotwidthsteplabel = new JLabel ("Stepsize:");
                plotwidthsteppanel.add(plotwidthsteplabel);
                plotwidthsteppanel.add(plotwidthstepspinner);
                plotwidthsteppopup.add(plotwidthsteppanel);
                plotwidthsteppopup.setLightWeightPopupEnabled(false);
                plotwidthsteppopup.addPopupMenuListener (new PopupMenuListener()
                    {
                    public void popupMenuCanceled(PopupMenuEvent popupMenuEvent) {
                            plotwidthsteppopup.setVisible(false);
                            }

                    public void popupMenuWillBecomeInvisible(PopupMenuEvent popupMenuEvent) {
                            plotwidthsteppopup.setVisible(true);
                            }

                    public void popupMenuWillBecomeVisible(PopupMenuEvent popupMenuEvent) {
                            plotwidthsteppopup.setVisible(false);
                            }
                    });





                JLabel plotwidthlabel = new JLabel ("Width");
                plotwidthlabel.setHorizontalAlignment(JLabel.RIGHT);
                plotwidthlabel.setPreferredSize(new Dimension(50, 25));
                plotwidthlabel.setToolTipText("Right click to set spinner stepsize");
                plotwidthlabel.setComponentPopupMenu(plotwidthsteppopup);
                plotsizepanel.add (plotwidthlabel);

                plotwidthspinnermodel = new SpinnerNumberModel(new Integer(plotSizeX), new Integer(1), null, new Integer(plotSizeXStep));
                plotwidthspinner = new JSpinner(plotwidthspinnermodel);
                plotwidthspinner.setComponentPopupMenu(plotwidthsteppopup);
                plotwidthspinner.setToolTipText("Right click to set spinner stepsize");
                plotwidthspinner.addChangeListener (new ChangeListener()
                    {
                    public void stateChanged(ChangeEvent ev)
                        {
                        plotSizeX = ((Integer)plotwidthspinner.getValue()).intValue();
                        updatePlot();
                        }
                    });

                plotwidthspinner.addMouseWheelListener( new MouseWheelListener()
                    {
                    public void mouseWheelMoved( MouseWheelEvent e )
                        {
                        int newValue =  new Integer(((Integer)plotwidthspinner.getValue()).intValue() - e.getWheelRotation()*plotSizeXStep);
                        if (newValue > 0) plotwidthspinner.setValue(newValue);
                        }
                    });
                plotsizepanel.add (plotwidthspinner);
                SpringUtil.makeCompactGrid (plotsizepanel, 3, 2, 0,0,0,0);
                mainpanelgroupe.add (plotsizepanel);


        //FINALIZE GROUP E PANEL (SCALING PANEL)

                SpringUtil.makeCompactGrid (mainpanelgroupe, 1, 3, 0,0,0,0);
                mainpanel.add(mainpanelgroupe);
                
        // MAIN PANEL GROUP F START

                JPanel updateplotbuttonpanel = new JPanel (new SpringLayout());
                updateplotbutton = new JButton(" Redraw Plot ");
                updateplotbutton.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                        totalPanOffsetX=0.0;
                        totalPanOffsetY=0.0;
                        newPanOffsetX=0.0;
                        newPanOffsetY=0.0;
                        leftDragReleased=false;
                        zoom = 0.0;
                        updatePlot();}
                });

                updateplotbuttonpanel.add (updateplotbutton);
                SpringUtil.makeCompactGrid (updateplotbuttonpanel,1,1, 20,15,6,0);
                mainpanelgroupf.add (updateplotbuttonpanel);

                JPanel formatoptionspanel = new JPanel (new SpringLayout());
                TitledBorder formatoptiontitle = BorderFactory.createTitledBorder("Format Options");
                formatoptiontitle.setTitleJustification(TitledBorder.CENTER);
                formatoptionspanel.setBorder(formatoptiontitle);
                usextickscheckbox = new JCheckBox ("X tick marks", xTics);
                usextickscheckbox.addItemListener(new ItemListener(){
                    public void itemStateChanged(ItemEvent e) {
                        if (e.getStateChange() == ItemEvent.DESELECTED)
                                xTics = false;
                        else if (e.getStateChange() == ItemEvent.SELECTED)
                                xTics = true;
                        updatePlot();}});
                usextickscheckbox.setHorizontalAlignment(JLabel.CENTER);
                formatoptionspanel.add (usextickscheckbox);

                useytickscheckbox = new JCheckBox ("Y tick marks", yTics);
                useytickscheckbox.addItemListener(new ItemListener(){
                    public void itemStateChanged(ItemEvent e) {
                        if (e.getStateChange() == ItemEvent.DESELECTED)
                                yTics = false;
                        else if (e.getStateChange() == ItemEvent.SELECTED)
                                yTics = true;
                        updatePlot();}});
                useytickscheckbox.setHorizontalAlignment(JLabel.CENTER);
                formatoptionspanel.add (useytickscheckbox);

                usexgridcheckbox = new JCheckBox ("X grid lines", xGrid);
                usexgridcheckbox.addItemListener(new ItemListener(){
                    public void itemStateChanged(ItemEvent e) {
                        if (e.getStateChange() == ItemEvent.DESELECTED)
                                xGrid = false;
                        else if (e.getStateChange() == ItemEvent.SELECTED)
                                xGrid = true;
                        updatePlot();}});
                usexgridcheckbox.setHorizontalAlignment(JLabel.CENTER);
                formatoptionspanel.add (usexgridcheckbox);

                useygridcheckbox = new JCheckBox ("Y grid lines", yGrid);
                useygridcheckbox.addItemListener(new ItemListener(){
                    public void itemStateChanged(ItemEvent e) {
                        if (e.getStateChange() == ItemEvent.DESELECTED)
                                yGrid = false;
                        else if (e.getStateChange() == ItemEvent.SELECTED)
                                yGrid = true;
                        updatePlot();}});
                useygridcheckbox.setHorizontalAlignment(JLabel.CENTER);
                formatoptionspanel.add (useygridcheckbox);

                usexnumberscheckbox = new JCheckBox ("X numbers", xNumbers);
                usexnumberscheckbox.addItemListener(new ItemListener(){
                    public void itemStateChanged(ItemEvent e) {
                        if (e.getStateChange() == ItemEvent.DESELECTED)
                                xNumbers = false;
                        else if (e.getStateChange() == ItemEvent.SELECTED)
                                xNumbers = true;
                        updatePlot();}});
                usexnumberscheckbox.setHorizontalAlignment(JLabel.CENTER);
                formatoptionspanel.add (usexnumberscheckbox);

                useynumberscheckbox = new JCheckBox ("Y numbers", yNumbers);
                useynumberscheckbox.addItemListener(new ItemListener(){
                    public void itemStateChanged(ItemEvent e) {
                        if (e.getStateChange() == ItemEvent.DESELECTED)
                                yNumbers = false;
                        else if (e.getStateChange() == ItemEvent.SELECTED)
                                yNumbers = true;
                        updatePlot();}});
                useynumberscheckbox.setHorizontalAlignment(JLabel.CENTER);
                formatoptionspanel.add (useynumberscheckbox);
                SpringUtil.makeCompactGrid (formatoptionspanel, 1,6, 2,2,6,2);
                mainpanelgroupf.add (formatoptionspanel);

                JPanel closebuttonpanel = new JPanel (new SpringLayout());
                closebutton = new JButton("        Exit        ");
                closebutton.setForeground(Color.RED);
                closebutton.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                            saveAndClose(); }});
                closebuttonpanel.add (closebutton);
                SpringUtil.makeCompactGrid (closebuttonpanel, 1, 1, 6,15,0,0);
                mainpanelgroupf.add (closebuttonpanel);

                SpringUtil.makeCompactGrid (mainpanelgroupf, 1, 3, 2,0,6,0);
                mainpanel.add(mainpanelgroupf);

//FINALIZE MAIN PANEL

                SpringUtil.makeCompactGrid (mainpanel, 5, 1, 6,6,6,6);



                mainFrame.setJMenuBar(mainmenubar);
                mainFrame.add (mainscrollpane);
                    mainFrame.pack();
                mainFrame.setResizable (true);

                if (rememberWindowLocations)
                        {
                        Dimension mainScreenSize = Toolkit.getDefaultToolkit().getScreenSize();
                        if (!Prefs.isLocationOnScreen(new Point(mainFrameLocationX,mainFrameLocationY)))
                            {
                            mainFrameLocationX = mainScreenSize.width/2 - mainFrame.getWidth()/2;
                            mainFrameLocationY = mainScreenSize.height/2 - mainFrame.getHeight()/2;
                            }
                        mainFrame.setLocation(mainFrameLocationX, mainFrameLocationY);
                        }
                if (!plotAutoMode)
                        mainFrame.setVisible (true);
                showMoreCurvesJPanel();
                }





        static void showMoreCurvesJPanel()
                {

                subFrame = new JFrame ("Multi-plot Y-data");
                subFrame.setIconImage(plotIcon.getImage());
                mainsubpanel = new JPanel (new SpringLayout());
//                mainsubpanel.addMouseListener(panelMouseListener);
                mainsubpanel.addMouseMotionListener(panelMouseMotionListener);                
                FileDrop fileDrop = new FileDrop(mainsubpanel, BorderFactory.createEmptyBorder(),new FileDrop.Listener()
                    {   public void filesDropped( java.io.File[] files )
                        {
                        openDragAndDropFiles(files);
                        }
                    });
                
                subscrollpane = new JScrollPane(mainsubpanel);
                subFrame.setDefaultCloseOperation(javax.swing.WindowConstants.DISPOSE_ON_CLOSE);
                subFrame.addWindowListener(new WindowAdapter(){
                    @Override
                    public void windowClosing(WindowEvent e){
                            subFrameLocationX = subFrame.getLocation().x;
                            subFrameLocationY = subFrame.getLocation().y;
                            setSubpanelVisible = false;
                            Prefs.set("plot2.subFrameLocationX",subFrameLocationX);
                            Prefs.set("plot2.subFrameLocationY",subFrameLocationY);
                            }});

                JPanel mainsubpanelgroupa = new JPanel (new SpringLayout());
                mainsubpanelgroupa.setBorder(BorderFactory.createTitledBorder(""));

                JPanel mainsubpanelgroupb = new JPanel (new SpringLayout());
                mainsubpanelgroupb.setBorder(BorderFactory.createTitledBorder(""));

       // SUBPANEL START

             
                
                if (useWideDataPanel)
                        {
//                        constructTopGroupTopLabels(mainsubpanelgroupa);
//                        constructOtherGroupTopLabels(mainsubpanelgroupa);
                        constructTopGroupBottomLabels(mainsubpanelgroupa);
                        constructOtherGroupBottomLabels(mainsubpanelgroupa);

                        }
                else
                        {
//                        constructTopGroupTopLabels(mainsubpanelgroupa);
                        constructTopGroupBottomLabels(mainsubpanelgroupa);
//                        constructOtherGroupTopLabels(mainsubpanelgroupb);
                        constructOtherGroupBottomLabels(mainsubpanelgroupb);
                        }
                for (cur=0; cur<maxCurves; cur++)
                        {
                        constructTopGroup(mainsubpanelgroupa, cur);
                        if (useWideDataPanel)
                                constructOtherGroup(mainsubpanelgroupa, cur);
                        else
                                constructOtherGroup(mainsubpanelgroupb, cur);
                        }

                if (useWideDataPanel)
                        {
                        SpringUtil.makeCompactGrid (mainsubpanelgroupa, maxCurves+1, mainsubpanelgroupa.getComponentCount()/(maxCurves+1), 5,5,5,5);
                        mainsubpanel.add(mainsubpanelgroupa);
                        SpringUtil.makeCompactGrid (mainsubpanel, 1, 1, 5,5,5,5);
                        }
                else
                        {
                        SpringUtil.makeCompactGrid (mainsubpanelgroupa, maxCurves+1, mainsubpanelgroupa.getComponentCount()/(maxCurves+1), 5,5,5,5);
                        mainsubpanel.add(mainsubpanelgroupa);
                        SpringUtil.makeCompactGrid (mainsubpanelgroupb, maxCurves+1, mainsubpanelgroupb.getComponentCount()/(maxCurves+1), 5,5,5,5);
                        mainsubpanel.add(mainsubpanelgroupb);
                        SpringUtil.makeCompactGrid (mainsubpanel, mainsubpanel.getComponentCount(), 1, 5,5,5,5);
                        }


                
                
                subFrame.add (subscrollpane);
                subFrame.pack();
                subFrame.setResizable (true);

                if (rememberWindowLocations)
                        {
                        Dimension mainScreenSize = Toolkit.getDefaultToolkit().getScreenSize();
                        if (!Prefs.isLocationOnScreen(new Point(subFrameLocationX,subFrameLocationY)))
                            {
                            subFrameLocationX = mainScreenSize.width/2 - subFrame.getWidth()/2;
                            subFrameLocationY = mainScreenSize.height/2 - subFrame.getHeight()/2;
                            }
                        subFrame.setLocation(subFrameLocationX, subFrameLocationY);
                        }

                if (openDataSetWindow || setSubpanelVisible)
                        {
                        subFrame.setVisible (true);
                        }
                }

//        static void constructTopGroupTopLabels(JPanel mainsubpanelgroup)
//                {
//                JLabel dummylabel120 = new JLabel ("Data");
//                dummylabel120.setHorizontalAlignment(JLabel.CENTER);
//                dummylabel120.setPreferredSize(new Dimension(10,15));
//                mainsubpanelgroup.add (dummylabel120);
//                JLabel dummylabel121 = new JLabel ("Plot");
//                dummylabel121.setToolTipText("Show the dataset on the plot");
//                dummylabel121.setHorizontalAlignment(JLabel.CENTER);
//                dummylabel121.setPreferredSize(new Dimension(10,15));
//                mainsubpanelgroup.add (dummylabel121);
//                JLabel autoscaletopincludelabel = new JLabel ("Auto");
//                autoscaletopincludelabel.setToolTipText("Include this dataset in the Y-axis autoscale range");
//                autoscaletopincludelabel.setHorizontalAlignment(JLabel.CENTER);
//                autoscaletopincludelabel.setPreferredSize(new Dimension(10,15));
//                mainsubpanelgroup.add (autoscaletopincludelabel);
//                JLabel mmagtoplabel = new JLabel ("");
//                mmagtoplabel.setToolTipText("Plot data in relative magnitude");
//                mmagtoplabel.setHorizontalAlignment(JLabel.CENTER);
//                mmagtoplabel.setPreferredSize(new Dimension(10,15));
//                mainsubpanelgroup.add (mmagtoplabel);
//                JLabel xcolumntoplabelsub = new JLabel ("");
//                xcolumntoplabelsub.setHorizontalAlignment(JLabel.CENTER);
//                xcolumntoplabelsub.setPreferredSize(new Dimension(10,15));
//                mainsubpanelgroup.add (xcolumntoplabelsub);
//                JLabel ycolumntoplabelsub = new JLabel ("");
//                ycolumntoplabelsub.setHorizontalAlignment(JLabel.CENTER);
//                ycolumntoplabelsub.setPreferredSize(new Dimension(10,15));
//                mainsubpanelgroup.add (ycolumntoplabelsub);
////                JLabel useerrortoplabel = new JLabel ("Use");
////                useerrortoplabel.setToolTipText("Plot Y-error bars using Y-error Column values");
////                useerrortoplabel.setHorizontalAlignment(JLabel.CENTER);
////                useerrortoplabel.setPreferredSize(new Dimension(10,15));
////                mainsubpanelgroup.add (useerrortoplabel);
//                JLabel errortoplabel = new JLabel ("");
//                errortoplabel.setHorizontalAlignment(JLabel.CENTER);
//                errortoplabel.setToolTipText("Error associated with Y-data");
//                errortoplabel.setPreferredSize(new Dimension(10,15));
//                mainsubpanelgroup.add (errortoplabel);
//                JLabel operatortoplabelsub = new JLabel ("");
//                operatortoplabelsub.setHorizontalAlignment(JLabel.CENTER);
//                operatortoplabelsub.setToolTipText("Performs the selected operation on Y-data Column and Y-operator Column and propagates error if enabled");
//                operatortoplabelsub.setPreferredSize(new Dimension(10,15));
//                mainsubpanelgroup.add (operatortoplabelsub);
//                JLabel opcolumntoplabelsub = new JLabel ("");
//                opcolumntoplabelsub.setHorizontalAlignment(JLabel.CENTER);
//                opcolumntoplabelsub.setPreferredSize(new Dimension(10,15));
//                mainsubpanelgroup.add (opcolumntoplabelsub);
////                JLabel useoperrortoplabel = new JLabel ("Use");
////                useoperrortoplabel.setToolTipText("Use Operator Error as error associated with Y-operator Column");
////                useoperrortoplabel.setPreferredSize(new Dimension(10,15));
////                useoperrortoplabel.setHorizontalAlignment(JLabel.CENTER);
////                mainsubpanelgroup.add (useoperrortoplabel);
//                JLabel operrortoplabel = new JLabel ("");
//                operrortoplabel.setHorizontalAlignment(JLabel.CENTER);
//                operrortoplabel.setPreferredSize(new Dimension(10,15));
//                mainsubpanelgroup.add (operrortoplabel);
//                JLabel markertoplabelsub = new JLabel ("");
//                markertoplabelsub.setHorizontalAlignment(JLabel.CENTER);
//                markertoplabelsub.setPreferredSize(new Dimension(10,15));
//                mainsubpanelgroup.add (markertoplabelsub);
//                JLabel markercolortoplabelsub = new JLabel ("");
//                markercolortoplabelsub.setHorizontalAlignment(JLabel.CENTER);
//                markercolortoplabelsub.setPreferredSize(new Dimension(10,15));
//                mainsubpanelgroup.add (markercolortoplabelsub);
//                JLabel linestoplabel = new JLabel ("Use");
//                linestoplabel.setToolTipText("Connect datapoints with lines");
//                linestoplabel.setHorizontalAlignment(JLabel.CENTER);
//                linestoplabel.setPreferredSize(new Dimension(10,15));
//                mainsubpanelgroup.add (linestoplabel);
//                }

        static void constructTopGroupBottomLabels(JPanel mainsubpanelgroup)
                {
                JLabel dummylabel20 = new JLabel ("<HTML><CENTER>Data<BR><CENTER>Set</HTML>");
                dummylabel20.setHorizontalAlignment(JLabel.CENTER);
                mainsubpanelgroup.add (dummylabel20);
                JLabel dummylabel21 = new JLabel ("Plot");
                dummylabel21.setToolTipText("Show the dataset on the plot");
                dummylabel21.setHorizontalAlignment(JLabel.CENTER);
                mainsubpanelgroup.add (dummylabel21);
                JLabel autoscaleincludelabel = new JLabel ("<HTML><CENTER>Auto<BR><CENTER>Scale</HTML>");
                autoscaleincludelabel.setToolTipText("Include this dataset in the Y-axis autoscale range");
                autoscaleincludelabel.setHorizontalAlignment(JLabel.CENTER);
                mainsubpanelgroup.add (autoscaleincludelabel);
                JLabel mmaglabel = new JLabel ("Mag");
                mmaglabel.setToolTipText("Plot data in relative magnitude");
                mmaglabel.setHorizontalAlignment(JLabel.CENTER);
                mainsubpanelgroup.add (mmaglabel);
                xcolumnlabelsub = new JLabel ("X-data");
                xcolumnlabelsub.setHorizontalAlignment(JLabel.CENTER);
                mainsubpanelgroup.add (xcolumnlabelsub);
//                ycolumnlabelsub = new JLabel ("Y-data");
//                ycolumnlabelsub.setHorizontalAlignment(JLabel.CENTER);
//                mainsubpanelgroup.add (ycolumnlabelsub);
                
                JPanel ycolumnlabelgroup = new JPanel();
                ycolumnlabelgroup.setLayout(new BoxLayout(ycolumnlabelgroup,BoxLayout.X_AXIS));
//                legendslabelgroup.add(Box.createGlue());
                ycolumnlabelgroup.add(Box.createHorizontalGlue());

                ycolumnlabelsub = new JLabel ("Y-data");
                ycolumnlabelsub.setHorizontalAlignment(JLabel.CENTER);
                ycolumnlabelgroup.add (ycolumnlabelsub);
                ycolumnlabelgroup.add(Box.createHorizontalStrut(5));
//                legendslabelgroup.add(Box.createGlue());
                helpIcon = createImageIcon("astroj/images/help.png", "Multi-plot Help");
                JButton ycolumnlabelconfigureButton = new JButton(helpIcon);
                ycolumnlabelconfigureButton.setToolTipText("Data naming convention help");
                ycolumnlabelconfigureButton.setMargin(new Insets(0,0,0,0));
                ycolumnlabelconfigureButton.addActionListener(new ActionListener(){
                            public void actionPerformed(ActionEvent ae) {
                                openDataHelpPanel();}});
                ycolumnlabelgroup.add(ycolumnlabelconfigureButton);
                ycolumnlabelgroup.add(Box.createHorizontalGlue());
//                legendslabelgroup.add(Box.createGlue());
                mainsubpanelgroup.add(ycolumnlabelgroup);                
                
                
                
                JLabel useerrorlabel = new JLabel ("<HTML><CENTER>Auto<BR><CENTER>Error</HTML>");
                useerrorlabel.setToolTipText("<HTML>Show automatic Y-error bars when dataset has predefined error available.<BR>"+
                                             "Compatible datasets are: Source-Sky_XX, rel_flux_XX, tot_C_cnts, tot_C_cnts-XX.</HTML>");
                useerrorlabel.setHorizontalAlignment(JLabel.CENTER);
                mainsubpanelgroup.add (useerrorlabel);
//                JLabel errorlabel = new JLabel ("Y-error");
//                errorlabel.setHorizontalAlignment(JLabel.CENTER);
//                errorlabel.setToolTipText("Error associated with Y-data");
//                mainsubpanelgroup.add (errorlabel);
                operatorlabelsub = new JLabel ("Function");
                operatorlabelsub.setHorizontalAlignment(JLabel.CENTER);
                operatorlabelsub.setToolTipText("Performs the selected operation on Y-data Column and Y-operator Column and propagates error if enabled");
                mainsubpanelgroup.add (operatorlabelsub);
                opcolumnlabelsub = new JLabel ("Y-operand");
                opcolumnlabelsub.setHorizontalAlignment(JLabel.CENTER);
                mainsubpanelgroup.add (opcolumnlabelsub);
//                JLabel useoperrorlabel = new JLabel ("Use");
//                useoperrorlabel.setToolTipText("Use Operator Error as error associated with Y-operator Column");
//                useoperrorlabel.setHorizontalAlignment(JLabel.CENTER);
//                mainsubpanelgroup.add (useoperrorlabel);
//                JLabel operrorlabel = new JLabel ("Y-op Error");
//                operrorlabel.setHorizontalAlignment(JLabel.CENTER);
//                mainsubpanelgroup.add (operrorlabel);
                markerlabelsub = new JLabel ("Symbol");
                markerlabelsub.setHorizontalAlignment(JLabel.CENTER);
                mainsubpanelgroup.add (markerlabelsub);
                markercolorlabelsub = new JLabel ("Color");
                markercolorlabelsub.setHorizontalAlignment(JLabel.CENTER);
                mainsubpanelgroup.add (markercolorlabelsub);
                JLabel lineslabel = new JLabel ("Lines");
                lineslabel.setToolTipText("Connect datapoints with lines");
                lineslabel.setHorizontalAlignment(JLabel.CENTER);
                mainsubpanelgroup.add (lineslabel);
                }

//        static void constructOtherGroupTopLabels(JPanel mainsubpanelgroup)
//                {
//                JLabel seconddatasettoplabel = new JLabel ("Data");
//                seconddatasettoplabel.setHorizontalAlignment(JLabel.CENTER);
//                if (!useWideDataPanel)
//                        mainsubpanelgroup.add (seconddatasettoplabel);
//                JLabel binstoplabel = new JLabel ("Samples");
//                binstoplabel.setHorizontalAlignment(JLabel.CENTER);
//                binstoplabel.setToolTipText("Plot binned data using the number of samples/bin indicated");
//                mainsubpanelgroup.add (binstoplabel);
//
//                JLabel autoscaletoplabel1 = new JLabel ("Scale Within Plot Range");
//                autoscaletoplabel1.setToolTipText("Auto shift dataset to center of plot and scale to 100% of plot range");
//                autoscaletoplabel1.setHorizontalAlignment(JLabel.CENTER);
//                mainsubpanelgroup.add(autoscaletoplabel1);
//
//                JLabel dummylabel139 = new JLabel ("");
//                mainsubpanelgroup.add (dummylabel139);
//
//                JLabel customscaletoplabel1 = new JLabel ("Scale Curve Directly");
//                customscaletoplabel1.setToolTipText("Directly scale and shift dataset");
//                customscaletoplabel1.setHorizontalAlignment(JLabel.CENTER);
//                mainsubpanelgroup.add(customscaletoplabel1);
//
//                JLabel morelegendstoplabel = new JLabel ("Legend Type");
//                morelegendstoplabel.setHorizontalAlignment(JLabel.CENTER);
//                morelegendstoplabel.setComponentPopupMenu(legendpopup);
//                morelegendstoplabel.setToolTipText("Right click to set legend preferences");
//                mainsubpanelgroup.add (morelegendstoplabel);
//
//                JLabel morelegendsnametoplabel = new JLabel ("");
//                morelegendsnametoplabel.setHorizontalAlignment(JLabel.CENTER);
//                morelegendsnametoplabel.setComponentPopupMenu(legendpopup);
//                morelegendsnametoplabel.setToolTipText("Right click to set legend preferences");
//                mainsubpanelgroup.add (morelegendsnametoplabel);
//
//                if (useWideDataPanel)
//                        mainsubpanelgroup.add (seconddatasettoplabel);
//                }

        static void constructOtherGroupBottomLabels(JPanel mainsubpanelgroup)
                {
                JLabel seconddatasetlabel = new JLabel ("<HTML><CENTER>Data<BR><CENTER>Set</HTML>");
                seconddatasetlabel.setHorizontalAlignment(JLabel.CENTER);
                seconddatasetlabel.setMaximumSize(new Dimension(25,25));
                if (!useWideDataPanel)
                        mainsubpanelgroup.add (seconddatasetlabel);
                JLabel binslabel = new JLabel ("Bin Size");
                binslabel.setMaximumSize(new Dimension(50, 25));
                binslabel.setPreferredSize(new Dimension(50, 25));
                binslabel.setHorizontalAlignment(JLabel.CENTER);
                binslabel.setToolTipText("Plot binned data using the number of samples/bin indicated");
                mainsubpanelgroup.add (binslabel);
//                JPanel autoscalelabelgroup = new JPanel();
//                autoscalelabelgroup.setLayout(new BoxLayout(autoscalelabelgroup,BoxLayout.X_AXIS));
//                JLabel autoscalelabel1 = new JLabel ("Auto");
//                autoscalelabel1.setToolTipText("Auto shift dataset to center of plot and scale to 100% of plot range");
//                autoscalelabel1.setHorizontalAlignment(JLabel.LEFT);
//                autoscalelabelgroup.add(autoscalelabel1);
//                autoscalelabelgroup.add(Box.createHorizontalGlue());
//                JLabel autoscalelabel2 = new JLabel ("Scale%");
//                autoscalelabel2.setToolTipText("Scale dataset to specified % of plot range");
//                autoscalelabel2.setHorizontalAlignment(JLabel.CENTER);
//                autoscalelabelgroup.add(autoscalelabel2);
//                autoscalelabelgroup.add(Box.createHorizontalStrut(7));
//                autoscalelabelgroup.add(Box.createHorizontalGlue());
//                JLabel autoscalelabel3 = new JLabel ("Shift%");
//                autoscalelabel3.setToolTipText("Shift dataset from plot center by specified % of plot range");
//                autoscalelabel3.setHorizontalAlignment(JLabel.CENTER);
//                autoscalelabelgroup.add(autoscalelabel3);
//                autoscalelabelgroup.add(Box.createHorizontalStrut(7));
//                mainsubpanelgroup.add (autoscalelabelgroup);



                JPanel customscalelabelgroup = new JPanel();
                
                customscalelabelgroup.setLayout(new BoxLayout(customscalelabelgroup,BoxLayout.X_AXIS));
                customscalelabelgroup.setMaximumSize(new Dimension(195,20));
//                JLabel dummylabel39 = new JLabel ("");
//                customscalelabelgroup.add (dummylabel39);
                customscalelabelgroup.add(Box.createHorizontalStrut(18));
//                customscalelabelgroup.add(Box.createHorizontalGlue());
                JLabel customscalelabel1 = new JLabel ("<HTML><CENTER>Scale<BR><CENTER>Rel</HTML>");
                customscalelabel1.setToolTipText("Select to arbitrarily scale data to fit on plot (\"Scale\" and \"Shift\" by a percentage of the plot Y-range)");
//                customscalelabel1.setHorizontalAlignment(JLabel.LEFT);
                customscalelabelgroup.add(customscalelabel1);
                customscalelabelgroup.add(Box.createHorizontalGlue());
                
                JLabel customscalelabel2 = new JLabel ("Scale");
                customscalelabel2.setToolTipText("Scale dataset by specified value or if \"Scale Rel\" is selected, by specified percentage of the plot Y-range");
//                customscalelabel2.setHorizontalAlignment(JLabel.CENTER);
                customscalelabelgroup.add(customscalelabel2);
                customscalelabelgroup.add(Box.createHorizontalStrut(25));
                
                JLabel customscalelabel3 = new JLabel ("then Shift");
                customscalelabel3.setToolTipText("Add specified constant value to scaled dataset or if \"Scale Rel\" is selected, add specified percentage of the plot Y-range");
//                customscalelabel3.setHorizontalAlignment(JLabel.CENTER);
                customscalelabelgroup.add(customscalelabel3);
                customscalelabelgroup.add(Box.createHorizontalStrut(20));
                mainsubpanelgroup.add (customscalelabelgroup);



                JPanel legendslabelgroup = new JPanel();
                legendslabelgroup.setLayout(new BoxLayout(legendslabelgroup,BoxLayout.X_AXIS));
                legendslabelgroup.setMaximumSize(new Dimension(75,20));
//                legendslabelgroup.add(Box.createGlue());
                legendslabelgroup.add(Box.createHorizontalStrut(1));

                JLabel morelegendslabel = new JLabel("<HTML><CENTER>Legend<BR><CENTER>Type</HTML>");
                morelegendslabel.setHorizontalAlignment(JLabel.CENTER);
                morelegendslabel.setComponentPopupMenu(legendpopup);
                morelegendslabel.setToolTipText("Right click to set legend preferences");
                legendslabelgroup.add (morelegendslabel);
                legendslabelgroup.add(Box.createHorizontalStrut(5));
//                legendslabelgroup.add(Box.createGlue());

                legendslabelconfigureButton = new JButton(configureIcon);
                legendslabelconfigureButton.setToolTipText("Set legend preferences");
                legendslabelconfigureButton.setMargin(new Insets(0,0,0,0));
                legendslabelconfigureButton.addActionListener(new ActionListener(){
                            public void actionPerformed(ActionEvent ae) {
                                legendpopup.show(legendslabelconfigureButton, mainsubpanel.getX() , mainsubpanel.getY()+25);}});
                legendslabelgroup.add(legendslabelconfigureButton);
                legendslabelgroup.add(Box.createHorizontalStrut(1));
//                legendslabelgroup.add(Box.createGlue());
                mainsubpanelgroup.add(legendslabelgroup);

                JLabel morelegendsnamelabel = new JLabel ("Custom Legend");
                morelegendsnamelabel.setHorizontalAlignment(JLabel.CENTER);
                morelegendsnamelabel.setComponentPopupMenu(legendpopup);
                morelegendsnamelabel.setToolTipText("Right click to set legend preferences");
                mainsubpanelgroup.add (morelegendsnamelabel);

                if (useWideDataPanel)
                        mainsubpanelgroup.add (seconddatasetlabel);
                }

        static void constructTopGroup(JPanel mainsubpanelgroup, final int c)
                {

                        curvelabel[c] = new JLabel (""+(c+1));
                        curvelabel[c].setHorizontalAlignment(JLabel.CENTER);
                        mainsubpanelgroup.add (curvelabel[c]);

                        usecurvebox[c] = new JCheckBox ("", plotY[c]);
                        usecurvebox[c].addItemListener(new ItemListener(){
                            public void itemStateChanged(ItemEvent e) {
                                if (e.getStateChange() == ItemEvent.DESELECTED)
                                        plotY[c] = false;
                                else if (e.getStateChange() == ItemEvent.SELECTED)
                                        plotY[c] = true;
                                updatePlot();}});
                        usecurvebox[c].setHorizontalAlignment(JLabel.CENTER);
                        mainsubpanelgroup.add (usecurvebox[c]);

                        autoscaleincludebox[c] = new JCheckBox ("", ASInclude[c]);
                        autoscaleincludebox[c].addItemListener(new ItemListener(){
                            public void itemStateChanged(ItemEvent e) {
                                if (e.getStateChange() == ItemEvent.DESELECTED)
                                        ASInclude[c] = false;
                                else if (e.getStateChange() == ItemEvent.SELECTED)
                                        ASInclude[c] = true;
                                updatePlot();}});
                        autoscaleincludebox[c].setHorizontalAlignment(JLabel.CENTER);
                        mainsubpanelgroup.add (autoscaleincludebox[c]);

                        usemmagbox[c] = new JCheckBox ("", mmag[c]);
                        usemmagbox[c].addItemListener(new ItemListener(){
                            public void itemStateChanged(ItemEvent e) {
                                if (e.getStateChange() == ItemEvent.DESELECTED)
                                        mmag[c] = false;
                                else if (e.getStateChange() == ItemEvent.SELECTED)
                                        mmag[c] = true;
                                updatePlot();}});
                        usemmagbox[c].setHorizontalAlignment(JLabel.CENTER);
                        mainsubpanelgroup.add (usemmagbox[c]);

                        xdatacolumn[c] = new JComboBox(columnswd);
                        xdatacolumn[c].setSelectedItem(xlabel[c]);
                        xdatacolumn[c].setPrototypeDisplayValue("123456789012345678");
                        xdatacolumn[c].addActionListener(new ActionListener(){
                            public void actionPerformed(ActionEvent ae) {
                                xlabel[c] = (String) xdatacolumn[c].getSelectedItem();
                                updatePlot();}});
                        mainsubpanelgroup.add (xdatacolumn[c]);

                        ydatacolumn[c] = new JComboBox(columns);
                        ydatacolumn[c].setSelectedItem(ylabel[c]);
                        ydatacolumn[c].setPrototypeDisplayValue("123456789012345678");
                        ydatacolumn[c].addActionListener(new ActionListener(){
                            public void actionPerformed(ActionEvent ae) {
                                ylabel[c] = (String) ydatacolumn[c].getSelectedItem();
                                usecurvebox[c].setSelected(true);
                                plotY[c] = true;
                                updatePlot();}});
                        mainsubpanelgroup.add (ydatacolumn[c]);

                        errorcolumnbox[c] = new JCheckBox ("", showErrors[c]);
                        errorcolumnbox[c].setToolTipText("<HTML>Show automatic Y-error bars when data has predefined error available.<BR>"+
                                                         "Compatible data columns are: Source-Sky_XX, rel_flux_XX, tot_C_cnts, tot_C_cnts-XX.</HTML>");                        
                        errorcolumnbox[c].addItemListener(new ItemListener(){
                            public void itemStateChanged(ItemEvent e) {
                                if (e.getStateChange() == ItemEvent.DESELECTED)
                                        showErrors[c] = false;
                                else if (e.getStateChange() == ItemEvent.SELECTED)
                                    {
                                    if (operatorIndex[c] == 6)
                                        {
                                        operatorIndex[c] = 0;
                                        operatorselection[c].setSelectedIndex(operatorIndex[c]);
                                        }
                                    showErrors[c] = true;
                                    }
                                updatePlot(); }});
                        errorcolumnbox[c].setHorizontalAlignment(JLabel.CENTER);
                        mainsubpanelgroup.add (errorcolumnbox[c]);

//                        errorcolumn[c] = new JComboBox(columnsErr);
//                        errorcolumn[c].setSelectedItem(errlabel[c]);
//                        errorcolumn[c].setPrototypeDisplayValue("123456789012345678");
//                        errorcolumn[c].addActionListener(new ActionListener(){
//                            public void actionPerformed(ActionEvent ae) {
//                                errlabel[c] = (String) errorcolumn[c].getSelectedItem();
//                                updatePlot();}});
//                        mainsubpanelgroup.add (errorcolumn[c]);

                        operatorselection[c] = new JComboBox(operators);
                        operatorselection[c].setSelectedIndex(operatorIndex[c]);
                        operatorselection[c].setPrototypeDisplayValue("123456789012");
                        operatorselection[c].addActionListener(new ActionListener(){
                            public void actionPerformed(ActionEvent ae) {
                                operatorIndex[c] = operatorselection[c].getSelectedIndex();
                                if (operatorIndex[c] == 6)
                                    {
                                    showErrors[c] = false;
                                    errorcolumnbox[c].setSelected(false);
                                    }
                                updatePlot();}});
                        mainsubpanelgroup.add (operatorselection[c]);

                        operatorcolumn[c] = new JComboBox(columns);
                        operatorcolumn[c].setSelectedItem(oplabel[c]);
                        operatorcolumn[c].setPrototypeDisplayValue("123456789012345678");
                        operatorcolumn[c].addActionListener(new ActionListener(){
                            public void actionPerformed(ActionEvent ae) {
                                oplabel[c] = (String) operatorcolumn[c].getSelectedItem();
                                updatePlot();}});
                        mainsubpanelgroup.add (operatorcolumn[c]);


//                        operrorcolumnbox[c] = new JCheckBox ("", showOpErrors[c]);
//                        operrorcolumnbox[c].addItemListener(new ItemListener(){
//                            public void itemStateChanged(ItemEvent e) {
//                                if (e.getStateChange() == ItemEvent.DESELECTED)
//                                        showOpErrors[c] = false;
//                                else if (e.getStateChange() == ItemEvent.SELECTED)
//                                        showOpErrors[c] = true;
//                                updatePlot(); }});
//                        operrorcolumnbox[c].setHorizontalAlignment(JLabel.CENTER);
//                        mainsubpanelgroup.add (operrorcolumnbox[c]);

//                        operrorcolumn[c] = new JComboBox(columnsErr);
//                        operrorcolumn[c].setSelectedItem(operrlabel[c]);
//                        operrorcolumn[c].setPrototypeDisplayValue("123456789012345678");
//                        operrorcolumn[c].addActionListener(new ActionListener(){
//                            public void actionPerformed(ActionEvent ae) {
//                                operrlabel[c] = (String) operrorcolumn[c].getSelectedItem();
//                                updatePlot();}});
//                        mainsubpanelgroup.add (operrorcolumn[c]);

                        markersymbolselection[c] = new JComboBox(markers);
//                        markersymbolselection[c].setPreferredSize(new Dimension(100, 25));
                        markersymbolselection[c].setSelectedIndex(markerIndex[c]);
                        markersymbolselection[c].setPrototypeDisplayValue("12345");
                        markersymbolselection[c].addActionListener(new ActionListener(){
                            public void actionPerformed(ActionEvent ae) {
                                markerIndex[c] = markersymbolselection[c].getSelectedIndex();
                                updatePlot();}});
                        mainsubpanelgroup.add (markersymbolselection[c]);

                        markercolorselection[c] = new JComboBox(colors);
//                        markercolorselection[c].setPreferredSize(new Dimension(100, 25));
                        markercolorselection[c].setSelectedIndex(colorIndex[c]);
                        markercolorselection[c].setPrototypeDisplayValue("12345678");
                        markercolorselection[c].addActionListener(new ActionListener(){
                            public void actionPerformed(ActionEvent ae) {
                                colorIndex[c] = markercolorselection[c].getSelectedIndex();
                                updatePlot();}});
                        mainsubpanelgroup.add (markercolorselection[c]);



                        uselinesbox[c] = new JCheckBox ("", lines[c]);
                        uselinesbox[c].addItemListener(new ItemListener(){
                            public void itemStateChanged(ItemEvent e) {
                                if (e.getStateChange() == ItemEvent.DESELECTED)
                                        lines[c] = false;
                                else if (e.getStateChange() == ItemEvent.SELECTED)
                                        lines[c] = true;
                                updatePlot();}});
                        uselinesbox[c].setHorizontalAlignment(JLabel.CENTER);
                        mainsubpanelgroup.add (uselinesbox[c]);

                }

        static void constructOtherGroup(JPanel mainsubpanelgroup, final int c)
                {
                if (!useWideDataPanel)
                        {
                        othercurvelabel[c] = new JLabel (""+(c+1));
                        othercurvelabel[c].setHorizontalAlignment(JLabel.CENTER);
                        mainsubpanelgroup.add (othercurvelabel[c]);
                        }

                binsizespinnermodel[c] = new SpinnerNumberModel(new Integer(binSize[c]),new Integer(1), null, new Integer(1));
                binsizespinner[c] = new JSpinner(binsizespinnermodel[c]);
                binsizespinner[c].setMaximumSize(new Dimension(50,25));
                binsizespinner[c].setPreferredSize(new Dimension(50,25));
                binsizespinner[c].addChangeListener (new ChangeListener()
                    {
                    public void stateChanged(ChangeEvent ev)
                        {
                        binSize[c] = ((Integer)binsizespinner[c].getValue()).intValue();
                        Prefs.set("plot.binSize"+c,binSize[c]);
                        updatePlot();
                        }
                    });
                binsizespinner[c].addMouseWheelListener( new MouseWheelListener()
                    {
                    public void mouseWheelMoved( MouseWheelEvent e )
                        {
                        int newValue =  new Integer(((Integer)binsizespinner[c].getValue()).intValue() - e.getWheelRotation());
                        if (newValue > 0) binsizespinner[c].setValue(newValue);
                        }
                    });
                mainsubpanelgroup.add (binsizespinner[c]);

//                autoscalepanelgroup[c] = new JPanel (new SpringLayout());
//                autoscalepanelgroup[c].setBorder(BorderFactory.createLineBorder(Color.lightGray));
//                customscalepanelgroup[c] = new JPanel (new SpringLayout());
//                customscalepanelgroup[c].setBorder(BorderFactory.createLineBorder(Color.lightGray));
//                autoscaleButton[c] = new JRadioButton("");
//                customscaleButton[c] = new JRadioButton("");
//                if (autoScale[c])
//                        {
//                        autoscaleButton[c].setSelected(true);
//                        customscaleButton[c].setSelected(false);
//                        }
//                else
//                        {
//                        autoscaleButton[c].setSelected(false);
//                        customscaleButton[c].setSelected(true);
//                        }
//
//                autoscalepanelgroup[c].add (autoscaleButton[c]);
//                customscalepanelgroup[c].add (customscaleButton[c]);
//                autoscaleRadioGroup[c] = new ButtonGroup();
//                autoscaleRadioGroup[c].add(autoscaleButton[c]);
//                autoscaleRadioGroup[c].add(customscaleButton[c]);
//
//                autoscaleButton[c].addActionListener(new ActionListener(){
//                    public void actionPerformed(ActionEvent ae) {
//                        autoScale[c] = true;
//                        updatePlot();}});
//                customscaleButton[c].addActionListener(new ActionListener(){
//                    public void actionPerformed(ActionEvent ae) {
//                        autoScale[c] = false;
//                        updatePlot();}});
//
//        //AUTO SCALE SPINNER
//
//                autoscalemodel[c] = new SpinnerNumberModel(new Double(autoScaleFactor[c]*100.), null, null, new Double(autoScaleStep[c]));
//                autoscalespinner[c] = new JSpinner(autoscalemodel[c]);
//                autoscalespinner[c].setToolTipText("Right click to set spinner stepsize");
//                autoscalespinner[c].setPreferredSize(new Dimension(75,25));
//                autoscalespinner[c].addChangeListener (new ChangeListener()
//                    {
//                    public void stateChanged(ChangeEvent ev)
//                        {
//                        autoScaleFactor[c]= ((Double)autoscalespinner[c].getValue()).doubleValue()/100. ;
//                        updatePlot();
//                        }
//                    });
//                autoscalespinner[c].addMouseWheelListener( new MouseWheelListener()
//                    {
//                    public void mouseWheelMoved( MouseWheelEvent e )
//                        {
//                        autoscalespinner[c].setValue(((Double)autoscalespinner[c].getValue()).doubleValue()
//                                - e.getWheelRotation()*autoScaleStep[c]);
//                        }
//                    });
//
//
//                autoscalepanelgroup[c].add (autoscalespinner[c]);
//
//                //AUTO SCALE SPINNER STEPSIZE POPUP
//
//                autoscalesteppopup[c] = new JPopupMenu();
//                autoscalesteppanel[c] = new JPanel();
//                autoscalestepmodel[c] = new SpinnerListModel(spinnerscalelist);
//                autoscalestepspinner[c] = new JSpinner(autoscalestepmodel[c]);
//                autoscalestepspinner[c].setValue(convertToText(autoScaleStep[c]));
//                autoscalestepspinner[c].addChangeListener (new ChangeListener()
//                    {
//                    public void stateChanged(ChangeEvent ev)
//                        {
//                        autoScaleStep[c] = Double.parseDouble((String)autoscalestepspinner[c].getValue());
//                        autoscalespinner[c].setModel(new SpinnerNumberModel(new Double(autoScaleFactor[c])*100., null, null, new Double(autoScaleStep[c])));
//                        Prefs.set("plot.autoScaleStep"+c,autoScaleStep[c]);
//                        }
//                    });
//                autoscalesteplabel[c] = new JLabel ("Stepsize:");
//                autoscalesteppanel[c].add(autoscalesteplabel[c]);
//                autoscalesteppanel[c].add(autoscalestepspinner[c]);
//                autoscalesteppopup[c].add(autoscalesteppanel[c]);
//                autoscalesteppopup[c].addPopupMenuListener (new PopupMenuListener() {
//                    public void popupMenuCanceled(PopupMenuEvent popupMenuEvent) { }
//                    public void popupMenuWillBecomeInvisible(PopupMenuEvent popupMenuEvent) { }
//                    public void popupMenuWillBecomeVisible(PopupMenuEvent popupMenuEvent) { } });
//                autoscalespinner[c].setComponentPopupMenu(autoscalesteppopup[c]);
//
//        //AUTO SHIFT SPINNER
//
//                autoshiftmodel[c] = new SpinnerNumberModel(new Double(autoShiftFactor[c]*100.), null, null, new Double(autoShiftStep[c]));
//                autoshiftspinner[c] = new JSpinner(autoshiftmodel[c]);
//                autoshiftspinner[c].setPreferredSize(new Dimension(75,25));
//                autoshiftspinner[c].setToolTipText("Right click to set spinner stepsize");
//                autoshiftspinner[c].addChangeListener (new ChangeListener()
//                    {
//                    public void stateChanged(ChangeEvent ev)
//                        {
//                        autoShiftFactor[c]= ((Double)autoshiftspinner[c].getValue()).doubleValue()/100. ;
//                        updatePlot();
//                        }
//                    });
//                autoshiftspinner[c].addMouseWheelListener( new MouseWheelListener()
//                    {
//                    public void mouseWheelMoved( MouseWheelEvent e )
//                        {
//                        autoshiftspinner[c].setValue(((Double)autoshiftspinner[c].getValue()).doubleValue()
//                                - e.getWheelRotation()*autoShiftStep[c]);
//                        }
//                    });
//                autoscalepanelgroup[c].add (autoshiftspinner[c]);
//
//                //AUTO SHIFT SPINNER STEPSIZE POPUP
//
//                autoshiftsteppopup[c] = new JPopupMenu();
//                autoshiftsteppanel[c] = new JPanel();
//                autoshiftstepmodel[c] = new SpinnerListModel(spinnerscalelist);
//                autoshiftstepspinner[c] = new JSpinner(autoshiftstepmodel[c]);
//                autoshiftstepspinner[c].setValue(convertToText(autoShiftStep[c]));
//                autoshiftstepspinner[c].addChangeListener (new ChangeListener()
//                    {
//                    public void stateChanged(ChangeEvent ev)
//                        {
//                        autoShiftStep[c] = Double.parseDouble((String)autoshiftstepspinner[c].getValue());
//                        autoshiftspinner[c].setModel(new SpinnerNumberModel(new Double(autoShiftFactor[c])*100., null, null, new Double(autoShiftStep[c])));
//                        Prefs.set("plot.autoShiftStep"+c,autoShiftStep[c]);
//                        }
//                    });
//                autoshiftsteplabel[c] = new JLabel ("Stepsize:");
//                autoshiftsteppanel[c].add(autoshiftsteplabel[c]);
//                autoshiftsteppanel[c].add(autoshiftstepspinner[c]);
//                autoshiftsteppopup[c].add(autoshiftsteppanel[c]);
//                autoshiftsteppopup[c].addPopupMenuListener (new PopupMenuListener() {
//                    public void popupMenuCanceled(PopupMenuEvent popupMenuEvent) { }
//                    public void popupMenuWillBecomeInvisible(PopupMenuEvent popupMenuEvent) { }
//                    public void popupMenuWillBecomeVisible(PopupMenuEvent popupMenuEvent) { } });
//                autoshiftspinner[c].setComponentPopupMenu(autoshiftsteppopup[c]);

                Insets buttonMargin = new Insets(0,0,0,0);
                forceIcon = createImageIcon("astroj/images/grab.png", "Transfer Scale Rel settings to absolute settings");

                customscalepanelgroup[c] = new JPanel (new SpringLayout());
                customscalepanelgroup[c].setMaximumSize(new Dimension(195,20));
                customscalepanelgroup[c].setBorder(BorderFactory.createLineBorder(Color.LIGHT_GRAY));
//                customscalepanelgroup[c].setBorder(BorderFactory.createCompoundBorder(BorderFactory.createLineBorder(new Color(122,138,153)),BorderFactory.createLineBorder(new Color(184,207,229))));

        // GRAB "FORCE" SCALE AND SHIFT VALUES
//                grabautopanel[c] = new JPanel (new SpringLayout());

                grabautobutton[c] = new JButton(forceIcon);
                grabautobutton[c].setMargin(buttonMargin);
                grabautobutton[c].setToolTipText("Transfer \"Scale Rel\" scale and shift values to absolute values");
                grabautobutton[c].addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                        customScaleFactor[c] = autoScaleFactor[c]*yRange/yWidth[c];
//                        customscalespinner[c].setValue(customScaleFactor[c]);
                        customShiftFactor[c] = yMid + yRange*autoShiftFactor[c] - customScaleFactor[c]*yMidpoint[c];
//                        customshiftspinner[c].setValue(customShiftFactor[c]);
                        force[c] = false;
                        forcebox[c].setSelected(false);
                        customscalespinner[c].setModel(new SpinnerNumberModel(new Double(customScaleFactor[c]), null, null, new Double(customScaleStep[c])));
                        customscalestepspinner[c].setValue(convertToText(customScaleStep[c]));
                        customshiftspinner[c].setModel(new SpinnerNumberModel(new Double(customShiftFactor[c]), null, null, new Double(customShiftStep[c])));
                        customshiftstepspinner[c].setValue(convertToText(customShiftStep[c]));
                        Prefs.set("plot.force"+c,force[c]);
                        updatePlot();
                    }
                });
                customscalepanelgroup[c].add (grabautobutton[c]);
//                SpringUtil.makeCompactGrid (grabautopanel[c],1,1, 0,0,0,0);
//                mainsubpanelgroup.add (grabautopanel[c]);
//                customscalepanelgroup[c].add (grabautopanel[c]);

        //FORCE CHECKBOX

                forcebox[c] = new JCheckBox ("", force[c]);
                forcebox[c].setToolTipText("Select to arbitrarily scale data to fit on plot (\"Scale\" and \"Shift\" by a percentage of the plot Y-range)");
                forcebox[c].addItemListener(new ItemListener(){
                    public void itemStateChanged(ItemEvent e) {
                        if (e.getStateChange() == ItemEvent.DESELECTED)
                            force[c] = false;
                        else if (e.getStateChange() == ItemEvent.SELECTED)
                            force[c] = true;
                        customscalespinner[c].setModel(new SpinnerNumberModel(new Double(force[c]?autoScaleFactor[c]*100.0:customScaleFactor[c]), null, null, new Double(force[c]?autoScaleStep[c]:customScaleStep[c])));
                        customscalestepspinner[c].setValue(convertToText(force[c]?autoScaleStep[c]:customScaleStep[c]));
                        customshiftspinner[c].setModel(new SpinnerNumberModel(new Double(force[c]?autoShiftFactor[c]*100:customShiftFactor[c]), null, null, new Double(force[c]?autoShiftStep[c]:customShiftStep[c])));
                        customshiftstepspinner[c].setValue(convertToText(force[c]?autoShiftStep[c]:customShiftStep[c]));
                        Prefs.set("plot.force"+c,force[c]);
                        updatePlot();}});
//                forcebox[c].setHorizontalAlignment(JLabel.CENTER);
                customscalepanelgroup[c].add (forcebox[c]);

        //SCALE SPINNER

                customscalemodel[c] = new SpinnerNumberModel(new Double(force[c]?autoScaleFactor[c]*100:customScaleFactor[c]), null, null, new Double(force[c]?autoScaleStep[c]:customScaleStep[c]));
                customscalespinner[c] = new JSpinner(customscalemodel[c]);
                customscalespinner[c].setToolTipText("<HTML>If 'Scale Rel' is deselected, the Y-dataset values are multiplied by this factor<BR>"+
                                                     "(and by the factor of 10 specified in the 'Y-axis' box of the 'Multi-plot Main' panel).<BR>"+
                                                     "If 'Scale Rel' is selected, the dataset values are arbitrarily scaled to fit on a 'Scale' percentage of the Y-axis range.<BR>"+
                                                     "Right click to set spinner stepsize.</HTML>");
                customscalespinner[c].setPreferredSize(new Dimension(75,25));

                customscaleeditor[c] = (JSpinner.NumberEditor)customscalespinner[c].getEditor();
                customscaleformat[c] = customscaleeditor[c].getFormat();
                customscaleformat[c].setMaximumFractionDigits(9);

                customscalespinner[c].addChangeListener (new ChangeListener()
                    {
                    public void stateChanged(ChangeEvent ev)
                        {
                        if (force[c])
                            autoScaleFactor[c]= ((Double)customscalespinner[c].getValue()).doubleValue()/100.0 ;
                        else
                            customScaleFactor[c]= ((Double)customscalespinner[c].getValue()).doubleValue() ;
                        updatePlot();
                        }
                    });
                customscalespinner[c].addMouseWheelListener( new MouseWheelListener()
                    {
                    public void mouseWheelMoved( MouseWheelEvent e )
                        {
                        if (force[c])
                            customscalespinner[c].setValue(((Double)customscalespinner[c].getValue()).doubleValue()-e.getWheelRotation()*autoScaleStep[c]);
                        else
                            customscalespinner[c].setValue(((Double)customscalespinner[c].getValue()).doubleValue()-e.getWheelRotation()*customScaleStep[c]);
                        }
                    });
                customscalepanelgroup[c].add (customscalespinner[c]);

        //SCALE SPINNER STEPSIZE POPUP

                customscalesteppopup[c] = new JPopupMenu();
                customscalesteppanel[c] = new JPanel();
                customscalestepmodel[c] = new SpinnerListModel(spinnerscalelist);
                customscalestepspinner[c] = new JSpinner(customscalestepmodel[c]);
                customscalestepspinner[c].setValue(convertToText(force[c]?autoScaleStep[c]:customScaleStep[c]));
                customscalestepspinner[c].addChangeListener (new ChangeListener()
                    {
                    public void stateChanged(ChangeEvent ev)
                        {
                        if (force[c])
                            {
                            autoScaleStep[c] = Double.parseDouble((String)customscalestepspinner[c].getValue());
                            Prefs.set("plot.autoScaleStep"+c,autoScaleStep[c]);
                            }
                        else
                            {
                            customScaleStep[c] = Double.parseDouble((String)customscalestepspinner[c].getValue());
                            Prefs.set("plot.customScaleStep"+c,customScaleStep[c]);
                            }
                        customscalespinner[c].setModel(new SpinnerNumberModel(new Double(force[c]?autoScaleFactor[c]*100:customScaleFactor[c]), null, null, new Double(force[c]?autoScaleStep[c]:customScaleStep[c])));
                        customscaleeditor[c] = (JSpinner.NumberEditor)customscalespinner[c].getEditor();
                        customscaleformat[c] = customscaleeditor[c].getFormat();
                        customscaleformat[c].setMaximumFractionDigits(9);
                        }
                    });
                customscalesteplabel[c] = new JLabel ("Stepsize:");
                customscalesteppanel[c].add(customscalesteplabel[c]);
                customscalesteppanel[c].add(customscalestepspinner[c]);
                customscalesteppopup[c].add(customscalesteppanel[c]);
                customscalesteppopup[c].addPopupMenuListener (new PopupMenuListener() {
                    public void popupMenuCanceled(PopupMenuEvent popupMenuEvent) { }
                    public void popupMenuWillBecomeInvisible(PopupMenuEvent popupMenuEvent) { }
                    public void popupMenuWillBecomeVisible(PopupMenuEvent popupMenuEvent) { } });
                customscalespinner[c].setComponentPopupMenu(customscalesteppopup[c]);

        //SHIFT SPINNER

                customshiftmodel[c] = new SpinnerNumberModel(new Double(force[c]?autoShiftFactor[c]*100:customShiftFactor[c]), null, null,
                                                             new Double(force[c]?invertYAxisSign*autoShiftStep[c]:invertYAxisSign*customShiftStep[c]));
                customshiftspinner[c] = new JSpinner(customshiftmodel[c]);

                customshifteditor[c] = (JSpinner.NumberEditor)customshiftspinner[c].getEditor();
                customshiftformat[c] = customshifteditor[c].getFormat();
                customshiftformat[c].setMaximumFractionDigits(6);


                customshiftspinner[c].setPreferredSize(new Dimension(75,25));
                customshiftspinner[c].setToolTipText("<HTML>If 'Scale Rel' is deselected, the 'Shift' value is added to the scaled Y-dataset values before plotting.<BR>"+
                                                     "If 'Scale Rel' is selected, the dataset values are arbitrarily shifted by 'Shift' percentage of the Y-axis plot range.<BR>"+
                                                     "Right click to set spinner stepsize.</HTML>");
                customshiftspinner[c].addChangeListener (new ChangeListener()
                    {
                    public void stateChanged(ChangeEvent ev)
                        {
                        if (force[c])
                            autoShiftFactor[c]=((Double)customshiftspinner[c].getValue()).doubleValue()/100.0;
                        else
                            customShiftFactor[c]=((Double)customshiftspinner[c].getValue()).doubleValue();
                        updatePlot();
                        }
                    });
                customshiftspinner[c].addMouseWheelListener( new MouseWheelListener()
                    {
                    public void mouseWheelMoved( MouseWheelEvent e )
                        {
                        if (force[c])
                            customshiftspinner[c].setValue(((Double)customshiftspinner[c].getValue()).doubleValue() -
                                      invertYAxisSign*e.getWheelRotation()*autoShiftStep[c]);
                        else
                            customshiftspinner[c].setValue(((Double)customshiftspinner[c].getValue()).doubleValue() -
                                      invertYAxisSign*e.getWheelRotation()*customShiftStep[c]);
                        }
                    });
                customscalepanelgroup[c].add (customshiftspinner[c]);

          //SHIFT SPINNER STEPSIZE POPUP

                customshiftsteppopup[c] = new JPopupMenu();
                customshiftsteppanel[c] = new JPanel();
                customshiftstepmodel[c] = new SpinnerListModel(spinnerscalelist);
                customshiftstepspinner[c] = new JSpinner(customshiftstepmodel[c]);
                customshiftstepspinner[c].setValue(convertToText(force[c]?autoShiftStep[c]:customShiftStep[c]));
                customshiftstepspinner[c].addChangeListener (new ChangeListener()
                    {
                    public void stateChanged(ChangeEvent ev)
                        {
                        if (force[c])
                            {
                            autoShiftStep[c] = Double.parseDouble((String)customshiftstepspinner[c].getValue());
                            Prefs.set("plot.autoShiftStep"+c,autoShiftStep[c]);
                            }
                        else
                            {
                            customShiftStep[c] = Double.parseDouble((String)customshiftstepspinner[c].getValue());
                            Prefs.set("plot.customShiftStep"+c,customShiftStep[c]);
                            }
                        customshiftspinner[c].setModel(new SpinnerNumberModel(new Double(force[c]?autoShiftFactor[c]*100:customShiftFactor[c]), null, null,
                                                             new Double(force[c]?invertYAxisSign*autoShiftStep[c]:invertYAxisSign*customShiftStep[c])));

                        customshifteditor[c] = (JSpinner.NumberEditor)customshiftspinner[c].getEditor();
                        customshiftformat[c] = customshifteditor[c].getFormat();
                        customshiftformat[c].setMaximumFractionDigits(6);
                        }
                    });
                customshiftsteplabel[c] = new JLabel ("Stepsize:");
                customshiftsteppanel[c].add(customshiftsteplabel[c]);
                customshiftsteppanel[c].add(customshiftstepspinner[c]);
                customshiftsteppopup[c].add(customshiftsteppanel[c]);
                customshiftsteppopup[c].addPopupMenuListener (new PopupMenuListener() {
                    public void popupMenuCanceled(PopupMenuEvent popupMenuEvent) { }
                    public void popupMenuWillBecomeInvisible(PopupMenuEvent popupMenuEvent) { }
                    public void popupMenuWillBecomeVisible(PopupMenuEvent popupMenuEvent) { } });
                customshiftspinner[c].setComponentPopupMenu(customshiftsteppopup[c]);
//                SpringUtil.makeCompactGrid (autoscalepanelgroup[c], 1, 3, 0,0,0,0);
                SpringUtil.makeCompactGrid (customscalepanelgroup[c], 1, customscalepanelgroup[c].getComponentCount(), 0,0,0,0);
//                mainsubpanelgroup.add (autoscalepanelgroup[c]);
                mainsubpanelgroup.add (customscalepanelgroup[c]);

        //LEGEND RADIO GROUP AND TEXT FIELD

                morelegendradiopanelgroup[c] = new JPanel (new SpringLayout());
                morelegendradiopanelgroup[c].setMaximumSize(new Dimension(75,20));
                morelegendradiopanelgroup[c].setBorder(BorderFactory.createLineBorder(Color.lightGray));
                ImageIcon noneLegendIcon = createImageIcon("astroj/images/nonelegend.png", "Disable legend for this data set");
                ImageIcon columnLegendIcon = createImageIcon("astroj/images/columnlegend.png", "Use the column name as the legend for this data set");
                ImageIcon customLegendIcon = createImageIcon("astroj/images/customlegend.png", "Use a custom legend for this data set");
                legendnoneButton[c] = new JRadioButton(noneLegendIcon);
                if (!useLegend[c] && !useColumnName[c])
                    {
                    legendnoneButton[c].setSelected(true);
                    legendnoneButton[c].setBackground(Color.LIGHT_GRAY);
                    }
                else
                    {
                    legendnoneButton[c].setSelected(false);
                    legendnoneButton[c].setBackground(Color.WHITE);
                    }
                morelegendradiopanelgroup[c].add (legendnoneButton[c]);
                legendcolumnNameButton[c] = new JRadioButton(columnLegendIcon);
                if (useColumnName[c])
                    {
                    legendcolumnNameButton[c].setSelected(true);
                    legendcolumnNameButton[c].setBackground(Color.LIGHT_GRAY);
                    }
                else
                    {
                    legendcolumnNameButton[c].setSelected(false);
                    legendcolumnNameButton[c].setBackground(Color.WHITE);
                    }
                morelegendradiopanelgroup[c].add (legendcolumnNameButton[c]);
                legendcustomNameButton[c] = new JRadioButton(customLegendIcon);
                if (!useColumnName[c] && useLegend[c])
                    {
                    legendcustomNameButton[c].setSelected(true);
                    legendcustomNameButton[c].setBackground(Color.LIGHT_GRAY);
                    }
                else
                    {
                    legendcustomNameButton[c].setSelected(false);
                    legendcustomNameButton[c].setBackground(Color.WHITE);
                    }
                morelegendradiopanelgroup[c].add (legendcustomNameButton[c]);

                morelegendRadioGroup[c] = new ButtonGroup();
                morelegendRadioGroup[c].add(legendnoneButton[c]);
                morelegendRadioGroup[c].add(legendcolumnNameButton[c]);
                morelegendRadioGroup[c].add(legendcustomNameButton[c]);
                legendnoneButton[c].addActionListener(new ActionListener(){
                    public void actionPerformed(ActionEvent ae) {
                        useColumnName[c] = false;
                        useLegend[c] = false;
                        legendnoneButton[c].setBackground(Color.LIGHT_GRAY);
                        legendcolumnNameButton[c].setBackground(Color.WHITE);
                        legendcustomNameButton[c].setBackground(Color.WHITE);
                        updatePlot();}});
                legendcustomNameButton[c].addActionListener(new ActionListener(){
                    public void actionPerformed(ActionEvent ae) {
                        useColumnName[c] = false;
                        useLegend[c] = true;
                        legendnoneButton[c].setBackground(Color.WHITE);
                        legendcolumnNameButton[c].setBackground(Color.WHITE);
                        legendcustomNameButton[c].setBackground(Color.LIGHT_GRAY);
                        updatePlot();}});
                legendcolumnNameButton[c].addActionListener(new ActionListener(){
                    public void actionPerformed(ActionEvent ae) {
                        useColumnName[c] = true;
                        useLegend[c] = false;
                        legendnoneButton[c].setBackground(Color.WHITE);
                        legendcolumnNameButton[c].setBackground(Color.LIGHT_GRAY);
                        legendcustomNameButton[c].setBackground(Color.WHITE);
                        updatePlot();}});
                SpringUtil.makeCompactGrid (morelegendradiopanelgroup[c], 1, 3, 0,0,0,0);
                legendnoneButton[c].setComponentPopupMenu(legendpopup);
                legendnoneButton[c].setToolTipText("Disable Legend -- Right click to set legend preferences");
                legendcustomNameButton[c].setComponentPopupMenu(legendpopup);
                legendcustomNameButton[c].setToolTipText("Use Custom Legend -- Right click to set legend preferences");
                legendcolumnNameButton[c].setComponentPopupMenu(legendpopup);
                legendcolumnNameButton[c].setToolTipText("Use Column Name as Legend -- Right click to set legend preferences");

                mainsubpanelgroup.add (morelegendradiopanelgroup[c]);

                // LEGEND TEXT FIELD
                morelegendField[c] = new JTextField (legend[c]);
                morelegendField[c].setBorder (BorderFactory.createEmptyBorder(0,4,0,4));
                morelegendField[c].getDocument().addDocumentListener (new DocumentListener()
                    {
                    public void insertUpdate (DocumentEvent ev)
                        {
                        legend[c] = morelegendField[c].getText();
                        updatePlot();
                        }
                    public void removeUpdate (DocumentEvent ev)
                        {
                        legend[c] = morelegendField[c].getText();
                        updatePlot();
                        }
                    public void changedUpdate (DocumentEvent ev)
                        {
                        legend[c] = morelegendField[c].getText();
                        updatePlot();
                        }
                    });
                if (useWideDataPanel)
                        morelegendField[c].setPreferredSize (new Dimension(150,20));
                else
                        morelegendField[c].setPreferredSize (new Dimension(250,20));
                morelegendField[c].setHorizontalAlignment (JTextField.LEFT);
                morelegendField[c].setComponentPopupMenu(legendpopup);
                morelegendField[c].setToolTipText("Right click to set legend preferences");
                mainsubpanelgroup.add (morelegendField[c]);

                if (useWideDataPanel)
                        {
                        curvelabel[c] = new JLabel (""+(c+1));
                        curvelabel[c].setHorizontalAlignment(JLabel.CENTER);
                        mainsubpanelgroup.add (curvelabel[c]);
                        }
                }

        /** Returns an ImageIcon, or null if the path was invalid. */
        static protected ImageIcon createImageIcon(String path, String description) {
            MultiPlot_ m = new MultiPlot_();
            java.net.URL imgURL = m.getClass().getResource(path);
            if (imgURL != null) {
                return new ImageIcon(imgURL, description);
            } else {
                IJ.log("Couldn't find icon file: " + path);
                return null;
            }
}

        static void setVMarkerText()
            {
            GenericDialog gd = new GenericDialog ("Set Vertical Marker Text", mainFrame.getX()+mainFrame.getWidth()/2-165,
                                                                              mainFrame.getY()+mainFrame.getHeight()/2-77);
            gd.addStringField("Vertical Marker 1 Text - Line 1:", vMarker1TopText, 40);
            gd.addStringField("Vertical Marker 1 Text - Line 2:", vMarker1BotText, 40);
            gd.addMessage ("");
            gd.addStringField("Vertical Marker 2 Text - Line 1:", vMarker2TopText, 40);
            gd.addStringField("Vertical Marker 2 Text - Line 2:", vMarker2BotText, 40);
            gd.addMessage ("");
            gd.showDialog();
            if (gd.wasCanceled()) return;
            vMarker1TopText = gd.getNextString();
            vMarker1BotText = gd.getNextString();
            vMarker2TopText = gd.getNextString();
            vMarker2BotText = gd.getNextString();
            Prefs.set("plot.vMarker1TopText", vMarker1TopText);
            Prefs.set("plot.vMarker1BotText", vMarker1BotText);
            Prefs.set("plot.vMarker2TopText", vMarker2TopText);
            Prefs.set("plot.vMarker2BotText", vMarker2BotText);
            }

        static void changePixelScale()
            {
            GenericDialog gd = new GenericDialog ("Change Pixel Scale", mainFrame.getX()+mainFrame.getWidth()/2-165,
                                                                        mainFrame.getY()+mainFrame.getHeight()/2-77);

            gd.addNumericField ("Pixel scale: ",pixelScale,4,8, "(seconds of arc per pixel)");
            gd.addMessage ("");

            gd.showDialog();
            if (gd.wasCanceled()) return;

            pixelScale = gd.getNextNumber();
            Prefs.set("plot.pixelScale", pixelScale);
            changepixelscalemenuitem.setText("Change pixel scale ("+pixelScale+")...");
            }

        static void changeMaxDataLength()
            {
            GenericDialog gd = new GenericDialog ("Change minimum data column length", mainFrame.getX()+mainFrame.getWidth()/2-165,
                                                                                       mainFrame.getY()+mainFrame.getHeight()/2-77);

            gd.addMessage ("The default length is 1000, but can be made smaller or larger\n"
                + "in order to optimize your system's speed and memory utilization.\n"
                + "The length will automatically double if the set size is reached, however\n"
                + "plot updates will be slightly delayed while memory is being reconfigured.");

            gd.addNumericField("Length: ", maxColumnLength, 0, 10, "(measurement table rows)");
            gd.addMessage ("");

            gd.showDialog();
            if (gd.wasCanceled()) return;

            maxColumnLength = (int)gd.getNextNumber();
            setupDataBuffers();
            Prefs.set("plot.maxColumnLength", maxColumnLength);
            }



        static void changePriorityColumns()
            {
            GenericDialog gd = new GenericDialog ("Change data name priorities in pulldown lists", mainFrame.getX()+mainFrame.getWidth()/2-165,
                                                                                                   mainFrame.getY()+mainFrame.getHeight()/2-77);
            gd.addMessage ("Enter data names to be placed at the top of data selection pulldown lists,\n"
                + "in the order desired, with each name separated by a comma (case insensitive).\n"
                + "All data column names starting with a comma separated value will be included.");
            gd.addStringField("Data Columns:", priorityColumns, 80);
//            gd.addStringField("Error Columns:", priorityErrColumns, 80);
            gd.addMessage ("");

            gd.showDialog();
            if (gd.wasCanceled()) return;

            priorityColumns = gd.getNextString();
//            priorityErrColumns = gd.getNextString();
            Prefs.set("plot.priorityColumns", priorityColumns);
////            Prefs.set("plot.priorityErrColumns", priorityErrColumns);
            }

        static void saveAndClose()
                {
                mainFrameLocationX = mainFrame.getLocation().x;
                mainFrameLocationY = mainFrame.getLocation().y;
                Prefs.set("plot2.mainFrameLocationX",mainFrameLocationX);
                Prefs.set("plot2.mainFrameLocationY",mainFrameLocationY);
                if (subFrame.isShowing())
                        {
                        subFrameLocationX = subFrame.getLocation().x;
                        subFrameLocationY = subFrame.getLocation().y;
                        Prefs.set("plot2.subFrameLocationX",subFrameLocationX);
                        Prefs.set("plot2.subFrameLocationY",subFrameLocationY);
                        }
                if (plotImage!=null) 
                    {
                    plotWindow = plotImage.getWindow();
                    if (plotWindow != null)
                            {
                            plotFrameLocationX = plotWindow.getLocation().x;
                            plotFrameLocationY = plotWindow.getLocation().y;
                            Prefs.set("plot2.plotFrameLocationX",plotFrameLocationX);
                            Prefs.set("plot2.plotFrameLocationY",plotFrameLocationY);
                            }
                    }

//                plotImageCanvas.removeMouseMotionListener(plotMouseMotionListener);
//                plotImageCanvas.removeMouseWheelListener(plotMouseWheelListener);
//                plotImageCanvas.removeMouseListener(plotMouseListener);
                
                savePreferences();
                if (delayedUpdateTimer != null) delayedUpdateTimer.cancel();
                if (delayedUpdateTask != null) delayedUpdateTask.cancel();                
                mainFrame.dispose();
                subFrame.dispose();
                mainFrame=null;
                subFrame=null;
                }

       static void openDragAndDropFiles(java.io.File[] files)
                {
                dragAndDropFiles = files;
                if (files.length > 0 && files[0].isFile() && files[0].getName().endsWith(".plotcfg"))
                    {
                    try {
                        InputStream is = new BufferedInputStream(new FileInputStream(files[0].getCanonicalPath()));
                        Prefs.ijPrefs.load(is);
                        is.close();
                        getPreferences();
                        setTable(table, true);
                        plotWindow.setVisible(true);
                        }
                    catch (Exception e) {
                        IJ.beep();
                        IJ.showMessage("Error reading plot configuration file");
                        }
                    }
                else if (files.length > 0 && files[0].isFile() && 
                        (files[0].getName().endsWith(Prefs.get("options.ext", ".xls")) ||
                         files[0].getName().endsWith(".txt") || files[0].getName().endsWith(".csv") ||
                         files[0].getName().endsWith(".prn") || files[0].getName().endsWith(".spc") || 
                         files[0].getName().endsWith(".xls") || files[0].getName().endsWith(".dat") ||
                         files[0].getName().endsWith(".tbl") ))
                    {

                    Thread t = new Thread()
                        {
                        public void run()
                            {
                            
                            IJ.wait(100);

                            try {
                                if (panelShiftDown)
                                    {
                                    appendDataAsRows(true, dragAndDropFiles[0].getCanonicalPath());
                                    } 
                                else if (panelControlDown)
                                    {
                                    appendDataAsColumns(true, dragAndDropFiles[0].getCanonicalPath());
                                    }                         
                                else
                                    {
                                    MeasurementTable newTable = MeasurementTable.getTableFromFile (dragAndDropFiles[0].getCanonicalPath());
                                    if (newTable == null)
                                        {
                                        IJ.beep();
                                        IJ.showMessage("Unable to open MeasurementTable "+dragAndDropFiles[0].getCanonicalPath());
                                        }
                                    else
                                        {
                                        newTable.show();
                                        Frame ntf = WindowManager.getFrame(newTable.shortTitle());
                                        ntf.setVisible(true);
                                        int lastDot = dragAndDropFiles[0].getCanonicalPath().lastIndexOf('.');
                                        String cfgPath = lastDot > 0 ? dragAndDropFiles[0].getCanonicalPath().substring(0, lastDot) + ".plotcfg":
                                                                       dragAndDropFiles[0].getCanonicalPath() + ".plotcfg";
                                        
                                        File cfgFile = new File(cfgPath);
                                        if (cfgFile != null && cfgFile.isFile())
                                            {
                                            InputStream is = new BufferedInputStream(new FileInputStream(cfgPath));
                                            Prefs.ijPrefs.load(is);
                                            is.close();
                                            getPreferences();
                                            }
                                        setTable(newTable, true);
                                        plotWindow.setVisible(true);
                                        }
                                    }
                                }
                            catch (Exception e) {
                                IJ.beep();
                                IJ.showMessage("Error reading measurements file");
                                }
                            }
                            
                        };

                    t.start();
                    Thread.yield();
                    
                    }
                }

        static void openData()
            {
            OpenDialog of = new OpenDialog("Select measurement table to open", "", Prefs.get("options.ext", ".xls"));
            String path = of.getDirectory() + of.getFileName();
            if (of.getDirectory() == null || of.getFileName() == null)
                return;
            try {
                MeasurementTable newTable = MeasurementTable.getTableFromFile (of.getDirectory() + of.getFileName());
                if (newTable == null)
                    {
                    IJ.beep();
                    IJ.showMessage("Unable to open measurement table " + path);
                    }
                else
                    {
                    newTable.show();
                    setTable(newTable, true);
                    plotWindow.setVisible(true);
                    }
                }
            catch (Exception e) {
                IJ.beep();
                IJ.showMessage("Error reading measurements table");
                }
            }
        

        static void appendDataAsRows(boolean fromFile, String inPath)
            {
            String filePath = inPath;
            
            if (fromFile && inPath == null)
                {
                OpenDialog of = new OpenDialog("Select measurement table to append as new rows", "", Prefs.get("options.ext", ".xls"));
                if (of.getDirectory() == null || of.getFileName() == null)
                    return;
                filePath = of.getDirectory() + of.getFileName();
                }

            try {
                MeasurementTable newTable;
                selectAnotherTableCanceled = false;
                if (fromFile)
                    {
                    newTable = MeasurementTable.getTableFromFile (filePath);
                    }
                else
                    {
                    newTable = selectAnotherTable(tableName);
                    }
                if (newTable == null)
                    {
                    if (!selectAnotherTableCanceled)
                        {
                        IJ.beep();
                        if (fromFile)
                            IJ.showMessage("Unable to open measurement table " + filePath);
                        else
                            IJ.showMessage("No additional measurement tables open");
                        }
                    }
                else
                    {
                    newTable.show();
                    
                    combinedTableName = MeasurementTable.shorterName(tableName);
                    if (combinedTableName.endsWith(Prefs.get("options.ext", ".xls")))
                        {
                        int position = combinedTableName.lastIndexOf(Prefs.get("options.ext", ".xls"));
                        if (position >= 0) combinedTableName = combinedTableName.substring(0, position);  
                        }
                    combinedTableName += "_" + MeasurementTable.shorterName(newTable.shortTitle());

                    MeasurementTable combinedTable = MeasurementTable.getTable(combinedTableName, tableName);                    
                    String[] oldColumns = table.getColumnHeadings().split("\t");
                    String[] newColumns = newTable.getColumnHeadings().split("\t");
                    int oldRows = combinedTable.getCounter();
                    int newRows = newTable.getCounter();

                    int combinedRows = oldRows + newRows;
                    int columnIndex = 0;
                    if (newColumns.length > 2 && newRows > 0)
                        {
                        for (int i=0; i<newRows;i++)
                            {
                            combinedTable.incrementCounter();
                            combinedTable.addLabel(newTable.getLabel(i));
                            }
                        
                        for (int i=2;i<oldColumns.length;i++)  //copy all new columns common with old table
                            {
                            columnIndex = newTable.getColumnIndex(oldColumns[i]);
                            for (int j=oldRows; j<combinedRows;j++)
                                {
                                combinedTable.setValue(oldColumns[i], j, columnIndex == ResultsTable.COLUMN_NOT_FOUND?
                                                        Double.NaN : newTable.getValueAsDouble(columnIndex, j-oldRows));
                                }
                            }
                        for (int i=2;i<newColumns.length;i++)  //copy all new columns not common with old table
                            {
                            columnIndex = table.getColumnIndex(newColumns[i]);
                            if (columnIndex == ResultsTable.COLUMN_NOT_FOUND)
                                {
                                for (int j=0; j<combinedRows;j++)
                                    {
                                    combinedTable.setValue(newColumns[i], j, j < oldRows? Double.NaN : newTable.getValue(newColumns[i], j-oldRows));
                                    }                                
                                }             
                            }
                        
                        table.show();
                        combinedTable.show();
                        setTable(combinedTable, true);
                        plotWindow.setVisible(true);
                        }
                    else if (newColumns.length < 3)
                        {
                        IJ.beep();
                        IJ.showMessage("No new columns to append");
                        }
                    else if (newRows < 1)
                        {
                        IJ.beep();
                        IJ.showMessage("No new rows to append");
                        } 
                    else
                        {
                        IJ.beep();
                        IJ.showMessage("Table append error");
                        } 
                    }
                }
            catch (Exception e) {
                IJ.beep();
                IJ.showMessage("Append Rows: Error reading measurement table");
                }
            }      

        static void appendDataAsColumns(boolean fromFile, String inPath)
            {
            String filePath = inPath;
            boolean uniqueHeadings = true;
            
            if (fromFile && inPath == null)
                {
                OpenDialog of = new OpenDialog("Select measurement table to append as new columns", "", Prefs.get("options.ext", ".xls"));
                if (of.getDirectory() == null || of.getFileName() == null)
                    return;
                filePath = of.getDirectory() + of.getFileName();
                }

            try {
                MeasurementTable newTable;
                selectAnotherTableCanceled = false;
                if (fromFile)
                    {
                    newTable = MeasurementTable.getTableFromFile (filePath);
                    }
                else
                    {
                    newTable = selectAnotherTable(tableName);
                    }
                if (newTable == null)
                    {
                    if (!selectAnotherTableCanceled)
                        {
                        IJ.beep();
                        if (fromFile)
                            IJ.showMessage("Unable to open measurement table " + filePath);
                        else
                            IJ.showMessage("No additional measurement tables open");
                        }
                    }
                else
                    {
                    newTable.show();
                    
                    combinedTableName = MeasurementTable.shorterName(tableName);
                    if (combinedTableName.endsWith(Prefs.get("options.ext", ".xls")))
                        {
                        int position = combinedTableName.lastIndexOf(Prefs.get("options.ext", ".xls"));
                        if (position >= 0) combinedTableName = combinedTableName.substring(0, position);  
                        }
                    combinedTableName += "_" + MeasurementTable.shorterName(newTable.shortTitle());

                    if (!appendTableDialog()) return;
                    MeasurementTable combinedTable = MeasurementTable.getTable(combinedTableName, tableName);    
                    String[] oldColumns = table.getColumnHeadings().split("\t");
                    String[] newColumns = newTable.getColumnHeadings().split("\t");
                    int oldRows = combinedTable.getCounter();
                    int newRows = newTable.getCounter();


                    for (int r = 0; r < oldRows; r++)
                        {
                        if (keepFileNamesOnAppend)
                            {                        
                            combinedTable.setLabel(combinedTable.getLabel(r)+(appendDestinationSuffix.equals("") ? " " : "("+appendDestinationSuffix+")") +
                                                (r < newRows ? " "+newTable.getLabel(r)+ "("+appendSourceSuffix+")": ""), r);
                            }
                        else
                            {
                            combinedTable.setLabel("merged", r);
                            }
                        }
                    
                    if (newRows > oldRows)
                        {
                        for (int r = oldRows; r < newRows; r++)
                            {
                            combinedTable.incrementCounter();
                            if (keepFileNamesOnAppend)
                                combinedTable.addLabel(newTable.getLabel(r)+ "("+appendSourceSuffix+")");
                            else
                                combinedTable.addLabel("merged");
                            for (int i=0;i<newColumns.length;i++)
                                {
                                combinedTable.setValue(i, r, Double.NaN);
                                }                            
                            }
                        }
                    if (!appendDestinationSuffix.equals("") && (oldColumns.length > 1))
                        {
                        for (int col=0;col<oldColumns.length - 2;col++)
                            {
                            combinedTable.setHeading(col, combinedTable.getColumnHeading(col)+appendDestinationSuffix);
                            }
                        
                        if (!xlabeldefault.equals("")) xlabeldefault += appendDestinationSuffix;                        
                        for (int curve=0; curve < maxCurves; curve++)
                            {
                            if (!xlabel[curve].equals("") && !xlabel[curve].equals("default")) xlabel[curve] += appendDestinationSuffix;
                            if (!ylabel[curve].equals("")) ylabel[curve] += appendDestinationSuffix;
                            if (!errlabel[curve].equals("")) errlabel[curve] += appendDestinationSuffix;
                            if (!oplabel[curve].equals("")) oplabel[curve] += appendDestinationSuffix;
                            if (!operrlabel[curve].equals("")) operrlabel[curve] += appendDestinationSuffix;
                            }
                        }
                    
                    
                    for (int i=0;i<newColumns.length;i++)
                        {
                        if (combinedTable.getColumnIndex(newColumns[i]+appendSourceSuffix)!=ResultsTable.COLUMN_NOT_FOUND)
                            {
                            uniqueHeadings = false;
                            break;
                            }
                        }
                    int combinedRows = combinedTable.getCounter();
                    if (uniqueHeadings && newColumns.length > 2 && newRows > 0)
                        {
                        for (int i=2;i<newColumns.length;i++)
                            {
                            for (int j=0;j<combinedRows;j++)
                                {
                                if (j < newRows)
                                    combinedTable.setValue(newColumns[i]+appendSourceSuffix, j, newTable.getValue(newColumns[i], j));
                                else
                                    combinedTable.setValue(newColumns[i]+appendSourceSuffix, j, Double.NaN);
                                }
                            }
                        table.show();
                        combinedTable.show();
                        setTable(combinedTable, true);
                        plotWindow.setVisible(true);
                        }
                    else if (!uniqueHeadings)
                        {
                        IJ.beep();
                        IJ.showMessage("Heading names are not unique");
                        } 
                    else if (newColumns.length < 3)
                        {
                        IJ.beep();
                        IJ.showMessage("No new columns to append");
                        }
                    else if (newRows < 1)
                        {
                        IJ.beep();
                        IJ.showMessage("No new rows to append");
                        } 
                    else
                        {
                        IJ.beep();
                        IJ.showMessage("Table append error");
                        } 
                    }
                }
            catch (Exception e) {
                IJ.beep();
                IJ.showMessage("Append Columns: Error reading measurement table");
                }
            }
        
        static boolean appendTableDialog()
            {
            GenericDialog gd = new GenericDialog ("Append settings", mainFrame.getX()+20,
                                                                     mainFrame.getY()+mainFrame.getHeight()/2-77);

            gd.addStringField("Enter name for combined table:", combinedTableName, 80);
            
            gd.addMessage ("The combined dataset column headings must all be unique.\n"+
                           "Enter suffixes to be appended to existing headings to make them unique.\n"+
                           "Either or both of the suffix fields can left blank if the headings are already unique.\n"+
                           "Short suffixes display better in the selection panels.");
            gd.addStringField("Original table column heading suffix:", "_", 80);
            gd.addStringField("Appended table column heading suffix:", "_", 80);
            gd.addCheckbox("Use concatenated source file names as label (label can become very long)***", keepFileNamesOnAppend);
            gd.addMessage("***if deselected, 'merged' will appear in all label columns ");

            gd.showDialog();
            if (gd.wasCanceled()) return false;

            combinedTableName = gd.getNextString().trim();
            appendDestinationSuffix = gd.getNextString().trim();
            appendSourceSuffix = gd.getNextString().trim();
            keepFileNamesOnAppend = gd.getNextBoolean();
            
            return true;
            } 
        
   static MeasurementTable selectAnotherTable(String firstTableName)
        {
        MeasurementTable secondTable;
        String secondTableName = "";
        String[] tables = MeasurementTable.getMeasurementTableNames();
        if (tables == null || tables.length == 0)
            {
            return null;
            }
        else
            {
            //filter out first table and B&C, Log, Errors windows if they exist
            int j = 0;
            for (int i=0; i < tables.length; i++)
                    if (tables[i].equals(firstTableName) || tables[i].equals("B&C") || tables[i].startsWith("Log") || tables[i].equals("Errors"))
                           j++;
            String[] filteredTables = new String[tables.length-j];

            j = 0;
            for (int i=0; i < tables.length; i++)
                if (!tables[i].equals(firstTableName) && !tables[i].equals("B&C") && !tables[i].startsWith("Log") && !tables[i].equals("Errors"))
                    {
                    filteredTables[j] = tables[i];
                    j++;
                    }
            
            if (filteredTables.length == 0)
                {
                return null;
                }
            else                            // IF MORE THAN ONE, ASK WHICH TABLE SHOULD BE USED
                {
                GenericDialog gd = new GenericDialog((firstTableName.equals("") ? "Select table" : "Select table to append"),
                                   mainFrame.getX()+mainFrame.getWidth()/2-165, mainFrame.getY()+mainFrame.getHeight()/2-77);
                gd.addChoice ((firstTableName.equals("") ? "Select table" : "Select table to append"), filteredTables, "Measurements");
                gd.showDialog();
                if (gd.wasCanceled())
                    {
                    selectAnotherTableCanceled = true;
                    return null;
                    }
                selectAnotherTableCanceled = false;
                secondTableName = gd.getNextChoice();
                }
            secondTable = MeasurementTable.getTable(MeasurementTable.longerName(secondTableName));
            return secondTable;
            }
        }           
        
        static void openConfig()
            {
            OpenDialog of = new OpenDialog("Open plot configuration", "", ".plotcfg");
            if (of.getDirectory() == null || of.getFileName() == null)
                return;
            try {
                InputStream is = new BufferedInputStream(new FileInputStream(of.getDirectory()+of.getFileName()));
                Prefs.ijPrefs.load(is);
                is.close();
                getPreferences();
                setTable(table, true);
                plotWindow.setVisible(true);
                }
            catch (Exception e) {
                IJ.beep();
                IJ.showMessage("Error reading plot configuration file");
                }
            }

        static public void openDataAndConfig(String filePath)
            {
            String inPath = filePath;
            if (inPath == null)
                {
                OpenDialog of = new OpenDialog("Select measurement table to open", "", Prefs.get("options.ext", ".xls"));
                if (of.getDirectory() == null || of.getFileName() == null)
                    return;
                inPath = of.getDirectory() + of.getFileName();
                }
            try {
                MeasurementTable newTable = MeasurementTable.getTableFromFile (inPath);
                if (newTable == null)
                    {
                    IJ.beep();
                    IJ.showMessage("Unable to open measurement table " + inPath);
                    }
                else
                    {
                    newTable.show();
                    int lastDot = inPath.lastIndexOf('.');
                    String cfgPath = lastDot > 0 ? inPath.substring(0, lastDot) + ".plotcfg": inPath + ".plotcfg";                    
                    File cfgFile = new File(cfgPath);
                    if (cfgFile != null && cfgFile.isFile())
                        {
                        InputStream is = new BufferedInputStream(new FileInputStream(cfgPath));
                        Prefs.ijPrefs.load(is);
                        is.close();
                        getPreferences();
                        }
                    setTable(newTable, true);
                    plotWindow.setVisible(true);
                    }
                }
            catch (Exception e) {
                IJ.beep();
                IJ.showMessage("Error reading measurements file");
                }
            }
        
        static public void loadDataOpenConfig(MeasurementTable newTable, String filePath)
            {
            String inPath = filePath;
            String cfgPath = "";
            try {
                int lastDot = inPath.lastIndexOf('.');
                cfgPath = lastDot > 0 ? inPath.substring(0, lastDot) + ".plotcfg": inPath + ".plotcfg";                    
                File cfgFile = new File(cfgPath);
                if (cfgFile != null && cfgFile.isFile())
                    {
                    InputStream is = new BufferedInputStream(new FileInputStream(cfgPath));
                    Prefs.ijPrefs.load(is);
                    is.close();
                    getPreferences();
                    }
                Frame ntf = WindowManager.getFrame(newTable.shortTitle());
                if (ntf != null) ntf.setVisible(true);                
                setTable(newTable, true);
                if (plotWindow != null) plotWindow.setVisible(true);
                }
            catch (Exception e) {
                IJ.beep();
                IJ.showMessage("Error reading plot configuration file " + cfgPath);
                }
            }        


        static void saveData()
            {
            if (table == null || table.getLastColumn()==-1)
                {
                IJ.beep();
                IJ.showMessage("No table to save");
                return;
                }
            SaveDialog sf = new SaveDialog("Save measurement table data",MeasurementTable.shorterName(table.shortTitle()), Prefs.get("options.ext", ".xls"));
            if (sf.getDirectory() == null || sf.getFileName() == null)
                return;
            try {
                table.saveAs(sf.getDirectory() + sf.getFileName());
                }
            catch(IOException ioe)
                {
                IJ.beep();
                IJ.showMessage("Error writing measurement table file");
                }
            }

        static void saveDataSubsetDialog()
            {
            int maxSubsetColumns = 20;
            boolean saveColumnHeadings = true;
            boolean saveHeadersAsComment = true; 
            boolean saveRowNumbers = true;
            boolean saveRowLabels = true;
            PrintWriter pw = null;            

            String[] subsetColumn = new String[maxSubsetColumns];
            saveColumnHeadings = Prefs.get("plot2.saveColumnHeadings", saveColumnHeadings);
            saveHeadersAsComment = Prefs.get("plot2.saveHeadersAsComment", saveHeadersAsComment);
            saveRowNumbers = Prefs.get("plot2.saveRowNumbers", saveRowNumbers);
            saveRowLabels = Prefs.get("plot2.saveRowLabels", saveRowLabels);
            
            for (int i=0; i<maxSubsetColumns; i++)
                {
                subsetColumn[i] = "";
                }            
            
            GenericDialog gd = new GenericDialog ("Save data subset", mainFrame.getX()+mainFrame.getWidth()/2-165,
                                                                          mainFrame.getY()+mainFrame.getHeight()/2-77);

            for (int i=0; i<maxSubsetColumns; i++)
                {
                subsetColumn[i] = Prefs.get("plot2.subsetColumn"+i, subsetColumn[i]);
                }
            
            gd.addMessage ("Select datasets in the order to be saved.\nLeave extra selections blank if no more data is to be saved.");
            for (int i=0; i<maxSubsetColumns; i++)
                {
                gd.addChoice("", columns, subsetColumn[i]);
                }
            gd.addCheckbox("Save column headings", saveColumnHeadings);
            gd.addCheckbox("Comment headings with '#'", saveHeadersAsComment);
            gd.addCheckbox("Save row numbers", saveRowNumbers);
            gd.addCheckbox("Save row labels", saveRowLabels);            
            
            gd.showDialog();
            if (gd.wasCanceled()) return;
            
            for (int i=0; i<maxSubsetColumns; i++)
                {
                subsetColumn[i] = gd.getNextChoice();
                }
            saveColumnHeadings = gd.getNextBoolean();
            saveHeadersAsComment = gd.getNextBoolean();
            saveRowNumbers = gd.getNextBoolean();
            saveRowLabels = gd.getNextBoolean();

            for (int i=0; i<maxSubsetColumns; i++)
                {
                Prefs.set("plot2.subsetColumn"+i, subsetColumn[i]);
                }            
            Prefs.set("plot2.saveColumnHeadings", saveColumnHeadings);
            Prefs.set("plot2.saveHeadersAsComment", saveHeadersAsComment);
            Prefs.set("plot2.saveRowNumbers", saveRowNumbers);
            Prefs.set("plot2.saveRowLabels", saveRowLabels);
            
            SaveDialog sf = new SaveDialog("Save data subset", MeasurementTable.shorterName(table.shortTitle()), Prefs.get("options.ext", ".xls"));
            if (sf.getDirectory() == null || sf.getFileName() == null)
                {
                IJ.beep();
                IJ.showMessage("Error: No save file path returned. Aborting.");
                return;
                }
            String savePath = sf.getDirectory() + sf.getFileName();
            int numColumns = 0;
            for (int i=0; i<maxSubsetColumns; i++)
                {
                if (!subsetColumn[i].trim().equals(""))
                    {
                    if (table.getColumnIndex(subsetColumn[i])==ResultsTable.COLUMN_NOT_FOUND)
                        {
                        IJ.beep();
                        IJ.showMessage("Error: Table column "+subsetColumn[i]+" not found. Aborting.");
                        return;
                        }
                    else
                        {
                        numColumns++;
                        }
                    }
                }
            if (numColumns == 0 && !saveRowLabels)
                {
                IJ.beep();
                IJ.showMessage("Error: No table columns selected and not saving row labels. Aborting.");
                return;
                }                
            
            int rows = table.getCounter();
            if (rows==0) 
                {
                IJ.beep();
                IJ.showMessage("Error: No rows in table to save. Aborting.");
                return;
                }

            char delimiter = savePath.endsWith(".csv")?',':'\t';
            String formatString = "0.";
            for (int i=0; i<MeasurementTable.DEFAULT_DECIMALS;i++)
                formatString += "0";
            DecimalFormat output = new DecimalFormat(formatString);
            try{
                FileOutputStream fos = new FileOutputStream(savePath);
                BufferedOutputStream bos = new BufferedOutputStream(fos);
                pw = new PrintWriter(bos);
                String line ="";
                String outText = "";
                double value = 0;
                boolean needDelimiter;
                if (saveColumnHeadings) 
                    {
                    line = (saveHeadersAsComment?"#":"")+(saveRowNumbers?delimiter:"")+(saveRowLabels?"Label"+delimiter:"");
                    needDelimiter = false;
                    for (int i=0; i<maxSubsetColumns; i++)
                        {
                        if (!subsetColumn[i].trim().equals(""))
                            {
                            line += (needDelimiter?delimiter:"")+subsetColumn[i];
                            needDelimiter = true;
                            }
                        }                    
                    pw.println(line);
                    }
                for (int row=0; row<rows; row++)
                    {
                    line = (saveRowNumbers?""+(row+1)+delimiter:"")+(saveRowLabels?table.getLabel(row)+delimiter:"");
                    needDelimiter = false;
                    for (int i=0; i<maxSubsetColumns; i++)
                        {
                        if (!subsetColumn[i].trim().equals(""))
                            {
                            value = table.getValue(subsetColumn[i],row);
                            outText = Double.isNaN(value) ? "":output.format(value);
                            line += ""+(needDelimiter?delimiter:"")+outText;
                            needDelimiter = true;
                            }
                        }                     
                    pw.println(line);
                    }
                }
            catch(IOException ioe) 
                {
                IJ.beep();
                IJ.showMessage("Error writing data subset file");
                }
            finally
                {
                pw.close();
                }            
            
            }
        
        static void savePlotImageAsPng()
            {
            ImagePlus image = WindowManager.getImage("Plot of "+tableName);
            if (image == null)
                {
                IJ.beep();
                IJ.showMessage("No plot image to save");
                return;
                }
            SaveDialog sf = new SaveDialog("Save plot image as PNG...",MeasurementTable.shorterName(table.shortTitle()), ".png");
            if (sf.getDirectory() == null || sf.getFileName() == null) return;
            IJ.runPlugIn(image, "ij.plugin.PNG_Writer", sf.getDirectory()+sf.getFileName());
            }

        static void savePlotImageAsJpg()
            {
            ImagePlus image = WindowManager.getImage("Plot of "+tableName);
            if (image == null)
                {
                IJ.beep();
                IJ.showMessage("No plot image to save");
                return;
                }
            SaveDialog sf = new SaveDialog("Save plot image as JPG...",MeasurementTable.shorterName(table.shortTitle()), ".jpg");
            if (sf.getDirectory() == null || sf.getFileName() == null) return;
            IJ.runPlugIn(image, "ij.plugin.JpegWriter", sf.getDirectory()+sf.getFileName());
            }

        static void saveConfig()
            {
            String fileName = table.shortTitle().startsWith("Measurements in ")? table.shortTitle().replace("Measurements in ", ""):table.shortTitle();
            int lastDot = fileName.lastIndexOf('.');
            if (lastDot >= 0) fileName = fileName.substring(0, lastDot);
            SaveDialog sf = new SaveDialog("Save plot configuration", fileName, ".plotcfg");
            if (sf.getDirectory() == null || sf.getFileName() == null || sf.getFileName().equals(""))
                return;

//            FileDialog fd = new FileDialog(new Frame(), "Save plot configuration...", FileDialog.LOAD);
//            fd.setFile(fileName+".plotcfg");
//            fd.setDirectory(OpenDialog.getDefaultDirectory());
//            fd.setVisible(true);
//            if (fd.getDirectory() == null || fd.getFile() == null || fd.getFile().equals(""))
//                return;

            String savepath = sf.getDirectory() + sf.getFileName();
            lastDot = savepath.lastIndexOf('.');
            if (lastDot >= 0) savepath = savepath.substring(0, lastDot);
            savepath += ".plotcfg";
            Properties prefs = new Properties();
            Enumeration e = Prefs.ijPrefs.keys();
            while (e.hasMoreElements()) {
                String key = (String) e.nextElement();
                if (key.indexOf(".plot.") == 0)
                    prefs.put(key, Prefs.ijPrefs.getProperty(key));
            }

            try
                {
                FileOutputStream fos = new FileOutputStream(savepath);
                BufferedOutputStream bos = new BufferedOutputStream(fos);
                prefs.store(bos, "AstroImageJ Plot Configuration");
                bos.close();
                }
            catch (IOException ioe)
                {
                IJ.beep();
                IJ.showMessage("Error writing plot configuration file");
                }
            }

        static void saveDataConfig()
            {
            SaveDialog sf = new SaveDialog("Save measurement table data and plot configuration",MeasurementTable.shorterName(table.shortTitle()), Prefs.get("options.ext", ".xls"));
            if (sf.getDirectory() == null || sf.getFileName() == null)
                return;
            String path = sf.getDirectory() + sf.getFileName();
            if (table == null || table.getLastColumn()==-1)
                {
                IJ.beep();
                IJ.showMessage("No table to save");
                return;
                }
            else
                {
                try {
                    table.saveAs(path);
                    }
                catch(IOException ioe)
                    {
                    IJ.beep();
                    IJ.showMessage("Error writing measurement table file");
                    }
                }
            int lastDot = path.lastIndexOf('.');
            if (lastDot >= 0) path = path.substring(0, lastDot);
            path += ".plotcfg";
            File outFile = new File(path);
            if (outFile == null || outFile.isDirectory())
                return;
            if (outFile.isFile()) outFile.delete();
            Properties prefs = new Properties();
            Enumeration e = Prefs.ijPrefs.keys();
            while (e.hasMoreElements()) {
                String key = (String) e.nextElement();
                if (key.indexOf(".plot.") == 0)
                    prefs.put(key, Prefs.ijPrefs.getProperty(key));
            }
            try
                {
                FileOutputStream fos = new FileOutputStream(path);
                BufferedOutputStream bos = new BufferedOutputStream(fos);
                prefs.store(bos, "AstroImageJ Plot Configuration");
                bos.close();
                }
            catch (IOException ioe)
                {
                IJ.beep();
                IJ.showMessage("Error writing plot configuration file");
                }
            }

        static public void saveDataImageConfig(boolean savePlot, boolean saveConfig, boolean saveData, boolean filenamesProvided,
                                               String imageFormat, String plotPath, String configPath, String dataPath)
            {
            SaveDialog sf = null;
            String outPath = "";
            String outBase = "";
            if (!filenamesProvided)
                {
                sf = new SaveDialog("Save "+(saveData?"measurement table data":"")+(saveData&&savePlot?", ":"")+
                    (savePlot?"plot image":"")+(saveData&&savePlot&&saveConfig?",":"")+((saveData||savePlot)&&saveConfig?" and ":"")+
                    (saveConfig?"plot configuration":""),MeasurementTable.shorterName(table.shortTitle()), "");
                if (sf.getDirectory() == null || sf.getFileName() == null)
                    return;
                outPath = sf.getDirectory() + sf.getFileName();
                int location = outPath.lastIndexOf('.');
                if (location >= 0) outPath = outPath.substring(0, location);
                outBase = outPath;
                if (outBase.endsWith(plotSuffix))
                    {
                    location = outBase.lastIndexOf(plotSuffix);
                    if (location >= 0) outBase = outBase.substring(0, location);
                    }
                else if (outBase.endsWith(configSuffix))
                    {
                    location = outBase.lastIndexOf(configSuffix);
                    if (location >= 0) outBase = outBase.substring(0, location);
                    }
                else if (outBase.endsWith(dataSuffix))
                    {
                    location = outBase.lastIndexOf(dataSuffix);
                    if (location >= 0) outBase = outBase.substring(0, location);
                    }
                }

            if (saveData)
                {
                if (table == null || table.getLastColumn()==-1)
                    {
                    IJ.beep();
                    IJ.showMessage("No data table to save");
                    }
                else
                    {
                    try {
                        table.saveAs(filenamesProvided?dataPath:outBase+dataSuffix+Prefs.get("options.ext", ".xls"));
                        }
                    catch(IOException ioe)
                        {
                        IJ.beep();
                        IJ.showMessage("Error writing measurement table file");
                        }
                    }
                }

            if (savePlot)
                {
                String imagepath = filenamesProvided ? plotPath : outBase+plotSuffix + "."+imageFormat;
                ImagePlus image = WindowManager.getImage("Plot of "+tableName);
                if (image == null)
                    {
                    IJ.beep();
                    IJ.showMessage("No plot image to save");
                    }
                else if (imageFormat.toLowerCase().equals("png"))
                    {
                    IJ.runPlugIn(image, "ij.plugin.PNG_Writer", imagepath);
                    }
                else if (imageFormat.toLowerCase().equals("jpg"))
                    {
                    IJ.runPlugIn(image, "ij.plugin.JpegWriter", imagepath);
                    }
                }
            if (saveConfig)
                {
                String cfgpath = filenamesProvided ? configPath : outBase+configSuffix+".plotcfg";
                File outFile = new File(cfgpath);
                if (outFile == null || outFile.isDirectory())
                    {
                    IJ.error("bad configuration save filename");
                    return;
                    }
                if (outFile.isFile()) outFile.delete();
                Properties prefs = new Properties();
                Enumeration e = Prefs.ijPrefs.keys();
                while (e.hasMoreElements()) {
                    String key = (String) e.nextElement();
                    if (key.indexOf(".plot.") == 0)
                        prefs.put(key, Prefs.ijPrefs.getProperty(key));
                }
                try
                    {
                    FileOutputStream fos = new FileOutputStream(cfgpath);
                    BufferedOutputStream bos = new BufferedOutputStream(fos);
                    prefs.store(bos, "AstroImageJ Plot Configuration");
                    bos.close();
                    }
                catch (IOException ioe)
                    {
                    IJ.beep();
                    IJ.showMessage("Error writing plot configuration file");
                    }
                }
            }
        
    

        static void setSaveStateDialog()
            {
            GenericDialog gd = new GenericDialog ("File suffix settings", mainFrame.getX()+mainFrame.getWidth()/2-165,
                                                                          mainFrame.getY()+mainFrame.getHeight()/2-77);

            plotSuffix = Prefs.get("Astronomy_Tool.plotSuffix", plotSuffix);
            configSuffix = Prefs.get("Astronomy_Tool.configSuffix", configSuffix);
            dataSuffix = Prefs.get("Astronomy_Tool.dataSuffix", dataSuffix);

            gd.addMessage ("Enter save file suffix settings.");
            gd.addStringField("Plot image suffix:", plotSuffix, 40);
            gd.addStringField("Plot config file suffix:", configSuffix, 40);
            gd.addStringField("Data table file suffix:", dataSuffix, 40);
            gd.addMessage ("Tip: make plot config and data table suffix the same so that the plot config");
            gd.addMessage ("will auto-load when a new data table file is opened by drag and drop.");

            gd.showDialog();
            if (gd.wasCanceled()) return;

            plotSuffix = gd.getNextString();
            configSuffix = gd.getNextString();
            dataSuffix = gd.getNextString();

            Prefs.set("Astronomy_Tool.plotSuffix", plotSuffix);
            Prefs.set("Astronomy_Tool.configSuffix", configSuffix);
            Prefs.set("Astronomy_Tool.dataSuffix", dataSuffix);
            }
        
        static void openHelpPanel()
            {
            String filename = "help/multiplot_help.htm";
            new HelpPanel(filename, "Multi-plot");
            }      
        static void openDataHelpPanel()
            {
            String filename = "help/multiplot_data_help.htm";
            new HelpPanel(filename, "Data Naming Convention");
            }         

        static void getPreferences()
                {
                tableName = Prefs.get("plot2.tableName",tableName);
                keepFileNamesOnAppend=Prefs.get("plot2.keepFileNamesOnAppend", keepFileNamesOnAppend);                
                plotAutoMode=Prefs.get("plot.automode",plotAutoMode);
                xlabeldefault=Prefs.get("plot.xlabeldefault",xlabeldefault);
                title=Prefs.get("plot.title", title);
                useTitle=Prefs.get("plot.useTitle", useTitle);
                subtitle=Prefs.get("plot.subtitle", subtitle);
                useSubtitle=Prefs.get("plot.useSubtitle", useSubtitle);
                titlePosX=Prefs.get("plot.titlePosX",titlePosX);
                titlePosY=Prefs.get("plot.titlePosY",titlePosY);
                subtitlePosX=Prefs.get("plot.subtitlePosX",subtitlePosX);
                subtitlePosY=Prefs.get("plot.subtitlePosY",subtitlePosY);
                legendPosX=Prefs.get("plot.legendPosX",legendPosX);
                legendPosY=Prefs.get("plot.legendPosY",legendPosY);
                xTics=Prefs.get("plot.xTics", xTics);
                yTics=Prefs.get("plot.yTics", yTics);
                xGrid=Prefs.get("plot.xGrid", xGrid);
                yGrid=Prefs.get("plot.yGrid", yGrid);
                xNumbers=Prefs.get("plot.xNumbers", xNumbers);
                yNumbers=Prefs.get("plot.yNumbers", yNumbers);
                autoScaleX=Prefs.get("plot.autoScaleX",autoScaleX);
                useFirstX=Prefs.get("plot.useFirstX",useFirstX);
                autoScaleY=Prefs.get("plot.autoScaleY",autoScaleY);
                xMin=Prefs.get("plot.xMin", xMin);
                xMax=Prefs.get("plot.xMax", xMax);
                xWidth=Prefs.get("plot.xWidth", xWidth);
                yMin=Prefs.get("plot.yMin", yMin);
                yMax=Prefs.get("plot.yMax", yMax);

                plotSizeX=(int)Prefs.get("plot.plotSizeX", plotSizeX);
                plotSizeY=(int)Prefs.get("plot.plotSizeY", plotSizeY);
                mainFrameLocationX=(int)Prefs.get("plot2.mainFrameLocationX", mainFrameLocationX);
                mainFrameLocationY=(int)Prefs.get("plot2.mainFrameLocationY", mainFrameLocationY);
                subFrameLocationX=(int)Prefs.get("plot2.subFrameLocationX", subFrameLocationX);
                subFrameLocationY=(int)Prefs.get("plot2.subFrameLocationY", subFrameLocationY);
                plotFrameLocationX=(int)Prefs.get("plot2.plotFrameLocationX", plotFrameLocationX);
                plotFrameLocationY=(int)Prefs.get("plot2.plotFrameLocationY", plotFrameLocationY);
                openDataSetWindow=Prefs.get("plot.openDataSetWindow", openDataSetWindow);
                rememberWindowLocations=Prefs.get("plot.rememberWindowLocations", rememberWindowLocations);
                usePixelScale=Prefs.get("plot.usePixelScale", usePixelScale);
                pixelScale=Prefs.get("plot.pixelScale", pixelScale);
                mmagrefs=(int)Prefs.get("plot.mmagrefs", mmagrefs);
                xExponent=(int)Prefs.get("plot.xExponent", xExponent);
                yExponent=(int)Prefs.get("plot.yExponent", yExponent);
                xWidthStep=Prefs.get("plot.xWidthStep", xWidthStep);
                showToolTips=Prefs.get("astroIJ.showToolTips",showToolTips);
                ToolTipManager.sharedInstance().setEnabled(showToolTips);
                xMaxStep=Prefs.get("plot.xMaxStep", xMaxStep);
                xMinStep=Prefs.get("plot.xMinStep", xMinStep);
                yMaxStep=Prefs.get("plot.yMaxStep", yMaxStep);
                yMinStep=Prefs.get("plot.yMinStep", yMinStep);
                vMarker1Value=Prefs.get("plot.vMarker1Value", vMarker1Value);
                vMarker2Value=Prefs.get("plot.vMarker2Value", vMarker2Value);
                vMarker1Step=Prefs.get("plot.vMarker1Step", vMarker1Step);
                vMarker2Step=Prefs.get("plot.vMarker2Step", vMarker2Step);
                showVMarker1=Prefs.get("plot.showVMarker1", showVMarker1);
                showVMarker2=Prefs.get("plot.showVMarker2", showVMarker2);
                vMarker1TopText=Prefs.get("plot.vMarker1TopText", vMarker1TopText);
                vMarker1BotText=Prefs.get("plot.vMarker1BotText", vMarker1BotText);
                vMarker2TopText=Prefs.get("plot.vMarker2TopText", vMarker2TopText);
                vMarker2BotText=Prefs.get("plot.vMarker2BotText", vMarker2BotText);
                invertYAxis=Prefs.get("plot.invertYAxis", invertYAxis);
                invertYAxisSign=invertYAxis?-1:1;
                negateMag=Prefs.get("plot.negateMag", negateMag);
                plotSizeXStep=(int)Prefs.get("plot.plotSizeXStep", plotSizeXStep);
                plotSizeYStep=(int)Prefs.get("plot.plotSizeYStep", plotSizeYStep);
                excludedHeadSamples=(int)Prefs.get("plot.excludedHeadSamples", excludedHeadSamples);
                excludedTailSamples=(int)Prefs.get("plot.excludedTailSamples", excludedTailSamples);
                excludedHeadSamplesStep=(int)Prefs.get("plot.excludedHeadSamplesStep", excludedHeadSamplesStep);
                excludedTailSamplesStep=(int)Prefs.get("plot.excludedTailSamplesStep", excludedTailSamplesStep);
                useXColumnName=Prefs.get("plot.useXColumnName", useXColumnName);
                useYColumnName=Prefs.get("plot.useYColumnName", useYColumnName);
                useXCustomName=Prefs.get("plot.useXCustomName", useXCustomName);
                useYCustomName=Prefs.get("plot.useYCustomName", useYCustomName);
                xLegend=Prefs.get("plot.xLegend", xLegend);
                yLegend=Prefs.get("plot.yLegend", yLegend);
                showXScaleInfo=Prefs.get("plot.showXScaleInfo", showXScaleInfo);
                showYScaleInfo=Prefs.get("plot.showYScaleInfo", showYScaleInfo);
                showYShiftInfo=Prefs.get("plot.showYShiftInfo", showYShiftInfo);
                showLScaleInfo=Prefs.get("plot.showLScaleInfo", showLScaleInfo);
                showLShiftInfo=Prefs.get("plot.showLShiftInfo", showLShiftInfo);
                showYBinInfo=Prefs.get("plot.showYBinInfo", showYBinInfo);
                showLBinInfo=Prefs.get("plot.showLBinInfo", showLBinInfo);
                showYmmagInfo=Prefs.get("plot.showYmmagInfo", showYmmagInfo);
                showLmmagInfo=Prefs.get("plot.showLmmagInfo", showLmmagInfo);
                showYSymbolInfo=Prefs.get("plot.showYSymbolInfo", showYSymbolInfo);
                showLSymbolInfo=Prefs.get("plot.showLSymbolInfo", showLSymbolInfo);
                useTwoLineLegend=Prefs.get("plot.useTwoLineLegend", useTwoLineLegend);
                useWideDataPanel=Prefs.get("plot.useWideDataPanel", useWideDataPanel);
                useBoldedDatum=Prefs.get("plot.useBoldedDatum", useBoldedDatum);
                legendLeft=Prefs.get("plot.legendLeft", legendLeft);
                legendRight=Prefs.get("plot.legendRight", legendRight);
                priorityColumns=Prefs.get("plot.priorityColumns", priorityColumns);
//                priorityErrColumns=Prefs.get("plot.priorityErrColumns", priorityErrColumns);
                prioritizeColumns=Prefs.get("plot.prioritizeColumns", prioritizeColumns);
                maxColumnLength=(int)Prefs.get("plot.maxColumnLength", 1000);

                plotSuffix = Prefs.get("Astronomy_Tool.plotSuffix", plotSuffix);
                configSuffix = Prefs.get("Astronomy_Tool.configSuffix", configSuffix);
                dataSuffix = Prefs.get("Astronomy_Tool.dataSuffix", dataSuffix);

                for (int i=0; i<maxCurves; i++)
                        {
                        moreOptions[i]=Prefs.get("plot.moreOptions"+i,moreOptions[i]);
//                        errlabel[i]=Prefs.get("plot.errlabel"+i,errlabel[i]);
                        oplabel[i]=Prefs.get("plot.oplabel"+i, oplabel[i]);
//                        operrlabel[i]=Prefs.get("plot.operrlabel"+i,operrlabel[i]);
//                        showOpErrors[i]=Prefs.get("plot.showOpErrors"+i,showOpErrors[i]);
                        xlabel[i]=Prefs.get("plot.xlabel"+i,xlabel[i]);
                        ylabel[i]=Prefs.get("plot.ylabel"+i,ylabel[i]);
                        lines[i]=Prefs.get("plot.lines"+i, lines[i]);
                        markerIndex[i]=(int)Prefs.get("plot.markerIndex"+i,markerIndex[i]);
                        colorIndex[i]=(int)Prefs.get("plot.colorIndex"+i, colorIndex[i]);
                        mmag[i]=Prefs.get("plot.mmag"+i, mmag[i]);
                        useColumnName[i]=Prefs.get("plot.useColumnName"+i, useColumnName[i]);
                        useLegend[i]=Prefs.get("plot.useLegend"+i, useLegend[i]);
                        legend[i] = Prefs.get("plot.legend"+i, legend[i]);
                        operatorIndex[i]=(int)Prefs.get("plot.operatorIndex"+i, operatorIndex[i]);
                        binSize[i]=(int)Prefs.get("plot.binSize"+i, binSize[i]);
                        plotY[i] = Prefs.get("plot.plotY"+i,plotY[i]);
                        force[i] = Prefs.get("plot.force"+i,force[i]);
                        showErrors[i] = Prefs.get("plot.showErrors"+i,showErrors[i]);
                        autoScaleFactor[i]=Prefs.get("plot.autoScaleFactor"+i,autoScaleFactor[i]);
                        autoScaleStep[i]=Prefs.get("plot.autoScaleStep"+i,autoScaleStep[i]);
                        autoShiftFactor[i]=Prefs.get("plot.autoShiftFactor"+i,autoShiftFactor[i]);
                        autoShiftStep[i]=Prefs.get("plot.autoShiftStep"+i,autoShiftStep[i]);
                        customScaleFactor[i]=Prefs.get("plot.customScaleFactor"+i,customScaleFactor[i]);
                        customScaleStep[i]=Prefs.get("plot.customScaleStep"+i,customScaleStep[i]);
                        customShiftFactor[i]=Prefs.get("plot.customShiftFactor"+i,customShiftFactor[i]);
                        customShiftStep[i]=Prefs.get("plot.customShiftStep"+i,customShiftStep[i]);
                        ASInclude[i]=Prefs.get("plot.ASInclude"+i,ASInclude[i]);
                        }
                }
        
        
        
        static void savePreferences()
                {
                Prefs.set("plot2.tableName",tableName);
                Prefs.set("plot2.keepFileNamesOnAppend", keepFileNamesOnAppend);
                Prefs.set("plot.useDefaultSettings", useDefaultSettings);
                Prefs.set("plot.xlabeldefault",xlabeldefault);
                Prefs.set("plot.xTics", xTics);
                Prefs.set("plot.yTics", yTics);
                Prefs.set("plot.xGrid", xGrid);
                Prefs.set("plot.yGrid", yGrid);
                Prefs.set("plot.xNumbers", xNumbers);
                Prefs.set("plot.yNumbers", yNumbers);
                Prefs.set("plot.autoScaleX", autoScaleX);
                Prefs.set("plot.autoScaleY", autoScaleY);
                Prefs.set("plot.useFirstX", useFirstX);
                Prefs.set("plot.xWidth", xWidth);
                Prefs.set("plot.xMin", xMin);
                Prefs.set("plot.xMax", xMax);
                Prefs.set("plot.yMin", yMin);
                Prefs.set("plot.yMax", yMax);
                Prefs.set("plot.vMarker1Value", vMarker1Value);
                Prefs.set("plot.vMarker2Value", vMarker2Value);
                Prefs.set("plot.vMarker1Step", vMarker1Step);
                Prefs.set("plot.vMarker2Step", vMarker2Step);
                Prefs.set("plot.showVMarker1", showVMarker1);
                Prefs.set("plot.showVMarker2", showVMarker2);
                Prefs.set("plot.vMarker1TopText", vMarker1TopText);
                Prefs.set("plot.vMarker1BotText", vMarker1BotText);
                Prefs.set("plot.vMarker2TopText", vMarker2TopText);
                Prefs.set("plot.vMarker2BotText", vMarker2BotText);
                Prefs.set("plot.invertYAxis", invertYAxis);
                Prefs.set("plot.negateMag", negateMag);
                Prefs.set("plot.plotSizeX",plotSizeX);
                Prefs.set("plot.plotSizeY",plotSizeY);
                Prefs.set("plot.title",title);
                Prefs.set("plot.useTitle",useTitle);
                Prefs.set("plot.subtitle",subtitle);
                Prefs.set("plot.useSubtitle",useSubtitle);
                Prefs.set("plot.titlePosX",titlePosX);
                Prefs.set("plot.titlePosY",titlePosY);
                Prefs.set("plot.subtitlePosX",subtitlePosX);
                Prefs.set("plot.subtitlePosY",subtitlePosY);
                Prefs.set("plot.openDataSetWindow",openDataSetWindow);
                Prefs.set("plot.legendPosX",legendPosX);
                Prefs.set("plot.legendPosY",legendPosY);
                Prefs.set("plot.rememberWindowLocations",rememberWindowLocations);
                Prefs.set("plot.usePixelScale",usePixelScale);
                Prefs.set("plot.pixelScale",pixelScale);
                Prefs.set("plot.mmagrefs",mmagrefs);
                Prefs.set("plot.xExponent",xExponent);
                Prefs.set("plot.yExponent",yExponent);
                Prefs.set("plot.xWidthStep",xWidthStep);
                Prefs.set("plot.xMaxStep",xMaxStep);
                Prefs.set("plot.xMinStep",xMinStep);
                Prefs.set("plot.yMaxStep",yMaxStep);
                Prefs.set("plot.yMinStep",yMinStep);
                Prefs.set("plot.vMarker1Step",vMarker1Step);
                Prefs.set("plot.vMarker2Step",vMarker2Step);
                Prefs.set("plot.plotSizeXStep",plotSizeXStep);
                Prefs.set("plot.plotSizeYStep",plotSizeYStep);
                Prefs.set("plot.excludedHeadSamples", excludedHeadSamples);
                Prefs.set("plot.excludedTailSamples", excludedTailSamples);
                Prefs.set("plot.excludedHeadSamplesStep", excludedHeadSamplesStep);
                Prefs.set("plot.excludedTailSamplesStep", excludedTailSamplesStep);
                Prefs.set("plot.useXColumnName", useXColumnName);
                Prefs.set("plot.useYColumnName", useYColumnName);
                Prefs.set("plot.useXCustomName", useXCustomName);
                Prefs.set("plot.useYCustomName", useYCustomName);
                Prefs.set("plot.xLegend", xLegend);
                Prefs.set("plot.yLegend", yLegend);
                Prefs.set("plot.showXScaleInfo", showXScaleInfo);
                Prefs.set("plot.showYScaleInfo", showYScaleInfo);
                Prefs.set("plot.showYShiftInfo", showYShiftInfo);
                Prefs.set("plot.showLScaleInfo", showLScaleInfo);
                Prefs.set("plot.showLShiftInfo", showLShiftInfo);
                Prefs.set("plot.showYBinInfo", showYBinInfo);
                Prefs.set("plot.showLBinInfo", showLBinInfo);
                Prefs.set("plot.showYmmagInfo", showYmmagInfo);
                Prefs.set("plot.showLmmagInfo", showLmmagInfo);
                Prefs.set("plot.showYSymbolInfo", showYSymbolInfo);
                Prefs.set("plot.showLSymbolInfo", showLSymbolInfo);
                Prefs.set("plot.showLSymbolInfo", showLSymbolInfo);
                Prefs.set("plot.useTwoLineLegend", useTwoLineLegend);
                Prefs.set("plot.useBoldedDatum", useBoldedDatum);
                Prefs.set("plot.legendLeft", legendLeft);
                Prefs.set("plot.legendRight", legendRight);
                Prefs.set("plot.priorityColumns", priorityColumns);
//                Prefs.set("plot.priorityErrColumns", priorityErrColumns);
                Prefs.set("plot.prioritizeColumns", prioritizeColumns);

                for (int i=0; i<maxCurves; i++)
                        {
                        Prefs.set("plot.xlabel"+i,xlabel[i]);
                        Prefs.set("plot.ylabel"+i,ylabel[i]);
//                        Prefs.set("plot.errlabel"+i,errlabel[i]);
                        Prefs.set("plot.oplabel"+i,oplabel[i]);
//                        Prefs.set("plot.operrlabel"+i,operrlabel[i]);
//                        Prefs.set("plot.showOpErrors"+i,showOpErrors[i]);
                        Prefs.set("plot.lines"+i, lines[i]);
                        Prefs.set("plot.markerIndex"+i,markerIndex[i]);
                        Prefs.set("plot.colorIndex"+i,colorIndex[i]);
                        Prefs.set("plot.mmag"+i, mmag[i]);
                        Prefs.set("plot.useColumnName"+i,useColumnName[i]);
                        Prefs.set("plot.useLegend"+i,useLegend[i]);
                        Prefs.set("plot.legend"+i,legend[i]);
                        Prefs.set("plot.operatorIndex"+i,operatorIndex[i]);
                        Prefs.set("plot.moreOptions"+i,moreOptions[i]);
                        Prefs.set("plot.binSize"+i,binSize[i]);
                        Prefs.set("plot.plotY"+i,plotY[i]);
                        Prefs.set("plot.force"+i,force[i]);
                        Prefs.set("plot.showErrors"+i,showErrors[i]);
                        Prefs.set("plot.autoScaleFactor"+i,autoScaleFactor[i]);
                        Prefs.set("plot.autoScaleStep"+i,autoScaleStep[i]);
                        Prefs.set("plot.autoShiftFactor"+i,autoShiftFactor[i]);
                        Prefs.set("plot.autoShiftStep"+i,autoShiftStep[i]);
                        Prefs.set("plot.customScaleFactor"+i,customScaleFactor[i]);
                        Prefs.set("plot.customScaleStep"+i,customScaleStep[i]);
                        Prefs.set("plot.customShiftFactor"+i,customShiftFactor[i]);
                        Prefs.set("plot.customShiftStep"+i,customShiftStep[i]);
                        Prefs.set("plot.ASInclude"+i,ASInclude[i]);
                        }
                }
	}

